package com.caringo.client;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.util.Map;

import org.apache.commons.httpclient.Header;
import org.apache.commons.httpclient.auth.AuthChallengeParser;
import org.apache.commons.httpclient.auth.AuthPolicy;
import org.apache.commons.httpclient.auth.MalformedChallengeException;


/**
 * BasicGetScspResponseHandler is a simple ScspResponseOutputStream implementation. If bodyOutputStream is given in the constructor,
 * the content from an OK (200) or Partial (206) response will be written to it, otherwise the response will end up in the
 * bodyString attribute. There's no sanity-checking done here, so it's dumb enough to try to store a multi-GB body in memory if you
 * don't provide the bodyOutputStream. <br>
 * <br>
 * Copyright (c) 2008 by Caringo, Inc. -- All rights reserved<br>
 * This is free software, distributed under the terms of the New BSD license.<br>
 * See the LICENSE.txt file included in this archive.<br>
 * <br>
 * @author jmike
 * @created Jan 19, 2008
 * @id $Id: BasicScspResponseOutputStream.java 52121 2012-08-28 17:31:56Z jafshar $
 */
public class BasicScspResponseOutputStream extends ScspResponseOutputStream {

    private int statusCode;
    private int retryCount;
    private Header[] headers;
    private Header[] trailers;
    private String statusLine;
    private long contentLength;
    private String bodyString;
    private String authenticationChallengeRealm;
    private String authenticationChallengeNonce;

    private OutputStream bodyOutputStream;
    private OutputStream bodySink;
    private long bodyRemaining;
    private boolean chunky;


    // /
    // / Construction
    // /

    public BasicScspResponseOutputStream(OutputStream bodyOutputStream) {
        super();
        this.bodyOutputStream = bodyOutputStream;
        this.retryCount = -1;
    }

    public BasicScspResponseOutputStream() {
        this(null);
    }

    // /
    // / ScspResponseHandler implementation
    // /

    /**
     * @see com.caringo.client.ScspResponseOutputStream#responseComplete()
     */
    @Override
    public void responseComplete(int retries) {
        if (bodySink != bodyOutputStream) {
            this.bodyString = ((ByteArrayOutputStream) bodySink).toString();
        }
        this.retryCount = retries;
    }

    /**
     * @throws IOException
     * @see com.caringo.client.ScspResponseOutputStream#bodyDataReceived(byte[], int, int)
     */
    @Override
    public void bodyDataReceived(byte[] data, int offset, int len) throws IOException {

        // new logic to get a better msg than assert.
        if (bodyRemaining < len && !chunky) {
            String msg = "Error attempting to receive body data.\n";
            msg += "bodyRemaining.:" + bodyRemaining + "\n";
            msg += "len...........:" + len + "\n";
            msg += "bodyRemaining must be <= len at all times.\n";
            throw new IOException(msg);
        }

        // original logic
        // assert bodyRemaining >= len;
        bodySink.write(data, offset, len);
        bodyRemaining -= len;
    }

    /**
     * @see com.caringo.client.ScspResponseOutputStream#responseReceived(int, org.apache.commons.httpclient.Header[], String, long)
     */
    @Override
    public boolean responseReceived(int statusCode, Header[] headers, String statusLine, long responseBodyLength) {
        this.statusCode = statusCode;
        this.headers = headers;
        this.bodyRemaining = responseBodyLength;
        this.contentLength = responseBodyLength;
        this.statusLine = statusLine;
        this.chunky = false;
        for (Header header:headers) {
        	if (header.getName().compareToIgnoreCase("transfer-encoding") == 0 &&
        		header.getValue().compareToIgnoreCase("chunked") == 0) {
        		chunky = true;
        		break;
        	}
        }
        	
        try {
            Map challenges = AuthChallengeParser.parseChallenges(headers);
            if (challenges.containsKey(AuthPolicy.DIGEST)) {
                Map challengeParams = AuthChallengeParser.extractParams((String)challenges.get(AuthPolicy.DIGEST));
                if (challengeParams.containsKey("realm")) {
                    this.authenticationChallengeRealm = (String)challengeParams.get("realm");
                }
                if (challengeParams.containsKey("nonce")) {
                    this.authenticationChallengeNonce = (String)challengeParams.get("nonce");
                }
            }
        }
        catch (MalformedChallengeException ex) {
            // We don't want to throw if the server returned something ugly, so silently fail and empty the authentication strings
            this.authenticationChallengeRealm = "";
            this.authenticationChallengeNonce = "";
        }
        
        // Just assume that we get a 206 partial content only if we asked for it.
        if ((statusCode == ScspStatus.SC_OK || statusCode == ScspStatus.SC_PARTIAL_CONTENT)
                && bodyOutputStream != null) {
            bodySink = bodyOutputStream;
        } else {
            bodySink = new ByteArrayOutputStream();
        }
        
        // Check for If-{Un}Modified-Since empty body codes. If-Modified-Since responds with a 304 if the 
        // content hasn't been modified since the date given.
        // Addresses an issue found in Bug 5179.
        // Fixing bug introduced in prior 5179 fix. We should NOT discard the body from a 412. Fixes bug 5218.
        if (statusCode == ScspStatus.SC_NOT_MODIFIED) {
        	bodyRemaining = 0;
        }
        return true;
    }
    /**
     * @see com.caringo.client.ScspResponseOutputStream#footersReceived(org.apache.commons.httpclient.Header[])
     */
    @Override
    public void footersReceived(Header[] footers) {
    	this.trailers = footers;
    }
    // /
    // / Getters/setters
    // /

    public String getBodyString() {
        return bodyString;
    }

    /**
     * This method will return the value for the specified header name (case insensitive).
     * Note that this will search both the headers and footers.
     * 
     * @param name
     *            - the header name
     * @return - A String object containing the specified header name. Case insensitive.
     */
    public String getHeader(String name) {
        String result = null;
        if (headers != null && headers.length > 0) {
            for (Header header : headers) {
                if (header.getName().equalsIgnoreCase(name)) {
                    return header.getValue();
                }
            }
        }
        if (trailers != null && trailers.length > 0) {
            for (Header header : trailers) {
                if (header.getName().equalsIgnoreCase(name)) {
                    return header.getValue();
                }
            }
        }
        return result;
    }

    /**
     * This method will return all the headers, both real headers and footers/trailer headers.
     * We don't currently have a need to split out the headers from the footers, so we're returning
     * the combined lists here to make it easier on clients. If in the future we do
     * need to return separate collections, do it here and also fix getHeader above.
     * 
     * @return - An array containing all headers.
     */
    public ScspHeaders getHeaders() {
        ScspHeaders result = new ScspHeaders();
        for (Header header:headers) {
            result.addValue(header.getName(), header.getValue());
        }
        if (trailers != null) {
	        for (Header header:trailers) {
	            result.addValue(header.getName(), header.getValue());
	        }
        }
        return result;
    }

    /**
     * this method will return the status code for the http method's execution.
     * 
     * @return - The status code for the http method's execution.
     */
    public int getStatusCode() {
        return statusCode;
    }

    /**
     * this method will return the number of times the request was retried.
     * 
     * @return - The retry count for the request execution.
     */
    public int getRetryCount() {
        return retryCount;
    }

    /**
     * not sure if this method works... if you want content length, then check headers. (this method should work, but not spending
     * the time now to figure out the problem.
     * 
     * @return - The content length.
     */
    public long getContentLength() {
        return contentLength;
    }
    
    public String toString()
    {
    	StringBuilder sb = new StringBuilder(getStatusLine());
		String eol = System.getProperty("line.separator");
		sb.append(eol);
		for (Header header:headers)
		{
			sb.append(header.toExternalForm());
		}
		sb.append(eol);

		sb.append(eol);
		if (bodySink != bodyOutputStream && contentLength < 2000)
    	{
    		// this seems very cheesy. There ought to be a better way of pulling off 
    		// the body when it's information
    		sb.append(bodySink.toString());
    	}
    	return sb.toString();
    }

	/**
	 * @return the statusLine
	 */
	public String getStatusLine() {
		return statusLine;
	}

    public void setAuthenticationChallengeRealm(String authenticationChallengeRealm) {
        this.authenticationChallengeRealm = authenticationChallengeRealm;
    }

    public String getAuthenticationChallengeRealm() {
        return authenticationChallengeRealm;
    }

    /**
     * @param authenticationChallengeNonce the authenticationChallengeNonce to set
     */
    public void setAuthenticationChallengeNonce(String authenticationChallengeNonce) {
        this.authenticationChallengeNonce = authenticationChallengeNonce;
    }

    /**
     * @return the authenticationChallengeNonce
     */
    public String getAuthenticationChallengeNonce() {
        return authenticationChallengeNonce;
    }

}
