/**
 * 
 */

package com.caringo.client;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.RandomAccessFile;

/**
 * Copyright (c) 2010 by Caringo, Inc. -- All rights reserved<br>
 * This is free software, distributed under the terms of the New BSD license.<br>
 * See the LICENSE.txt file included in this archive.<br>
 * <br>
 * 
 * @author pray
 * @created Aug 24, 2010
 * @id $Id: $
 */
 
public class ResettableFileInputStream extends InputStream {
 
	RandomAccessFile f;
	long mark;

     /**
      * Create a ResettableFileInputStream from a File object
      * 
      * @param file
      *            the file to be opened for reading.
      */
	public ResettableFileInputStream(File file) throws IOException {
		this.f = new RandomAccessFile(file, "r");
        this.mark = 0;
	}

     /**
      * Create a ResettableFileInputStream from a file name
      * 
      * @param name
      *            the name of the file to be opened for reading.
      */
    public ResettableFileInputStream(String name) throws IOException {
        this.f = new RandomAccessFile(name, "r");
        this.mark = 0;
    }

    /**
     * Mark the current position of the input file.
     *
     * Note that this ignores readlimit, which means the mark is never invalidated.
     * 
     * @see java.util.io.InputStream#mark(int readlimit)
     */
	@Override
	public void mark(int readlimit) {
        // ignore readLimit, which means the mark is never invalidated.
		try {
			mark = f.getFilePointer();
		} catch (IOException e) {
			// ignore IOExceptions. Will be caught on later calls.
		}
	}
 
    /**
     * Return whether mark is supported for this stream type.
     *
     * Always returns true.
     * 
     * @see java.util.io.InputStream#markSupported()
     */
	@Override
	public boolean markSupported() {
		return true;
	}

    /**
     * Read a byte from the file.
     * 
     * @see java.util.io.InputStream#read()
     */
	@Override
	public int read() throws IOException {
		return f.read();
	}
 
    /**
     * Read from input file.
     *
     * @see java.util.io.InputStream#read(byte[] b, int off, int len)
     */
	@Override
	public int read(byte[] b) throws IOException {
		return f.read(b);
	}

    /**
     * Read from input file.
     *
     * @see java.util.io.InputStream#read(byte[] b, int off, int len)
     */
	@Override
	public int read(byte[] b, int off, int len) throws IOException {
		return f.read(b, off, len);
	}
 
    /**
     * Reset the input file.
     *
     * @see java.util.io.InputStream#reset()
     */
	@Override
	public void reset() throws IOException {
		f.seek(this.mark);
	}
 
    /**
     * Return the number of bytes left to read from input file.
     *
     * @see java.util.io.InputStream#available()
     */     
	@Override
	public int available() throws IOException {
		return (int) (f.length() - f.getFilePointer());
	}
 
    /**
     * Close the input file.
     *
     * @see java.util.io.InputStream#close()
     */     
	@Override
	public void close() throws IOException {
		f.close();
	}
 
    /**
     * Return the number of bytes left to read from input file.
     *
     * @see java.util.io.InputStream#available()
     */     
	@Override
	public long skip(long n) throws IOException {
		long s = f.getFilePointer() + n;
		if (s > f.length()) {
            n = f.length() - f.getFilePointer();
        }
    	f.seek(f.getFilePointer() + n);
		return n;
	}
}

