/**
 * 
 */
package com.caringo.client;

import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Helper class for parsing and storing a WWW-Authenticate challenge and holding challenge reply user name and password.
 * <p>
 * Copyright (c) 2009 by Caringo, Inc. -- All rights reserved<br>
 * This is free software, distributed under the terms of the New BSD license.<br>
 * See the LICENSE.txt file included in this archive.<br>
 * <br>
 * The client only needs to supply the user name, realm, and password in a new
 * ScspAuthentication object if it's not trying to preemptively authenticate.
 * If it is retrying on the same resource with a fresh (that is, non-stale) nonce,
 * it can parse the ScspAuthentication object from a response, fill in the user
 * name and password, and pass the object back into a new request for the same
 * resource.
 * <br>
 * Note that the realm and nonce values are currently unused elsewhere in this package. These can be useful to implementers using
 * other http libraries than commons HttpClient 3.x.
 * 
 * @author pray
 * @created September 20, 2009
 * @id $Id: ScspAuthentication.java 52121 2012-08-28 17:31:56Z jafshar $
 */
public class ScspAuthentication {

    private String userName;
    private String password;
    private String realm;
    private String nonce;

    /**
     * Construct a <code>ScspAuthentication</code> object with user name, password, realm, and nonce.
     * 
     * @param userName -
     *            authentication user name
     * @param password -
     *            authentication password
     * @param realm -
     *            authentication realm
     * @param nonce -
     *            authentication nonce
     */
    public ScspAuthentication(String userName, String password, String realm, String nonce) {
        this.userName = userName;
        this.password = password;
        this.realm = realm;
        this.nonce = nonce;
    }

    // constructor for use when building an authentication object from a challenge header
    // 
    private ScspAuthentication(String realm, String nonce) {
        this.realm = realm;
        this.nonce = nonce;
        this.userName = null;
        this.password = null;
    }

    /**
     * Construct an ScspAuthentication object from a WWW-Authenticate challenge header. Retrives the realm and nonce from the header
     * value.
     * 
     * @param headerValue -
     *            Contents of the WWW-Authenticate: following the ':'.
     * 
     * @return a new ScspAuthentication with the realm and nonce filled in if the header can be parsed; otherwise, null.
     */
    public static ScspAuthentication fromAuthenticateHeader(String headerValue) {
        // the header looks like
        String realm = "";
        String nonce = "";
        Pattern realmPattern = Pattern.compile("realm=(\"[^\"]+\")");
        Matcher realmMatcher = realmPattern.matcher(headerValue);
        if (realmMatcher.find()) {
            realm = realmMatcher.group(1);
        }
        Pattern noncePattern = Pattern.compile("nonce=(\"[^\"]+\")");
        Matcher nonceMatcher = noncePattern.matcher(headerValue);
        if (realmMatcher.find()) {
            nonce = nonceMatcher.group(1);
        }

        if (realm.length() > 0 && nonce.length() > 0) {
            return new ScspAuthentication(realm, nonce);
        } else {
            return null;
        }
    }

    /**
     * Set the user name
     * 
     * @param userName -
     *            the userName to set
     */
    public void setUserName(String userName) {
        this.userName = userName;
    }

    /**
     * Get the user name
     * 
     * @return the userName
     */
    public String getUserName() {
        return userName;
    }

    /**
     * Set the password.
     * 
     * @param password
     *            the password to set
     */
    public void setPassword(String password) {
        this.password = password;
    }

    /**
     * Get the password.
     * 
     * @return the password
     */
    public String getPassword() {
        return password;
    }

    /** 
     * Set the realm.
     *  
     * @param realm
     *            the realm to set
     */
    public void setRealm(String realm) {
        this.realm = realm;
    }

    /**
     * Get the realm.
     * 
     * @return the realm
     */
    public String getRealm() {
        return realm;
    }

    /**
     * Set the nonce value.
     * 
     * @param nonce
     *            the nonce to set
     */
    public void setNonce(String nonce) {
        this.nonce = nonce;
    }

    /**
     * Get the nonce value.
     * 
     * @return the nonce
     */
    public String getNonce() {
        return nonce;
    }

}
