/**
 * 
 */
package com.caringo.client;

import java.util.Map;
import java.util.HashMap;
import java.util.ArrayList;
import java.lang.StringBuffer;
import java.lang.IllegalArgumentException;
import java.util.Iterator;
import java.io.UnsupportedEncodingException;

import com.caringo.client.request.ScspRequestHandler;
import com.caringo.client.ScspHeaders;

/**
 * Properly formats an SCSP authorization header
 * 
 * To retrieve a list of validation or parsing errors, check the 'errors' member,
 * which will contain a list of errors.  This class can be used to format an SCSP
 * authorization header entry to be added to an ScspHeaders object, or can also
 * parse authorization information from a response header.
 * <p>
 * It is recommended that the client calls validate() on an ScspAuthorization object
 * to reduce potential ambiguity due to duplicate or invalid authorization entries.
 * <p>
 * Copyright (c) 2010 by Caringo, Inc. -- All rights reserved<br>
 * This is free software, distributed under the terms of the New BSD license.<br>
 * See the LICENSE.txt file included in this archive.<br>
 * <br>
 * 
 * @author bguetzlaff
 * @created October 29, 2010
 * @id $Id: ScspAuthorization.java 52121 2012-08-28 17:31:56Z jafshar $
 */
public class ScspAuthorization {

    public static final String AUTHORIZATION_HEADER_NAME = "Castor-Authorization";

    // Generic operations that may have authorization applied to them
    public static final String ALL_OP = "";
    public static final String VIEW_OP = "view";
    public static final String CHANGE_OP = "change";

    // Specific operations that may have authorization applied to them
    public static final String PUT_OP = "put";
    public static final String COPY_OP = "copy";
    public static final String APPEND_OP = "append";
    public static final String GET_OP = "get";
    public static final String HEAD_OP = "head";
    public static final String DELETE_OP = "delete";
    public static final String POST_OP = "post";

    // Convenience realms
    public static final String ANY_REALM = "";

    private HashMap<String, ArrayList<String> > ops;
    private ArrayList<String> errors;

    /**
     * Construct a <code>ScspAuthorization</code> object.
     */
    public ScspAuthorization() {
        this.ops = new HashMap<String, ArrayList<String> >();
        this.errors = new ArrayList<String>();
    }

    public void addAuthorization(String operation, String realm) throws IllegalArgumentException {
        if (null == operation) {
            operation = ScspAuthorization.ALL_OP;
        }
        String workingop = operation.toLowerCase();
            
        if (null == realm) {
            realm = ANY_REALM;
        }
        
        if ((ScspAuthorization.ALL_OP == operation) && (ANY_REALM == realm)) {
            throw new IllegalArgumentException("Empty rules are not allowed; either a specific operation or a specific realm must be provided");
        }
            
        if (!ops.containsKey(workingop)) {
            ops.put(workingop, new ArrayList<String>());
        }
        
        ops.get(workingop).add(realm);
    }
            
    public void addAuthorization(String operation, ArrayList<String> realms) throws IllegalArgumentException { // allows x.addAuthorization(op1, y.getAuthorization(op2))
        if (null == operation) {
            operation = ScspAuthorization.ALL_OP;
        }
        String workingop = operation.toLowerCase();
        
        if (!ops.containsKey(workingop)) {
            ops.put(workingop, new ArrayList<String>());
        }
        
        if (realms.isEmpty()) {
            throw new IllegalArgumentException("At least one realm must be provided");
        }
        
        ops.get(workingop).addAll(realms);
    }
            
    public void resetAuthorization(String operation) {
        String workingop = operation.toLowerCase();
        if (ops.containsKey(workingop)) {
            ops.remove(workingop);
        }
    }
            
    public boolean hasAuthorization(String operation) {
        return ops.containsKey(operation.toLowerCase());
    }
    
    /**
     * Returns a list of realms authorized for the given operation
     */
    public ArrayList<String> getAuthorization(String operation) {
        return ops.get(operation.toLowerCase());
    }
        
    /**
     * Returns the full SCSP authorization header.  If an encoding exception occurred while formatting the header,
     * then an empty string will be returned.  Call getErrors() for details.
     */
    public String toString() {
        return AUTHORIZATION_HEADER_NAME + ": " + getAuthSpec();
    }
    
    public final ArrayList<String> getErrors() {
        return errors;
    }
        
    public void resetErrors() {
        errors.clear();
    }
    
    private String join(ArrayList<String> coll, String delim) {
        if (coll.isEmpty()) {
            return "";
        }
        
        Iterator<String> iter = coll.iterator();
        StringBuffer buffer = new StringBuffer(iter.next());
        while (iter.hasNext()) {
            buffer.append(delim).append(iter.next());
        }
        
        return buffer.toString();
    }
        
    /**
     * Returns ONLY the authorization specification portion of the full SCSP authorization header.
     * If an encoding exception occurred while formatting the header, then an empty string will be
     * returned.  Call getErrors() for details.
     */
    public String getAuthSpec() {
        ArrayList<String> authSpecs = new ArrayList<String>();
        ArrayList<Byte> dontEscape = new ArrayList<Byte>();
        
        dontEscape.add((byte)' ');
        dontEscape.add((byte)':');
        dontEscape.add((byte)',');
        
        for (Map.Entry<String, ArrayList<String>> authSpec : ops.entrySet()) {
            for (String realm : authSpec.getValue()) {
                try {
                    realm = ScspRequestHandler.pathEscape(realm, dontEscape);
                } catch (UnsupportedEncodingException ex) {
                    errors.add("Caught an UnsupportedEncodingException while creating auth spec: " + ex.toString());
                    return "";
                }
                
                if (ScspAuthorization.ALL_OP == authSpec.getKey()) {
                    authSpecs.add(realm);
                } else {
                    authSpecs.add(authSpec.getKey() + "=" + realm);
                }
            }
        }
                    
        return join(authSpecs, ", ");
    }
    
    private static ArrayList<String> splitKeepAll(String text, String delim) {
        ArrayList<String> result = new ArrayList<String>();
        int oldIndex = 0;
        int newIndex = text.indexOf(delim, oldIndex);
        
        while (-1 != newIndex) {
            result.add(text.substring(oldIndex, newIndex));
            oldIndex = newIndex + delim.length();
            newIndex = text.indexOf(delim, oldIndex);
        }
        
        if (oldIndex <= text.length()) {
            result.add(text.substring(oldIndex));
        }
        
        return result;
    }
        
    private static void parseInto(String header, ScspAuthorization authObj) {
        HashMap<String, ArrayList<String> > ops = authObj.ops;
        ArrayList<String> parseErrors = authObj.errors;
        
        try {
            // If we were given the full header, lop off the header name and ":"
            if (header.startsWith(AUTHORIZATION_HEADER_NAME)) {
                header = header.substring(AUTHORIZATION_HEADER_NAME.length() + 1);
            }
            
            for (String spec : splitKeepAll(header, ",")) {
                spec = spec.trim();
                
                String op;
                String realm;
                
                if (spec.contains("=")) {
                    if (spec == "=") {
                        op = ScspAuthorization.ALL_OP;
                        realm = ANY_REALM;
                    } else {
                        String[] pair = spec.split("=", 2);
                        op = pair[0];
                        realm = pair[1];
                    }
                } else {
                    op = ScspAuthorization.ALL_OP;
                    realm = spec;
                }
                
                op = op.trim().toLowerCase();
                realm = realm.trim();
                
                if ((ScspAuthorization.ALL_OP == op) && (ANY_REALM == realm)) {
                    parseErrors.add("Empty authorization specification (\",=,\") found when parsing header (" + header + ")");
                }
                
                if (ops.containsKey(op)) {
                    ops.get(op).add(realm);
                    parseErrors.add("Duplicate realm (\"" + realm + "\") defined for operation (\"" + op + "\"); authorization behavior will be ambiguous");
                } else {
                    ops.put(op, new ArrayList<String>());
                    ops.get(op).add(realm);
                }
            }
        } catch (Exception ex) {
            parseErrors.add("Error while parsing authorization header string (" + header + "): " + ex.toString());
        }
    }
    
    /**
     * If the Castor-Authorization header is present in the given headers, this will create a corresponding
     * ScspAuthorization object.
     * 
     * It is recommended that you check the result of getErrors() on the returned ScspAuthorization object for
     * any parsing errors, as these can be indicative of ambiguous authorization rules.  It is also highly
     * recommended that you call validate() on the returned object, as validate() is able to catch more
     * errors.
     */
    public static ScspAuthorization fromHeaders(ScspHeaders headers) {
        if (!headers.containsName(AUTHORIZATION_HEADER_NAME)) {
            return new ScspAuthorization();
        }
            
        ScspAuthorization result = new ScspAuthorization();
        
        for (String header : headers.getHeaderValues(AUTHORIZATION_HEADER_NAME)) {
            parseInto(header, result);
        }
            
        return result;
    }
    
    /**
     * Parses the given Castor-Authorization header into an ScspAuthorization object.
     * 
     * It is recommended that you check the result of getErrors() on the returned ScspAuthorization object for
     * any parsing errors, as these can be indicative of ambiguous authorization rules.  It is also highly
     * recommended that you call validate() on the returned object, as validate() is able to catch more
     * errors.
     */
    public static ScspAuthorization fromString(String header) {
        ScspAuthorization result = new ScspAuthorization();
        parseInto(header, result);
        return result;
    }
        
    /**
     * Ensures that the operations and realms within this object are of a valid format.
     * 
     * @return true if validation was successful, false otherwise.  If false is returned, it is recommended that you
     *      check the result of getErrors() on the ScspAuthorization object for validation errors.
     */
    public boolean validate() {
        ArrayList<String> illegalRealmChars = new ArrayList<String>();
        illegalRealmChars.add(":");
        illegalRealmChars.add(",");
        
        ArrayList<String> validOps = new ArrayList<String>();
        validOps.add(ScspAuthorization.ALL_OP);
        validOps.add(ScspAuthorization.VIEW_OP);
        validOps.add(ScspAuthorization.CHANGE_OP);
        validOps.add(ScspAuthorization.PUT_OP);
        validOps.add(ScspAuthorization.COPY_OP);
        validOps.add(ScspAuthorization.APPEND_OP);
        validOps.add(ScspAuthorization.GET_OP);
        validOps.add(ScspAuthorization.HEAD_OP);
        validOps.add(ScspAuthorization.DELETE_OP);
        validOps.add(ScspAuthorization.POST_OP);
        
        resetErrors();
        
        try {
            for (Map.Entry<String, ArrayList<String>> entry : ops.entrySet()) {
                String op = entry.getKey();
                ArrayList<String> realms = entry.getValue();

                if (!validOps.contains(op)) {
                    errors.add("Unsupported operation (\"" + op + "\") for authorization");
                }

                if (1 < realms.size()) {
                    errors.add("Multiple realms ([\"" + join(realms, "\", \"") + "\"]) defined for operation (\"" + op + "\"); authorization behavior will be ambiguous");
                }

                if (1 > realms.size()) {
                    // Shouldn't happen, but for completeness...
                    errors.add("Missing realm for operation (\"" + op + "\"); authorization behavior will be ambiguous");
                }

                for (String realm : realms) {
                    if ((op == ScspAuthorization.ALL_OP) && (realm == ANY_REALM)) {
                        errors.add("Empty authorization specification (ALL_OP=ANY_REALM, or \"=\") found");
                    }
                        
                    for (String illegal : illegalRealmChars) {
                        if (realm.contains(illegal)) {
                            errors.add("Illegal character ('" + illegal + "') found in realm (\"" + realm + "\") for operation (\"" + op + "\")");
                        }
                    }
                }
            }
        } catch (Exception ex) {
            errors.add("Error while validating ScspAuthorization object: " + ex.toString());
        }
            
        return errors.isEmpty();
    }

}
