package com.caringo.client;

import com.caringo.client.request.ScspRequestHandler;
import org.apache.commons.httpclient.params.HttpClientParams;

/**
 * Abstract base class for all SCSP commands.<br>
 * <p>
 * Copyright (c) 2009 by Caringo, Inc. -- All rights reserved<br>
 * This is free software, distributed under the terms of the New BSD license.<br>
 * See the LICENSE.txt file included in this archive.<br>
 * <br>
 * 
 * @author pray
 * @created September 22, 2009
 * @id ${Id}
 */
public abstract class ScspCommand {

    private ScspRequestHandler client;
    private ScspHeaders headers;
    private ScspQueryArgs queryArgs;
    private String path;

    // Construct a ScspCommand object with a request handler
    //
    protected ScspCommand(ScspRequestHandler client) {
        this.client = client;
        this.headers = new ScspHeaders();
        this.queryArgs = new ScspQueryArgs();
        this.path = "";
    }

    // Build a list of query args. Override to build the query args for a specific command.
    // 
    // returns a List of NameValuePairs to use for building query args.
    //    
    final protected ScspRequestHandler getPreparedClient() {
        ScspAuthentication auth = headers.getAuthentication();
        if (headers.getAuthentication() != null) {
            client.setAuthentication(auth.getUserName(), auth.getPassword(), auth.getNonce(), auth.getRealm());
        } else {
            client.setAuthentication("", "", "", "");
        }
        return client;
    }

    /**
     * Execute the command.
     * 
     * Implementations of this method should execute the command and retrieve the headers and body.
     * 
     * @throws ScspExecutionException -
     *             Exception indicating that something bad happened talking to CAStor.
     */
    public abstract ScspResponse execute() throws ScspExecutionException;

    /**
     * Set the query args for the command.
     * 
     * @param queryArgs
     *            the <code>ScspQueryArgs</code> to set
     */
    public void setQueryArgs(ScspQueryArgs queryArgs) {
    	// do a copy so that shared queryArgs instances outside this class aren't stomped on.
        this.queryArgs = new ScspQueryArgs();
        if (queryArgs != null) {
        	this.queryArgs.addAll(queryArgs.getArgList());
        }
    }

    /**
     * Get the query args from the command.
     * 
     * @return the <code>ScspQueryArgs</code> queryArgs
     */
    public ScspQueryArgs getQueryArgs() {
        return queryArgs;
    }
    
    /**
     * Set the path for the command.
     * 
     * @param path
     *            the path to set
     */
    public void setPath(String path) {
    	if(path == null) {
    		this.path = "";
    	}
    	else {
    		this.path = path;
    	}
    }

    /**
     * Get the path from the command.
     * 
     * @return the path
     */
    public String getPath() {
        return this.path;
    }

    /**
     * Set the user agent to pass in HTTP requests to CAStor.
     * 
     * @param userAgent -
     *            the user agent to pass to CAStor.
     */
    public void setUserAgent(String userAgent) {
        this.client.setUserAgent(userAgent);
    }

    /**
     * Get the user agent.
     * 
     * @return the user agent.
     */
    public String getUserAgent() {
        return this.client.getUserAgent();
    }

    /**
     * Set the headers for the command.
     * 
     * @param headers
     *            the <code>ScspHeaders</code> to set
     */
    public void setHeaders(ScspHeaders headers) {
    	// do a copy so that shared headers instances outside this class aren't stomped on.
        this.headers.removeAll();
        this.headers.setAuthentication(null);
        if (headers != null) {
        	this.headers.addAll(headers);
            this.headers.setAuthentication(headers.getAuthentication());
        }
    }

    /**
     * Get the headers from the command.
     * 
     * @return the <code>ScspHeaders</code> headers.
     */
    public ScspHeaders getHeaders() {
        return headers;
    }

    /**
     * Validate that this command agrees with its headers and query args.
     * 
     * @return true if this comprises a valid SCSP command and false if it doesn't.
     */
    public boolean validate() {
        return ScspValidations.validateHost(this.headers) &&
               ScspValidations.validateDomain(this.queryArgs) &&
               ScspValidations.validateAdmin(this.queryArgs);
    }

}
