package com.caringo.client;

import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.TimeZone;
import java.lang.IllegalArgumentException;

/**
 * Helper class for converting dates into <code>RFC2616</code> dates where required, such as in Policies and Cache-Coherency
 * values.
 * <p>
 * Copyright (c) 2010 by Caringo, Inc. -- All rights reserved.
 * This is free software, distributed under the terms of the New BSD license.<br>
 * See the LICENSE.txt file included in this archive.<br>
 * <br>
 * 
 * @author jmike
 * @created Oct 24, 2008
 * @id $Id: ScspDate.java 52121 2012-08-28 17:31:56Z jafshar $
 */
public class ScspDate implements Comparable<Object> {

    public static final DateFormat DATE_FORMAT = new SimpleDateFormat("EEE, dd MMM yyyy HH:mm:ss z");
    static {
        DATE_FORMAT.setTimeZone(TimeZone.getTimeZone("GMT"));
    }

    private Date date = null;

    /**
     * Construct an absolute date.
     * 
     * @param absoluteDate -
     *            The absolute date.
     * @throws IllegalArgumentException
     *             If an invalid date is given
     */
    public ScspDate(Date absoluteDate) throws IllegalArgumentException {
        this.date = normalizeDate(absoluteDate);
    }

    /**
     * Construct a relative date.
     * 
     * @param offset -
     *            Offset from "now", in seconds
     * @throws IllegalArgumentException
     *             If an invalid offset is given
     */
    public ScspDate(long offset) throws IllegalArgumentException {
        Calendar cal = Calendar.getInstance();
        cal.setTimeInMillis(System.currentTimeMillis() + (offset * 1000));
        this.date = normalizeDate(cal.getTime());
    }
    
    /**
     * Construct a date, either absolute or relative. If absolute, it must be specified in <code>RFC 2616</code> format.
     * 
     * @param dateStr -
     *            A <code>String</code> representation of the date.
     * @throws ParseException
     *             If a parsing exception occurs.
     * @throws NumberFormatException
     *             If a number formatting exception occurs.
     */
    public ScspDate(String dateStr) throws ParseException, NumberFormatException {
        this.date = DATE_FORMAT.parse(dateStr);
    }

    /**
     * Return the Policy date as a <code>Date</code> object.
     * 
     * @return - The policy date as a <code>Date</code> object.
     */
    public Date toDate() {
        return date;
    }

    /**
     * Return the date as a formatted string.
     * 
     * @return A <code>String</code> object.
     */
    @Override
    public String toString() {
        return DATE_FORMAT.format(toDate());
    }

    /*
     * (non-Javadoc)
     */
    @Override
    public int hashCode() {
        final int PRIME = 31;
        int result = 1;
        result = PRIME * result + date.hashCode();
        return result;
    }

    /**
     * Returns whether the two dates represent the same date.
     * 
     * @return <code>true</code> if obj is a ScspDate and represents the same date and time.
     */
    @Override
    public boolean equals(Object obj) {
        if (this == obj)
            return true;
        if (obj == null)
            return false;
        if (getClass() != obj.getClass())
            return false;
        final ScspDate other = (ScspDate) obj;
        
        return (date.equals(other.date));
    }

    /*
     * (non-Javadoc)
     */
    public int compareTo(Object o) {
        int result = 0;
        if (o == null) { // null == terminal which is > all other dates
            result = -1;
        } else if (o != this) {
            ScspDate other = (ScspDate) o;
            result = this.date.compareTo(other.date);
        }
        return result;
    }

    /*
     * Normalizes a date, removing any extraneous information (truncates milliseconds, pretty much)
     */
    private Date normalizeDate(Date raw) throws IllegalArgumentException {
        try {
            return DATE_FORMAT.parse(DATE_FORMAT.format(raw));
        } catch (ParseException ex) {
            throw new IllegalArgumentException("Failed to normalize date due to: " + ex.getMessage());
        } catch (NumberFormatException ex) {
            throw new IllegalArgumentException("Failed to normalize date due to: " + ex.getMessage());
        }
    }

}
