/**
 * 
 */
package com.caringo.client;

import java.util.HashMap;
import java.util.Map;
import java.util.List;
import java.util.ArrayList;

/**
 * Encapsulation of a set of HTTP query or response headers. Provides a way to store and retrieve header names with one or more
 * values and methods for parsing and building header strings.<br>
 * Note that headers are stored and retrieved case-insensitive.
 * <p>
 * Copyright (c) 2009 by Caringo, Inc. -- All rights reserved<br>
 * This is free software, distributed under the terms of the New BSD license.<br>
 * See the LICENSE.txt file included in this archive.<br>
 * <br>
 * 
 * @author pray
 * @created September 20, 2009
 * @id ${Id}
 */
public final class ScspHeaders {
    private HashMap<String, ArrayList<String>> headers;
    private ScspAuthentication authentication;

    /**
     * Default constructor. Constructs an object with an empty header map.
     */
    public ScspHeaders() {
        headers = new HashMap<String, ArrayList<String>>();
        authentication = null;
    }

    /**
     * Replace all of the headers with copies from the input headers.
     * 
     * @param headers -
     *            A Map of Strings->List of Strings to copy into me.
     */
    public void setHeaders(HashMap<String, ArrayList<String>> headers) {
        this.headers.clear();
        copyHeadersFromMap(headers);
    }

    /**
     * Return the headers as a map of strings to lists of string.
     * 
     * Updating this map will not affect the values stored in the ScspHeaders instance.
     * 
     * @return a map (string->list of strings) of the headers.
     */
    public HashMap<String, ArrayList<String>> getHeaderMap() {
        HashMap<String, ArrayList<String>> copy = new HashMap<String, ArrayList<String>>();
        copy.putAll(headers);
        return copy;
    }

    /**
     * Return the header values for a particular header name.
     * 
     * @param name -
     *            the name of the header to retrieve the values for.
     * @return a list of strings containing the header values for the name. If null, the header name is not stored.
     */
    public ArrayList<String> getHeaderValues(String name) {
        String headerName = getMapNameIgnoreCase(name);
        if (headerName != null) {
            return headers.get(headerName);
        } else {
            return null;
        }
    }

    /**
     * Returns whether there are any headers with a name matching the input name.
     * 
     * @return true if there are any headers with the input name.
     */
    public boolean containsName(String name) {
        return getMapNameIgnoreCase(name) != null;
    }

    /**
     * Retrieve the header names that match the input prefix
     * 
     * @param prefix -
     *            Prefix to check
     * 
     * @return list of matching header names
     */
    public ArrayList<String> getNamesWithPrefix(String prefix) {
        String workingPrefix = prefix.toUpperCase();
        ArrayList<String> result = new ArrayList<String>();
        for (Map.Entry<String, ArrayList<String>> header : headers.entrySet()) {
            String workingHeader = header.getKey().toUpperCase();
            if (workingHeader.startsWith(workingPrefix)) {
                result.add(header.getKey());
            }
        }
        return result;
    }

    /**
     * Add a header value to the list of headers for name. Does not remove any headers already set for name. Will add a new header
     * if the header doesn't already exist.
     * 
     * @param name -
     *            header name to set.
     * @param value -
     *            the value to set.
     */
    public void addValue(String name, String value) {
        addEntryIgnoreCase(name, value);
    }

    /**
     * Replace a header value in the list of headers for name. Removes any headers already set for name. Will add a new header
     * if the header doesn't already exist.
     * 
     * @param name -
     *            header name to set.
     * @param value -
     *            the value to set.
     */
    public void replaceValue(String name, String value) {
        remove(name);
        addEntryIgnoreCase(name, value);
    }

    /**
     * Copies all of the headers from other into this. This method does not remove any existing header names of values.
     * 
     * @param other -
     *            header object to copy into this.
     */
    public void addAll(ScspHeaders other) {
        copyHeadersFromMap(other.headers);
    }

    /**
     * Removes all entries that have a matching header name.
     * 
     * @param name -
     *            header name to remove.
     */
    public void remove(String name) {
        String headerName = getMapNameIgnoreCase(name);
        if (headerName != null) {
            headers.remove(headerName);
        }
    }

    /**
     * Removes all of the stored headers.
     */
    public void removeAll() {
        headers.clear();
    }

    /**
     * Retrieves all the headers as a list of <code>ScspHeader</code>.
     * 
     * @return list of <code>ScspHeader</code>.
     */
    public ArrayList<ScspHeader> toHeaderList() {
        ArrayList<ScspHeader> headerList = new ArrayList<ScspHeader>();
        for (Map.Entry<String, ArrayList<String>> mapEntry : headers.entrySet()) {
            for (String singleValue : mapEntry.getValue()) {
                headerList.add(new ScspHeader(mapEntry.getKey(), singleValue));
            }
        }
        return headerList;
    }

    /**
     * Add a range header.
     * 
     * @param start -
     *            Start of range. If value is < 0, then start string value for the range will be empty,
     *            indicating a read from the end of the stream. Ex: addRange(-1, 30) reads the last 30 bytes.
     * @param end -
     *            End of range. If value < 0, then end string value for the range will be empty,
     *            indicating a read until the end of the stream. Ex: addRange(30, -1) reads from byte 30 on.
     */
    public void addRange(long start, long end) {
        String sStart;
        String sEnd;
        if (start < 0) {
            sStart = "";
        } else {
            sStart = String.valueOf(start);
        }
        if (end < 0) {
            sEnd = "";
        } else {
            sEnd = String.valueOf(end);
        }
        String value = "bytes=" + sStart + "-" + sEnd;
        addValue("Range", value);
    }

    /**
     * Add a SCSP Lifepoint header.
     * 
     * @param date -
     *            Lifepoint date. Null indicates empty ('[]') lifepoint date
     * @param deleteConstraint
     * @param reps -
     *            Null indicates an empty reps value; Only values > 0 will be used
     * @param custom -
     *            A <code>List</code> of <code>String</code> containing any custom lifepoint attributes. This parameter may be <code>null</code>.
     */
    public void addLifepoint(ScspDate date, ScspDeleteConstraint deleteConstraint, Integer reps, List<String> custom) {
        ScspLifepoint lifepoint = new ScspLifepoint(date, reps, deleteConstraint, custom);

        addValue("Lifepoint", lifepoint.toString());
    }

    /**
     * Add a SCSP Lifepoint header.
     * 
     * @param date -
     *            Lifepoint date. Null indicates empty ('[]') lifepoint date
     * @param deleteConstraint
     * @param reps -
     *            Null indicates an empty reps value; Only values > 0 will be used
     */
    public void addLifepoint(ScspDate date, ScspDeleteConstraint deleteConstraint, Integer reps) {
        ScspLifepoint lifepoint = new ScspLifepoint(date, reps, deleteConstraint);

        addValue("Lifepoint", lifepoint.toString());
    }

    /**
     * Retrieve the count of distinct (case-insensitive) header names.
     * 
     * @return the number of header names.
     */
    public int nameCount() {
        return headers.size();
    }

    /**
     * Retrieve the number of header values.
     * 
     * @return the number of header values.
     */
    public int headerCount() {
        // slow, but it works
        return toHeaderList().size();
    }

    /**
     * Return whether the header name and value are stored.
     * 
     * @param name - the header name
     * @param value - the header value
     * @return <code>true</code> if there is a header with the matching name and value.
     */
    public boolean containsValue(String name, String value) {
        if (containsName(name)) {
            return getHeaderValues(name).contains(value);
        } else {
            return false;
        }
    }

    /**
     * Return whether this equals other.
     * 
     * @param other - the headers to compare to.
     * @return - True if everything in other is in me, and everything in me is in other; false otherwise. Note that this does not
     *         attempt to pull apart compound headers, thus {("H", "a,b")} is not equal to {("H", "a"), ("H", "b")} even though they
     *         are (most likely) equivalent for HTTP.
     */
    public boolean equals(ScspHeaders other) {
        for (Map.Entry<String, ArrayList<String>> myEntry : headers.entrySet()) {
            for (String value : myEntry.getValue()) {
                if (!other.containsValue(myEntry.getKey(), value)) {
                    return false;
                }
            }
        }
        for (Map.Entry<String, ArrayList<String>> otherEntry : other.headers.entrySet()) {
            for (String value : otherEntry.getValue()) {
                if (!containsValue(otherEntry.getKey(), value)) {
                    return false;
                }
            }
        }

        return true;
    }

    /**
     * Set the authentication
     * 
     * @param authentication
     *            the authentication to set
     */
    public void setAuthentication(ScspAuthentication authentication) {
        this.authentication = authentication;
    }

    /**
     * Get the authentication
     * 
     * @return the authentication
     */
    public ScspAuthentication getAuthentication() {
        return authentication;
    }

    // copy all of the headers from the input map into me. This will merge any matching header names
    // into the existing list for a header.
    private void copyHeadersFromMap(HashMap<String, ArrayList<String>> map) {
        for (Map.Entry<String, ArrayList<String>> entry : map.entrySet()) {
            addEntryListIgnoreCase(entry.getKey(), entry.getValue());
        }
    }

    // merge a list of header values into a header or create a new header map entry if it is not already
    // there.
    private void addEntryListIgnoreCase(String name, ArrayList<String> values) {
        String headerName = getMapNameIgnoreCase(name);
        if (headerName != null) {
            ArrayList<String> currentValues = headers.get(headerName);
            currentValues.addAll(values);
            headers.put(headerName, currentValues);
        } else {
            headers.put(name, values);
        }
    }

    // merge a header value into a header or create a new header map entry if it is not already
    // there.
    private void addEntryIgnoreCase(String name, String value) {
        String headerName = getMapNameIgnoreCase(name);
        if (headerName != null) {
            ArrayList<String> currentValues = headers.get(headerName);
            currentValues.add(value);
            headers.put(headerName, currentValues);
        } else {
            ArrayList<String> newValues = new ArrayList<String>();
            newValues.add(value);
            headers.put(name, newValues);
        }
    }

    // Retrieve a list of header names that match the input name ignoring case.
    private String getMapNameIgnoreCase(String name) {
        for (Map.Entry<String, ArrayList<String>> header : headers.entrySet()) {
            if (header.getKey().compareToIgnoreCase(name) == 0) {
                return header.getKey();
            }
        }
        return null;
    }

}
