/**
 * 
 */
package com.caringo.client;

import java.text.ParseException;

/**
 * Encapsulation SCSP Integrity Seals.<br>
 * <p>
 * Typical usage would be to parse an instance from a Location header returned by a write command using
 * <code>ScspIntegritySeal.valueOf</code> and use the <code>ScspQueryArgs</code>object returned by <code>toQueryArgs</code>
 * to set a command object's query args.
 * <p>
 * Copyright (c) 2009 by Caringo, Inc. -- All rights reserved<br>
 * This is free software, distributed under the terms of the New BSD license.<br>
 * See the LICENSE.txt file included in this archive.<br>
 * <br>
 * 
 * @author pray
 * @created September 20, 2009
 * @id ${Id}
 */
public class ScspIntegritySeal {
    /**
     * Known hash types
     */
    public static final String MD5 = "md5";
    public static final String SHA1 = "sha1";
    public static final String SHA256 = "sha256";
    public static final String SHA384 = "sha384";
    public static final String SHA512 = "sha512";

    private String hashType;
    private String hash;

    /**
     * Create an integrity seal from a hash type and hash value.
     * 
     * @param hashType -
     *            query arg hash type. Typically one of the known hash types.
     * @param hash -
     *            hash value.
     */
    public ScspIntegritySeal(String hashType, String hash) {
        this.hashType = hashType;
        this.hash = hash;
    }

    /**
     * Set the hash type.
     * 
     * @param hashType
     *            the hashType to set
     */
    public void setHashType(String hashType) {
        this.hashType = hashType;
    }

    /**
     * Get the hashType
     * 
     * @return the hashType
     */
    public String getHashType() {
        return hashType;
    }

    /**
     * Set the hash value
     * 
     * @param hash
     *            the hash to set
     */
    public void setHash(String hash) {
        this.hash = hash;
    }

    /**
     * Get the hash value
     * 
     * @return the hash
     */
    public String getHash() {
        return hash;
    }

    /**
     * Build a query arg string of the form 'hashtype='hashvalue'&'hash='hash
     * 
     * @return string suitable for passing as a query arg.
     */
    public String toQueryArgString() {
        StringBuilder sb = new StringBuilder("hashtype=");
        return sb.append(hashType).append("&hash=").append(hash).toString();
    }

    /**
     * Retrieve a <code>ScspQueryArgs</code> object containing the hashtype and hash.
     * 
     * @return a <code>ScspQueryArgs</code> object containing the hashtype and hash.
     */
    public ScspQueryArgs toQueryArgs() {
        ScspQueryArgs args = new ScspQueryArgs();
        args.setValue("hash", hash);
        args.setValue("hashtype", hashType);
        return args;
    }

    /**
     * Get a <code>ScspIntegritySeal<code> instance from a location header.
     * 
     * @param header - The header to parse.
     * @return new ScspIntegritySeal if one can be parsed from the header; null if it can't
     * @throws ParseException - if the header isn't a Location header or if the args in the header are unknown hashes
     */
    public static ScspIntegritySeal fromLocationHeader(ScspHeader header) throws ParseException {
        if (!header.getName().equalsIgnoreCase("Location")) {
            // Let's double check.
            // 
            throw new ParseException("Header wasn't Location. Header was " + header.getName(), 0);
        }
        String locationValue = header.getValue();
        int argIndex = locationValue.indexOf("?");
        if (argIndex <= 0) {
            return null; // no args in Location header
        } else {
            String args = locationValue.substring(argIndex + 1);
            return valueOf(args);
        }
    }

    /**
     * Get a <code>ScspIntegritySeal<code> instance from a location header value.
     * 
     * @param headerString - The header value to parse (doesn't include the "Location:" header name).
     * @return new ScspIntegritySeal if one can be parsed from the header value; null if it can't
     * @throws ParseException - if the args in the header value are unknown hashes
     */
    public static ScspIntegritySeal valueOf(String headerString) throws ParseException {
        ScspQueryArgs args = ScspQueryArgs.valueOf(headerString);
        if ((args.containsName("hash") && !args.containsName("hashtype"))
                || (!args.containsName("hash") && args.containsName("hashtype"))) {
            throw new ParseException("Missing hash or hashtype argument", 0);
        }
        if (args.containsName("hashtype")) {
            return new ScspIntegritySeal(args.getValue("hashtype"), args.getValue("hash"));
        } else {
            return null;
        }
    }
}
