package com.caringo.client;

import java.util.Date;
import java.util.Arrays;
import java.util.List;
import java.util.ArrayList;
import java.util.regex.Pattern;
import java.util.regex.Matcher;
import java.text.ParseException;

import com.caringo.client.request.InvalidPolicyException;

/**
 * Represents a CAStor lifepoint.
 * 
 * Copyright (c) 2009 by Caringo, Inc. -- All rights reserved Caringo
 * This is free software, distributed under the terms of the New BSD license.<br>
 * See the LICENSE.txt file included in this archive.<br>
 * <br>
 * 
 * @author jmike
 * @created Sep 21, 2008
 * @id $Id: ScspLifepoint.java 52121 2012-08-28 17:31:56Z jafshar $
 */
public class ScspLifepoint implements Comparable<ScspLifepoint> {

    private Integer minReps = null;
    private ScspDeleteConstraint deleteConstraint = null;
    private ScspDate endDate = null;
    private List<String> customAttributes = null;

    /**
     * Creates a non-terminal policy (has an end time).
     * 
     * @param endDate -
     *            The policy's end date.
     * @param minreps -
     *            The minimum number of replications. This parameter may be <code>null</code>.
     * @param delete -
     *            A <code>DeleteConstraint</code>. This parameter may be <code>null</code>.
     * @param custom -
     *            A <code>List</code> of <code>String</code> containing any custom lifepoint attributes. This parameter may be <code>null</code>.
     * @throws InvalidPolicyException
     *             If the policy is invalid.
     */
    public ScspLifepoint(ScspDate endDate, Integer minreps, ScspDeleteConstraint delete, List<String> custom) {
        super();
        this.endDate = endDate;
        this.minReps = minreps;
        this.deleteConstraint = delete;
        this.customAttributes = custom;
    }

    /**
     * Creates a non-terminal policy (has an end time).
     * 
     * @param endDate -
     *            The policy's end date.
     * @param minreps -
     *            The minimum number of replications. This parameter may be <code>null</code>.
     * @param delete -
     *            A <code>DeleteConstraint</code>. This parameter may be <code>null</code>.
     * @throws InvalidPolicyException
     *             If the policy is invalid.
     */
    public ScspLifepoint(ScspDate endDate, Integer minreps, ScspDeleteConstraint delete) {
        super();
        this.endDate = endDate;
        this.minReps = minreps;
        this.deleteConstraint = delete;
    }

    /**
     * Creates a non-terminal policy (has an end time).
     * 
     * @param endDate -
     *            The policy's end data.
     * @param minreps -
     *            The minimum number of replications. This parameter may be <code>null</code>.
     */
    public ScspLifepoint(ScspDate endDate, Integer minreps) {
        super();
        this.endDate = endDate;
        this.minReps = minreps;
    }

    /**
     * Creates a non-terminal policy (has an end time).
     * 
     * @param endDate -
     *            The policy's end date.
     * @param delete -
     *            A <code>DeleteConstraint<.code>. This parameter may be <code>null</code>.
     */
    public ScspLifepoint(ScspDate endDate, ScspDeleteConstraint delete) {
        super();
        this.endDate = endDate;
        this.deleteConstraint = delete;
    }

    /**
     * Creates a non-terminal policy (has an end time).
     * 
     * @param endDate -
     *            The policy's end date.
     * @param custom -
     *            A <code>List</code> of <code>String</code> containing any custom lifepoint attributes. This parameter may be <code>null</code>.
     */
    public ScspLifepoint(ScspDate endDate, List<String> custom) {
        super();
        this.endDate = endDate;
        this.customAttributes = custom;
    }

    /**
     * Creates a terminal policy (never ends).
     * 
     * @param minreps -
     *            The minimum number of replications. This parameter may be <code>null</code>.
     * @param delete -
     *            A <code>DeleteConstraint<.code>. This parameter may be <code>null</code>.
     * @param custom -
     *            A <code>List</code> of <code>String</code> containing any custom lifepoint attributes. This parameter may be <code>null</code>.
     */
    public ScspLifepoint(Integer minreps, ScspDeleteConstraint delete, List<String> custom) {
        super();
        this.minReps = minreps;
        this.deleteConstraint = delete;
        this.customAttributes = custom;
    }

    /**
     * Creates a terminal policy (never ends).
     * 
     * @param minreps -
     *            The minimum number of replications. This parameter may be <code>null</code>.
     * @param delete -
     *            A <code>DeleteConstraint<.code>. This parameter may be <code>null</code>.
     */
    public ScspLifepoint(Integer minreps, ScspDeleteConstraint delete) {
        super();
        this.minReps = minreps;
        this.deleteConstraint = delete;
    }

    /**
     * Creates a terminal policy (never ends).
     * 
     * @param minreps -
     *            The minimum number of replications. This parameter may be <code>null</code>.
     */
    public ScspLifepoint(Integer minreps) {
        super();
        this.minReps = minreps;
    }

    /**
     * Creates a terminal policy (never ends).
     * 
     * @param delete -
     *            A <code>DeleteConstraint</code>. This parameter may be <code>null</code>.
     */
    public ScspLifepoint(ScspDeleteConstraint delete) {
        super();
        this.deleteConstraint = delete;
    }

    /**
     * Creates a terminal policy (never ends).
     * 
     * @param custom -
     *            A <code>List</code> of <code>String</code> containing any custom lifepoint attributes. This parameter may be <code>null</code>.
     */
    public ScspLifepoint(List<String> custom) {
        super();
        this.customAttributes = custom;
    }

    /**
     * Determines if the <code>Lifepoint</code> is terminal.
     * 
     * @return True if the end date is <code>null</code>, false if it has a value.
     */
    public boolean isTerminal() {
        return endDate == null;
    }

    /**
     * Returns the <code>DeleteConstraint</code>.
     * 
     * @return The <code>DeleteConstraint</code>.
     */
    public ScspDeleteConstraint getDeleteConstraint() {
        return deleteConstraint;
    }

    /**
     * Date getter. Note that if the underlying <code>PolicyDate</code> is relative, this will likely return a different date each
     * time it is called.
     * 
     * @return The end date as a <code>Date</code> object.
     */
    public Date getEndDate() {
        return endDate == null ? null : endDate.toDate();
    }

    /**
     * Returns the end date as a <code>PolicyDate</code> object.
     * 
     * @return - Then end date as a <code>PolicyDate</code> object.
     */
    public ScspDate getEndPolicyDate() {
        return this.endDate;
    }

    /**
     * Returns the minimum number of replications.
     * 
     * @return The minimum number of replications.
     */
    public Integer getMinReps() {
        return minReps;
    }

    /**
     * Returns the custom attributes.
     * 
     * @return The custom attributes specified for the lifepoint.
     */
    public List<String> getCustomAttributes() {
        return customAttributes;
    }

    /**
     * @param deleteConstraint -
     *            The <code>DeleteConstraint</code> to set.
     */
    public void setDeleteConstraint(ScspDeleteConstraint deleteConstraint) {
        this.deleteConstraint = deleteConstraint;
    }

    /**
     * @param endDate -
     *            The end date to set.
     */
    public void setEndDate(ScspDate endDate) {
        this.endDate = endDate;
    }

    /**
     * @param minReps -
     *            The minimum number of replications to set.
     */
    public void setMinReps(int minReps) {
        this.minReps = minReps;
    }

    /**
     * @param custom -
     *            A <code>List</code> of <code>String</code> containing any custom lifepoint attributes. This parameter may be <code>null</code>.
     */
    public void setCustomAttributes(List<String> custom) {
        this.customAttributes = custom;
    }

    /**
     * @param attribute -
     *            A custom lifepoint attribute. This parameter may NOT be <code>null</code>.
     * @throws InvalidPolicyException
     *            If the custom attribute is <code>null</code>.
     */
    public boolean addCustomAttribute(String attribute) throws InvalidPolicyException {
        if (null == attribute) {
            throw new InvalidPolicyException("Cannot add a null custom attribute");
        }
        
        if (null == this.customAttributes) {
            this.customAttributes = new ArrayList<String>();
        }
        
        return this.customAttributes.add(attribute);
    }

    /**
     * Gets the text representation of this lifepoint.
     * 
     * @return A <code>String</code>.
     */
    @Override
    public String toString() {
        StringBuffer buf = new StringBuffer();
        boolean attributeAdded = false;

        buf.append("[");
        
        if (endDate != null) {
            buf.append(endDate.toString());
        }
        
        buf.append("] ");

        if (minReps != null) {
            buf.append("reps=").append(minReps);
            attributeAdded = true;
        }

        if (deleteConstraint != null) {
            if (attributeAdded) {
                buf.append(", ");
            }
            
            buf.append(deleteConstraint.headerString);
            attributeAdded = true;
        }
        
        if (null != this.customAttributes) {
            for (String attribute : this.customAttributes) {
                if (attributeAdded) {
                    buf.append(", ");
                }
                
                buf.append(attribute);
                attributeAdded = true;
            }
        }
        
        return buf.toString();
    }

    // Build the Lifepoint from a string header value

    public static ScspLifepoint valueOf(String value) throws ParseException {
        Integer minReps = null;
        ScspDeleteConstraint deleteConstraint = null;
        ScspDate endDate = null;
        ArrayList<String> customAttrs = new ArrayList<String>();

        Pattern datePattern = Pattern.compile("\\[([^\\]])*\\]");
        Matcher dateMatcher = datePattern.matcher(value);
        if (dateMatcher.find()) {
            String dateString = dateMatcher.group(0).trim();
            dateString = dateString.substring(1, dateString.length() - 1); // strip the brackets
            if (dateString.length() > 0) {
                endDate = new ScspDate(dateString);
            }
            
            int offset = dateMatcher.end(0);
            String remainder = value.substring(dateMatcher.end(0));
            ArrayList<String> tokens = new ArrayList<String>(Arrays.asList(remainder.split(",")));
            
            for (String token : tokens) {
                if (token.contains(ScspDeleteConstraint.DELETABLE.headerString)) {
                    if (null != deleteConstraint) {
                        throw new ParseException(value, offset);
                    }
                    
                    deleteConstraint = ScspDeleteConstraint.DELETABLE;
                } else if (token.contains(ScspDeleteConstraint.NOT_DELETABLE.headerString)) {
                    if (null != deleteConstraint) {
                        throw new ParseException(value, offset);
                    }
                    
                    deleteConstraint = ScspDeleteConstraint.NOT_DELETABLE;
                } else if (token.contains(ScspDeleteConstraint.MUST_DELETE.headerString)) {
                    if (null != deleteConstraint) {
                        throw new ParseException(value, offset);
                    }
                    
                    deleteConstraint = ScspDeleteConstraint.MUST_DELETE;
                } else {
                    Pattern repsPattern = Pattern.compile("reps=(\\d+)");
                    Matcher repsMatcher = repsPattern.matcher(token);
                    
                    if (repsMatcher.find()) {
                        if (null != minReps) {
                            throw new ParseException(value, offset);
                        }
                    
                        minReps = Integer.valueOf(repsMatcher.group(1));
                    } else {
                        // This is a non-standard attribute
                        customAttrs.add(token.trim());
                    }
                }
                
                offset += token.length() + 1;
            }
            
            return new ScspLifepoint(endDate, minReps, deleteConstraint, customAttrs);
        } else {
            throw new ParseException(value, 0);
        }
    }

    /**
     * This method needs a description.
     * 
     * @return An <code>int</code>.
     */
    @Override
    public int hashCode() {
        final int PRIME = 31;
        int result = 1;
        result = PRIME * result + ((deleteConstraint == null) ? 0 : deleteConstraint.hashCode());
        result = PRIME * result + ((endDate == null) ? 0 : endDate.hashCode());
        result = PRIME * result + ((minReps == null) ? 0 : minReps.hashCode());
        result = PRIME * result + ((customAttributes == null) ? 0 : customAttributes.hashCode());
        return result;
    }

    /**
     * This method needs a description.
     * 
     * @return A <code>boolean</code>.
     */
    @Override
    public boolean equals(Object obj) {
        if (this == obj)
            return true;
        if (obj == null)
            return false;
        if (getClass() != obj.getClass())
            return false;
            
        final ScspLifepoint other = (ScspLifepoint) obj;
        
        if (deleteConstraint == null) {
            if (other.deleteConstraint != null)
                return false;
        } else if (!deleteConstraint.equals(other.deleteConstraint))
            return false;
            
        if (endDate == null) {
            if (other.endDate != null)
                return false;
        } else if (!endDate.equals(other.endDate))
            return false;
            
        if (minReps == null) {
            if (other.minReps != null)
                return false;
        } else if (!minReps.equals(other.minReps))
            return false;
            
        if (customAttributes == null) {
            if (other.customAttributes != null)
                return false;
        } else if (!customAttributes.equals(other.customAttributes))
            return false;
            
        return true;
    }

    /**
     * Compares two <code>LifePoint</code> objects.
     * 
     * @return -1, 0 or 1. A better explanation is needed here by the author.
     * 
     */
    public int compareTo(ScspLifepoint other) {
        int result = 0;
        if (other == null) {
            result = -1;
        } else if (other != this) {

            // date
            if (this.endDate == null) {
                result = other.endDate == null ? 0 : 1;
            } else {
                result = this.endDate.compareTo(other.endDate);
            }

            // reps
            if (result == 0) {
                if (this.minReps != other.minReps) {
                    if (this.minReps == null) {
                        // other.minReps isn't null
                        result = -1;
                    } else if (other.minReps == null) {
                        // this.minReps isn't null
                        result = 1;
                    } else {
                        result = this.minReps < other.minReps ? -1 : 1;
                    }
                }
            }

            // delete constraint
            if (result == 0) {
                if (deleteConstraint == null) {
                    result = other.deleteConstraint == null ? 0 : 1;
                } else {
                    result = deleteConstraint.compareTo(other.deleteConstraint);
                }
            }

            // delete constraint
            if (result == 0) {
                if (customAttributes == null) {
                    result = other.customAttributes == null ? 0 : 1;
                } else if (customAttributes.size() != other.customAttributes.size()) {
                    result = customAttributes.size() < other.customAttributes.size() ? -1 : 1;
                } else {
                    for (int i = 0; i < customAttributes.size(); ++i) {
                        result = customAttributes.get(i).compareTo(other.customAttributes.get(i));
                        
                        if (0 != result) {
                            break;
                        }
                    }
                }
            }
        }
        return result;
    }

}
