package com.caringo.client;

import java.text.ParseException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

/**
 * Represents a set of policies. Can render a set of Lifepoint headers.
 * 
 * Copyright (c) 2008 by Caringo, Inc. -- All rights reserved Caringo
 * This is free software, distributed under the terms of the New BSD license.<br>
 * See the LICENSE.txt file included in this archive.<br>
 * <br>
 * 
 * @author jmike
 * @created Sep 21, 2008
 * @id $Id: ScspPolicy.java 52121 2012-08-28 17:31:56Z jafshar $
 */
public class ScspPolicy extends ArrayList<ScspLifepoint> {

    private static final long serialVersionUID = -7744577415211451132L;

    //
    // Constructors
    //

    /**
     * Default constructor.
     */
    public ScspPolicy() {
        super();
    }

    //
    // Public interface
    //

    /**
     * Converts the policy's <code>LifePoint</code> objects to headers.
     * 
     * @return - A collection of <code>LifePoint</code> headers.
     */
    public ScspHeaders toHeaders() {
        ScspHeaders result = new ScspHeaders();
        for (ScspLifepoint lifepoint : this) {
            result.addValue("Lifepoint", lifepoint.toString());
        }
        return result;
    }

    /**
     * Converts the policy's <code>LifePoint</code> objects to a collection of strings.
     * 
     * @return - A collection of <code>LifePoint</code> strings.
     */
    public Collection<String> toHeaderStrings() {
        Collection<String> result = new ArrayList<String>(size());
        for (ScspLifepoint lifepoint : this) {
            result.add("Lifepoint: " + lifepoint.toString());
        }
        return result;
    }

    /**
     * Converts a policy's <code>LifePoint</code> objects to a concatenated string of headers.
     * 
     * @return - The policy's <code>LifePoint</code> objects as a string. Each <code>LifePoint</code> is terminated with a
     *         <code>newline</code>.
     */
    @Override
    public String toString() {
        StringBuilder out = new StringBuilder();
        for (Iterator<ScspLifepoint> iterator = this.iterator(); iterator.hasNext();) {
            ScspLifepoint lp = iterator.next();
            if (out.length() > 0) {
                out.append("\n");
            }
            out.append("Lifepoint: ").append(lp.toString());
        }

        return out.toString();
    }

    /**
     * Constructs a <code>ScspPolicy</code> object from a <code>ScspHeaders</code> object's lifepoints.
     * 
     * @param scspHeaders
     *              The header object to scan for lifepoints
     * @return - A <code>ScspPolicy</code> object made up of the lifepoint headers in the input headers.
     * @throws ParseException
     */
    public static ScspPolicy getPolicyFromHeaders(ScspHeaders scspHeaders) throws ParseException {
        ScspPolicy policy = new ScspPolicy();
        ArrayList<String> headers = scspHeaders.getHeaderValues("lifepoint");
        if (headers != null) {
            for (String value : headers) {
                ScspLifepoint lp = ScspLifepoint.valueOf(value);
                policy.add(lp);
            }
        }
        return policy;
    }

}
