/**
 * 
 */
package com.caringo.client;

import java.util.HashMap;
import java.util.Map;
import java.util.regex.Pattern;
import java.util.regex.Matcher;

/**
 * Encapsulation of a set of query args. Provides a way to store and retrieve query arg names with one value and methods for parsing
 * and building query arg strings.<br>
 * <p>
 * Copyright (c) 2009 by Caringo, Inc. -- All rights reserved<br>
 * This is free software, distributed under the terms of the New BSD license.<br>
 * See the LICENSE.txt file included in this archive.<br>
 * <br>
 * <br>
 * 
 * @author pray
 * @created September 20, 2009
 * @id ${Id}
 */
public final class ScspQueryArgs {
    private HashMap<String, String> args;

    /**
     * Default constructor. Constructs an object with an empty query arg map.
     */
    public ScspQueryArgs() {
        args = new HashMap<String, String>();
    }

    /**
     * Set the list of query args.
     * 
     * @param args
     *            the arg list to set
     */
    public void setArgs(HashMap<String, String> args) {
        this.args = args;
    }

    /**
     * Return the list of query args.
     * 
     * @return the argList
     */
    public HashMap<String, String> getArgList() {
        return args;
    }

    /**
     * Retrieve the value stored for the input name.
     * 
     * @param name -
     *            The query arg name.
     * @return The value stored for the input name.
     */
    public String getValue(String name) {
        String key = getMapNameIgnoreCase(name);
        if (key != null) {
            return args.get(key);
        }
        else {
            return null;
        }
    }

    /**
     * Return whether the input name is stored
     * 
     * @param name -
     *            The query arg name to search for
     * @return <code>true</code>if the query arg name can be found.
     */
    public boolean containsName(String name) {
        return getMapNameIgnoreCase(name) != null;
    }

    /**
     * Set the value for the input arg name. Note that this will overwrite a previously stored value.
     * 
     * @param name -
     *            The name to set.
     * @param value -
     *            The value to set.
     */
    public void setValue(String name, String value) {
        String key = getMapNameIgnoreCase(name);
        if (key == null) {
            key = name;
        }   
        args.put(key, value);
    }

    /**
     * Copy all of the newArg values into this.
     * 
     * @param newArgs -
     *            map of strings->strings containing args->arg values.
     */
    public void addAll(HashMap<String, String> newArgs) {
        for (Map.Entry<String, String> arg : newArgs.entrySet()) {      
            setValue(arg.getKey(), arg.getValue());
        }
    }

    /**
     * Remove the query arg matching the input name.
     * 
     * @param name -
     *            Name of the query arg to remove.
     */
    public void remove(String name) {
        String key = getMapNameIgnoreCase(name);
        if (key != null) {
            args.remove(key);
        }
    }

    /**
     * Build a query arg string from the stored values.
     * 
     * @return a string of query args
     */
    public String toQueryArgString() {
        StringBuilder sb = new StringBuilder();
        for (Map.Entry<String, String> mapEntry : args.entrySet()) {
            if (sb.length() > 0) {
                sb.append("&");
            }
            sb.append(mapEntry.getKey()).append("=").append(mapEntry.getValue());
        }

        return sb.toString();
    }

    /**
     * Simple parse of an list of query arguments. Looks for string=string pairs separated by '&' and puts them in a new
     * ScspQueryArg. Note that this is not a complete query arg string validator.
     * 
     * @param argString -
     *            the string to parse.
     * @return A new <code>ScspQueryArgs</code> instance containing the args and values from the input string.
     */
    public static ScspQueryArgs valueOf(String argString) {
        ScspQueryArgs result = new ScspQueryArgs();
        final Pattern argPattern = Pattern.compile("([^=&]+)=([^&]*)(&)?");
        Matcher argMatcher = argPattern.matcher(argString);
        while (argMatcher.find()) {
            result.setValue(argMatcher.group(1), argMatcher.group(2));
        }
        return result;
    }

    // Retrieve a list of header names that match the input name ignoring case.
    private String getMapNameIgnoreCase(String name) {
        for (Map.Entry<String, String> arg : args.entrySet()) {
            if (arg.getKey().compareToIgnoreCase(name) == 0) {
                return arg.getKey();
            }
        }
        return null;
    }
}
