package com.caringo.client;

import java.io.OutputStream;

import com.caringo.client.request.ScspRequestHandler;

/**
 * Represents an SCSP READ command.<br>
 * <p>
 * Copyright (c) 2009 by Caringo, Inc. -- All rights reserved<br>
 * This is free software, distributed under the terms of the New BSD license.<br>
 * See the LICENSE.txt file included in this archive.<br>
 * <br>
 * 
 * @author pray
 * @created September 20, 2009
 * @id $Id: ScspRead.java 52121 2012-08-28 17:31:56Z jafshar $
 */
public class ScspRead extends ScspBaseStreamCommand {

    private OutputStream outputStream;

    // Package-private
    //
    // Construct a ScspRead object with a request handler, UUID, and output stream.
    // 
    ScspRead(ScspRequestHandler client, String UUID, OutputStream outputStream) {
        super(client, UUID);
        this.outputStream = outputStream;
    }

    /**
     * Return the output stream. This will likely be in a very different state after execution.
     * 
     * @return the output stream.
     */
    public OutputStream getOutputStream() {
        return outputStream;
    }

    /*
     * (non-Javadoc)
     * 
     * @see com.caringo.client.ScspCommand#execute()
     */
    @Override
    public ScspResponse execute() throws ScspExecutionException {
        BasicScspResponseOutputStream output = new BasicScspResponseOutputStream(getOutputStream());
        getPreparedClient().read(makePath(), getHeaders(), getQueryArgs().toQueryArgString(), output);
        int httpResponseStatusCode = output.getStatusCode();
        ScspResponse.ScspResultCode resultCode = ScspResponse.ScspResultCode.ScspRCSuccess;
        // Check for both a OK & partial read. Assume that we're getting a partial only if we ask for it.
        if (httpResponseStatusCode != 200 && httpResponseStatusCode != 206 && httpResponseStatusCode != 304) {
            resultCode = ScspResponse.ScspResultCode.ScspRCFailure;
        }
        return new ScspResponse(resultCode, httpResponseStatusCode, output.getStatusLine(), output.getHeaders(), output
                .getBodyString(), output.getRetryCount());
    }

    /*
     * (non-Javadoc)
     * 
     * @see com.caringo.client.ScspCommand#validate()
     */
    @Override
    public boolean validate() {
        boolean result = super.validate() && ScspValidations.validateAlias(getQueryArgs(), false)
                && ScspValidations.validateHashType(getQueryArgs(), false)
                && ScspValidations.validateNewHashType(getQueryArgs(), false) && ScspValidations.validateValidate(getQueryArgs())
                && (!getQueryArgs().containsName("indirect") || getQueryArgs().getValue("indirect").compareTo("HEAD") != 0);

        if (result) {
            // need to check that if there's a hash, there's a hashtype and a newhashtype
            boolean hasHash = getQueryArgs().containsName("hash");
            boolean hasHashType = getQueryArgs().containsName("hashtype");
            boolean hasNewHashType = getQueryArgs().containsName("newhashtype");
            // hasHashType <-> hasHash
            // hasNewHashType -> hasHash && hasHashType
            result = ((hasHash && hasHashType) ||
                      (!hasHash && !hasHashType)) && 
                     ((hasNewHashType && hasHash) || // shortcut here since hasHash <-> hasHashType from prior clause
                    !hasNewHashType);
        }
        return result;
    }
}
