/**
 * 
 */
package com.caringo.client;

/**
 * Response values returned by command and client execution methods. Contains a result code and the raw command response
 * headers, status code, and status line along with the response body when applicable.
 * <p>
 * Copyright (c) 2009 by Caringo, Inc. -- All rights reserved<br>
 * This is free software, distributed under the terms of the New BSD license.<br>
 * See the LICENSE.txt file included in this archive.<br>
 * <br>
 * 
 * @author pray
 * @created September 20, 2009
 * @id ${Id}
 */
public class ScspResponse {

    /**
     * Result status codes.
     * 
     */
    public enum ScspResultCode {
        ScspRCSuccess, ScspRCFailure
    }

    private int httpStatusCode;
    private String statusLine;
    private ScspHeaders responseHeaders;
    private String responseBody;
    private ScspResultCode resultCode;
    private int retries;

    /**
     * 
     * Construct a response from a result code, status code, status line, response heders, and response body.
     * Note that the response headers aren't passed in or stored as raw text due to the underlying HttpClient implementation.
     * 
     * @param resultCode - <code>ScspResultCode</code> indicating success or failure
     * @param httpStatusCode - Raw HTTP result code
     * @param statusLine - Raw HTTP status line
     * @param responseHeaders - list of HTTP response headers. 
     * @param responseBody - Raw HTTP response body.
     */
    public ScspResponse(ScspResultCode resultCode, int httpStatusCode, String statusLine, ScspHeaders responseHeaders, String responseBody, int retries ) {
        this.resultCode = resultCode;
        this.httpStatusCode = httpStatusCode;
        this.statusLine = statusLine;
        this.responseHeaders = responseHeaders;
        this.responseBody = responseBody;
        this.retries = retries;
    }
    /**
     * Set the raw HTTP status code.
     * 
     * @param httpStatusCode - The httpStatusCode to set
     */
    public void setHttpStatusCode(int httpStatusCode) {
        this.httpStatusCode = httpStatusCode;
    }
    
    /** 
     * Get the raw HTTP status code.
     * 
     * @return the httpStatusCode
     */
    public int getHttpStatusCode() {
        return httpStatusCode;
    }
    
    /** 
     * Set the HTTP response headers.
     * 
     * @param responseHeaders the responseHeaders to set
     */
    public void setResponseHeaders(ScspHeaders responseHeaders) {
        this.responseHeaders = responseHeaders;
    }
    
    /**
     * Get the HTTP response headers.
     * 
     * @return the responseHeaders
     */
    public ScspHeaders getResponseHeaders() {
        return responseHeaders;
    }
    
    /**
     * Set the raw HTTP response body.
     * 
     * @param responseBody the responseBody to set
     */
    public void setResponseBody(String responseBody) {
        this.responseBody = responseBody;
    }
    /**
     * Get the raw HTTP response body.
     * 
     * @return the responseBody
     */
    public String getResponseBody() {
        return responseBody;
    }
    
    /**
     * Set the raw HTTP status line.
     * 
     * @param statusLine the statusLine to set
     */
    public void setStatusLine(String statusLine) {
        this.statusLine = statusLine;
    }
    
    /**
     * Get the raw HTTP status line.
     * @return the statusLine
     */
    public String getStatusLine() {
        return statusLine;
    }
    
    /**
     * Set the SCSP result code.
     * 
     * @param resultCode the resultCode to set
     */
    public void setResultCode(ScspResultCode resultCode) {
        this.resultCode = resultCode;
    }
    
    /**
     * Get the SCSP result code.
     * 
     * @return the resultCode
     */
    public ScspResultCode getResultCode() {
        return resultCode;
    }
    
    /**
     * Set the number of times the request was retried.
     * 
     * @param retries the retry count value to set
     */
    public void setResultCode(int retries) {
        this.retries = retries;
    }
    
    /**
     * Get the retry count
     * 
     * @return the retry count
     */
    public int getRetries() {
        return retries;
    }

    // for testing. Cats the status line, headers, and body
    public String toString() {
        String eol = System.getProperty("line.separator");
        StringBuilder sb = new StringBuilder(statusLine);
        sb.append(eol);
        for (ScspHeader header:responseHeaders.toHeaderList()) {
            sb.append(header.getName()).append(": ").append(header.getValue()).append(eol);
        }
        sb.append(responseBody).append(eol);
        sb.append("Retries: ").append(getRetries()).append(eol);
        sb.append("ResultCode: ").append(getResultCode()).append(eol);
        return sb.toString();
    }

}
