/*
 * Copyright (c) 2009 by Caringo, Inc. -- All rights reserved
 */
package com.caringo.client;

import java.io.IOException;
import java.io.OutputStream;

import org.apache.commons.httpclient.Header;

import com.caringo.client.request.ScspRequestHandler;

/**
 * ScspResponseOutputStream is the callback interface to the ScspRequester SCSP client. The normal lifecycle of an
 * ScspResponseHandler is that the {@link #responseReceived} method will be called first. Then, if there are body data to be
 * delivered, the {@link #bodyDataReceived} method will be called, possibly repeatedly, until all data are delivered (the body
 * length is given to you in the responseReceived() call, so you may do validation of the quanity of data delivered). Next, 
 * the {@link #footersReceived} method is called to set the footers/trailer headers from the response stream. Finally, the
 * {@link #responseComplete} method is called to indicate completion. <br>
 * <br>
 * Copyright (c) 2009 by Caringo, Inc. -- All rights reserved<br>
 * This is free software, distributed under the terms of the New BSD license.<br>
 * See the LICENSE.txt file included in this archive.<br>
 * <br>
 * 
 * @see ScspRequestHandler
 * @see BasicScspResponseOutputStream
 * 
 * @author jmike
 * @created Jan 19, 2008
 * @id $Id: ScspResponseOutputStream.java 52121 2012-08-28 17:31:56Z jafshar $
 */
public abstract class ScspResponseOutputStream extends OutputStream {

    /**
     * Initial response callback
     * 
     * @param statusCode
     *            the code of the response
     * @param headers
     *            the headers from the response
     * @param statusLine
     *            the status line from the response
     * @param responseBodyLength
     *            the length of the response body. <em>NB:</em> For HEAD requests, this <em>may not</em> be equal to the
     *            Content-length header.
     * @param retryCount
     *            the number of times the request was retried.
     * 
     * @return true if you want the content, false if not.
     */
    public abstract boolean responseReceived(int statusCode, Header[] headers, String statusLine, long responseBodyLength);

    /**
     * Receive a chunk of body data. This may be called as many times as necessary to deliver all of the data
     * 
     * @param data
     *          buffer of received data
     * @param offset
     *          location in output stream to write the data to
     * @param len
     *          length of received data buffer
     * @throws IOException
     */
    public abstract void bodyDataReceived(byte[] data, int offset, int len) throws IOException;

    /**
     * Final response callback. This adds the footers/trailer headers from the end of the response stream to me.
     * Note on terminology: 'trailer headers' and 'footers' seem to be synonymous for HttpClient. CAStor uses
     * 'trailer header'. We'll use both.
     * 
     * @param footers
     *            the footers/trailer headers from the response
     */
    public abstract void footersReceived(Header[] footers);

    /**
     * The response is complete, the body has been delivered, you're done.
     * 
     * @param retryCount     *            the number of times the request was retried.
     * @throws ScspExecutionException
     */
    public abstract void responseComplete(int retryCount) throws ScspExecutionException;

    public void write(int c) throws IOException {
        write(new byte[] { (byte) c });
    }

}
