/**
 * 
 */
package com.caringo.client;

import java.util.regex.Pattern;
import java.util.regex.Matcher;

/**
 * Utilty methods for performing header and query arg validations.<br>
 * <p>
 * Copyright (c) 2009 by Caringo, Inc. -- All rights reserved<br>
 * This is free software, distributed under the terms of the New BSD license.<br>
 * See the LICENSE.txt file included in this archive.<br>
 * <br>
 * 
 * @author pray
 * @created September 20, 2009
 * @id ${Id}
 */
public final class ScspValidations {

    private static final String tokenRegexp = "[^\\s\\(\\)<>\\@,;:\\\\/\\\"/\\[\\]\\?\\=\\{\\}]+";
    private static final String quotedStringRegexp = "\"[^\"\\\\]*(\\\\.[^\"\\\\]*)*\"";
    private static final Pattern argPattern = Pattern.compile("^;\\s*" + tokenRegexp + "=" + "(" + tokenRegexp + "|"
            + quotedStringRegexp + ")");

    // Examples of the compiled RFC patterns
    // private static final Pattern tokenPattern = Pattern.compile("^" + tokenRegexp);
    // private static final Pattern quotedStringPattern = Pattern.compile("^" + quotedStringRegexp);

    /**
     * Validate that an alias arg stored in args has the correct value.
     * 
     * @param args -
     *            <code>ScspQueryArgs</code> to check.
     * @param must -
     *            <code>true</code> if the args must contain an alias arg.
     * @return <code>!must</code> if the args don't contain an alias value; <code>true</code> if the args contain a valid alias
     *         value; and <code>false</code>otherwise.
     */
    public static boolean validateAlias(ScspQueryArgs args, boolean must) {
        return validateQueryArg(args, must, "alias", new String[] { "yes", "no", "true", "false" });
    }

    /**
     * Validate that a validate arg stored in args has the correct value.
     * 
     * @param args -
     *            <code>ScspQueryArgs</code> to check.
     * @return <code>true</code> if the args don't contain a validate name or if the validate value is valid.
     */
    public static boolean validateValidate(ScspQueryArgs args) {
        return validateQueryArg(args, false, "validate", new String[] { "yes", "no", "true", "false" });
    }

    /**
     * Validate that a countReps arg stored in args has the correct value.
     * 
     * @param args -
     *            <code>ScspQueryArgs</code> to check.
     * @return <code>true</code> if the args don't contain a countReps name or if the countReps value is valid.
     */
    public static boolean validateCountReps(ScspQueryArgs args) {
        return validateQueryArg(args, false, "countReps", new String[] { "yes", "no", "true", "false" });
    }

    /**
     * Validate that a replicate arg stored in args has the correct value.
     * 
     * @param args -
     *            <code>ScspQueryArgs</code> to check.
     * @return <code>true</code> if the args don't contain a replicate name or if the replicate value is valid.
     */
    public static boolean validateReplicateOnWrite(ScspQueryArgs args) {
        return validateQueryArg(args, false, "replicate", new String[] { "immediate" });
    }

    /**
     * Validate that a hashtype arg stored in args has the correct value.
     * 
     * @param args -
     *            <code>ScspQueryArgs</code> to check.
     * @return <code>true</code> if the args don't contain a hashtype name or if the hashtype value is valid.
     */
    public static boolean validateHashType(ScspQueryArgs args, boolean must) {
        return validateHashType(args, must, "hashtype");
    }

    /**
     * Validate that the input value is a hex string of the correct length.
     * 
     * @param value -
     *            the value to check.
     * @param length -
     *            the length the input string must be.
     * @return <code>true</code> if value is a hex string with the correct length.
     */
    public static boolean validateHexString(String value, int length) {
        boolean result = value.length() == length;
        if (result) {
            result = !Pattern.matches(".*[^a-fA-F\\d].*", value);
        }
        return result;
    }

    /**
     * Validate that a newhashtype arg stored in args has the correct value.
     * 
     * @param args -
     *            <code>ScspQueryArgs</code> to check.
     * @return <code>true</code> if the args don't contain a newhashtype name or if the newhashtype value is valid.
     */
    public static boolean validateNewHashType(ScspQueryArgs args, boolean must) {
        return validateHashType(args, must, "newhashtype");
    }

    /**
     * Validate all of the Content-Type headers in the input headers.
     * 
     * See RFC2616 Sections 14.17 and 3.7 for the correct syntax for a Content-Type header.
     * 
     * @param headers -
     *            <code>ScspHeaders</code> to check.
     * @return <code>true</code> if all of the Content-Type headers are valid.
     */
    public static boolean validateContentType(ScspHeaders headers) {
        boolean result = true;
        if (headers.containsName("Content-Type")) {
            // rfc 2616 secs 14.17 & 3.7
            // "Content-Type" ":" media-type
            // media-type = type "/" subtype *( ";" parameter )
            // type = token
            // subtype = token
            for (String value : headers.getHeaderValues("Content-Type")) {
                // Validate all of them.
                final Pattern mimeTypePattern = Pattern.compile("^" + tokenRegexp + "/" + tokenRegexp);
                String[] matched = eatRegexp(value.trim(), mimeTypePattern);
                if (matched != null) {
                    value = matched[1].trim();
                } else {
                    result = false;
                    break;
                }
                while (value.length() > 0 && result) {
                    matched = eatRegexp(value, argPattern);
                    if (matched != null) {
                        value = matched[1].trim();
                    } else {
                        // there was some string left, and we didn't match anything.
                        result = false;
                    }
                }
                if (!result) {
                    break;
                }
            }

        }
        return result;
    }

    /**
     * Validate all of the Content-Type headers in the input headers.
     * 
     * See RFC2616 Section 19.5.1 for the correct syntax for a Content-Disposition header.
     * 
     * @param headers -
     *            <code>ScspHeaders</code> to check.
     * @return <code>true</code> if all of the Content-Disposition headers are valid.
     */
    static public boolean validateContentDisposition(ScspHeaders headers) {
        boolean result = true;
        if (headers.containsName("Content-Disposition")) {
            // rfc 2616 sec 19.5.1
            // content-disposition = "Content-Disposition" ":"
            // disposition-type *( ";" disposition-parm )
            // disposition-type = "attachment" | disp-extension-token
            // disposition-parm = filename-parm | disp-extension-parm
            // filename-parm = "filename" "=" quoted-string
            // disp-extension-token = token
            // disp-extension-parm = token "=" ( token | quoted-string )

            for (String value : headers.getHeaderValues("Content-Disposition")) {
                // Validate all of them.
                // "attachment" is just a token, so we can collapse the disposition-type to token
                final Pattern dispositionTypePattern = Pattern.compile("^" + tokenRegexp);
                String[] matched = eatRegexp(value, dispositionTypePattern);
                if (matched != null) {
                    value = matched[1].trim();
                } else {
                    result = false;
                    break;
                }
                if (matched != null) {
                    while (value.length() > 0) {
                        // filename-parm | disp-extension-parm
                        // is equivalent to "filename" "=" quoted-string | token "=" (token | quoted-string)
                        // note that we have to match filename=quoted-string separately from the token = alternative
                        // since filename is a token and we'll get false filename=token matches

                        // try filename=quoted-string
                        final Pattern fileNamePattern = Pattern.compile("^;\\s*filename=" + quotedStringRegexp);
                        matched = eatRegexp(value, fileNamePattern);
                        if (matched != null) {
                            value = matched[1].trim();
                        } else {
                            // no filename=quoted-string, try argument
                            matched = eatRegexp(value, argPattern);
                            if (matched != null) {
                                // but let's make sure the matched token wasn't filename
                                String token = matched[0];
                                if (token.length() > 0 && token.startsWith(";")) {
                                    token = token.substring(1).trim();
                                    if (token.startsWith("filename")) {
                                        result = false;
                                    }
                                } else {
                                    result = false;
                                }
                                value = matched[1].trim();
                            } else {
                                // there was some string left, and we didn't match anything.
                                result = false;
                                break;
                            }
                        }
                        if (!result) {
                            break;
                        }
                    }
                }
                if (!result) {
                    break;
                }
            }

        }
        return result;
    }
    /**
     * Validate the Allow headers. 
     *
     * If not mutable, the only allowable method in an Allow header is 'DELETE'.
     * If mutable, the allowable methods are 'PUT', 'COPY', 'APPEND', and 'DELETE'.
     *
     * @param headers -
     *      <code>ScspHeaders</code> to check
     * @param mutable - 
     *      <code>true</code> if we should check the values for a mutable stream
     * @return <code>false</code> if there are any methods in any Allow headers that aren't as described
     *   in the allowable methods and <code>true</code> otherwise.
     */
    private static final String[] MUTABLE_ALLOW_METHODS = {"PUT", "COPY", "APPEND", "DELETE"};
    private static final String[] IMMUTABLE_ALLOW_METHODS = {"DELETE"};
    private static boolean arrayContains(String val, String[] arr) {
        for (String av : arr) {
            if (val.compareToIgnoreCase(av) == 0) {
                return true;
            }
        }
        return false;
    }
    public static boolean validateAllow(ScspHeaders headers, boolean mutable) {

        String allowables[];
        if (mutable) {
            allowables = MUTABLE_ALLOW_METHODS;
        }
        else {
            allowables = IMMUTABLE_ALLOW_METHODS;
        }
        if (headers.containsName("Allow")) {
            for (String val : headers.getHeaderValues("Allow")) {
                for (String method : val.split(",")) {
                    if (! arrayContains(method.toUpperCase().trim(), allowables)) {
                        return false;
                    }
                }
            }
        }
        return true;
    }

    private static boolean validateHost(String host) {
        boolean result = Pattern.matches("[a-zA-Z0-9][\\w\\-\\.]*", host);
        return result;
    }

    /**
     * Validate Host headers. 
     *
     *
     * @param headers -
     *      <code>ScspHeaders</code> to check
     * @return <code>false</code> if there are 0 or 1 Host headers and the header value, if there is one, is a valid host value.
     */
     public static boolean validateHost(ScspHeaders headers) {
        return !headers.containsName("Host") || 
            (headers.getHeaderValues("Host").size() == 1 && validateHost(headers.getHeaderValues("Host").get(0)));
    }
    
    /**
     * Validate that an admin arg stored in args has the correct value.
     * 
     * @param args -
     *            <code>ScspQueryArgs</code> to check.
     * @return <code>true</code> if the args don't contain an admin name or if the admin value is valid.
     */
     public static boolean validateAdmin(ScspQueryArgs args) {
        return validateQueryArg(args, false, "admin", new String[] { "yes", "no", "true", "false" });
    }
    
    /**
     * Validate that a domain arg stored in args has the correct value.
     * 
     * @param args -
     *            <code>ScspQueryArgs</code> to check.
     * @return <code>true</code> if the args don't contain a domain name or if the domain value is valid.
     */
    public static boolean validateDomain(ScspQueryArgs args) {
        return !args.containsName("domain") || validateHost(args.getValue("domain"));
    }

    // verify that a hashtype or newhashtype value matches one of the known values.
    //
    private static boolean validateHashType(ScspQueryArgs args, boolean must, String arg) {
        return validateQueryArg(args, must, arg, new String[] { "md5", "sha1", "sha256", "sha384", "sha512" });
    }

    // verify that a query arg has one of the acceptable value. If must, then result is false if the
    // arg isn't found.
    private static boolean validateQueryArg(ScspQueryArgs args, boolean must, String arg, String[] acceptableValues) {
        boolean result = false;
        if (args.containsName(arg)) {
            String value = args.getValue(arg);
            for (String acceptable : acceptableValues) {
                if (value.compareToIgnoreCase(acceptable) == 0) {
                    result = true;
                    break;
                }
            }

        } else {
            result = !must;
        }
        return result;
    }

    // consume a regular expression pattern from an string. in the returned array,
    // result[0] is the matched string and result[1] is the remainder. The returned
    // value is null if there is no match.
    static String[] eatRegexp(String val, Pattern regexp) {
        Matcher matcher = regexp.matcher(val);
        if (matcher.find()) {
            String[] result = new String[2];
            result[0] = matcher.group(0);
            result[1] = val.substring(matcher.end(0));
            return result;
        } else
            return null;
    }
}
