package com.caringo.client;

import java.io.InputStream;

import com.caringo.client.request.ScspRequestHandler;

/**
 * Represents an SCSP Write command.<br>
 * <p>
 * Copyright (c) 2009 by Caringo, Inc. -- All rights reserved<br>
 * This is free software, distributed under the terms of the New BSD license.<br>
 * See the LICENSE.txt file included in this archive.<br>
 * <br>
 * 
 * @author pray
 * @created September 20, 2009
 * @id $Id: ScspWrite.java 52121 2012-08-28 17:31:56Z jafshar $
 */
public class ScspWrite extends ScspCommand {

    private InputStream inputStream;
    private long streamLength;

    // Package-private
    //
    // Construct a ScspWrite object with a request handler, input stream, and input data length.
    // 
    ScspWrite(ScspRequestHandler client, InputStream inputStream, long streamLength) {
        super(client);
        this.inputStream = inputStream;
        this.streamLength = streamLength;
    }

    /**
     * Return the source input stream. Note that execute may change the state of the input stream.
     * 
     * @return the input stream.
     */
    public InputStream getInputStream() {
        return inputStream;
    }

    /**
     * Return the length of the input stream.
     * 
     * @return The length of the input stream.
     */
    public long getStreamLength() {
        return streamLength;
    }

    /*
     * (non-Javadoc)
     * 
     * @see com.caringo.client.ScspCommand#execute()
     */
    @Override
    public ScspResponse execute() throws ScspExecutionException {
        BasicScspResponseOutputStream output = new BasicScspResponseOutputStream(null);
        String contentType = "";
        if (getHeaders().containsName("Content-type")) {
            contentType = getHeaders().getHeaderValues("Content-Type").get(0);
        }
        getPreparedClient().write(getPath(), getInputStream(), getStreamLength(), contentType, getHeaders(),
                getQueryArgs().toQueryArgString(), output);
        int httpResponseStatusCode = output.getStatusCode();
        ScspResponse.ScspResultCode resultCode = ScspResponse.ScspResultCode.ScspRCSuccess;
        // Do we need a partial content result check?
        if (httpResponseStatusCode != 201 && httpResponseStatusCode != 202) {
            resultCode = ScspResponse.ScspResultCode.ScspRCFailure;
        }
        return new ScspResponse(resultCode, httpResponseStatusCode, output.getStatusLine(), output.getHeaders(), output
                .getBodyString(), output.getRetryCount());
    }

    /*
     * (non-Javadoc)
     * 
     * @see com.caringo.client.ScspCommand#validate()
     */
    @Override
    public boolean validate() {
        boolean mutable = getQueryArgs().containsName("alias") && getQueryArgs().getValue("alias").compareTo("yes") == 0;
        boolean result = super.validate()
                && getHeaders().getNamesWithPrefix("Castor-System-").size() == 0
                && ScspValidations.validateAlias(getQueryArgs(), false) && ScspValidations.validateReplicateOnWrite(getQueryArgs())
                && ScspValidations.validateHashType(getQueryArgs(), false)
                && ScspValidations.validateContentDisposition(getHeaders()) && ScspValidations.validateContentType(getHeaders())
                && ScspValidations.validateAllow(getHeaders(), mutable);

        return result;
    }

}
