/**
 *
 */
package com.caringo.client.examples;

import com.caringo.client.*;

import java.io.*;
import java.util.ArrayList;
import java.util.Date;
import java.math.BigInteger;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;

/**
 * Copyright (c) 2009 by Caringo, Inc. -- All rights reserved<br>
 * This is free software, distributed under the terms of the New BSD license.<br>
 * See the LICENSE.txt file included in this archive.<br>
 * <br>
 *
 * @author pray
 * @created September 20, 2009
 * @id $Id: SCSPClientExample.java 52121 2012-08-28 17:31:56Z jafshar $
 */
public final class SCSPClientExample {

	/**
	 * @param args
	 */
	public static void main(String[] args) {
		String[] hosts = new String[1];
        //hosts[0] = "cas.caringo.com";
        hosts[0] = "10.136.1.1";
		ScspClient client = new ScspClient(hosts,80,4,4,120,300,300);
		try
		{
			client.start();
		}
		catch (Exception ex)
		{
			ex.printStackTrace(); // might as well continue, we're going down regardless
		}

		try
		{
			RunExamples(client);
		}

		finally
		{
			client.stop();
		}

	}

    static void RunMutableMethodExamples(ScspClient client)
    //[[DOC SNIPPET START | Basic Examples | SCSP operations on unnamed mutable objects (no authentication)]]
    //This example shows how to WRITE, READ, UPDATE, COPY, APPEND, INFO, and DELETE mutable objects.
    //Here, "mutable" means mutable unnamed objects, also referred to as "anchor streams".
    {
        System.out.println(">>>>>> Mutable Method Example <<<<<<<<");
        String testData = "Hello world";
        String testUpdateData = "I'm Here";
        String testAppendData = "I'm Gone";
        ByteArrayInputStream inputStream = new ByteArrayInputStream(testData.getBytes());
        ByteArrayInputStream updateInputStream = new ByteArrayInputStream(testUpdateData.getBytes());
        ByteArrayInputStream appendInputStream = new ByteArrayInputStream(testAppendData.getBytes());
        ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
        try
        {
            System.out.println(">>>>>> Write mutable <<<<<<<<");
            ScspQueryArgs args = new ScspQueryArgs();
            ScspResponse wcResponse = client.writeMutable("", inputStream, testData.getBytes().length, args, new ScspHeaders());
            String uuid = wcResponse.getResponseHeaders().getHeaderValues("Content-UUID").get(0);
            System.out.println(wcResponse.toString());

            System.out.println(">>>>>> Update mutable <<<<<<<<");
            ScspResponse ucResponse = client.updateMutable("", uuid, updateInputStream, testUpdateData.getBytes().length,
                                                           args, new ScspHeaders());
            System.out.println(ucResponse.toString());

            System.out.println(">>>>>> Read mutable <<<<<<<<");
            ScspResponse rcResponse = client.readMutable("", uuid, outputStream, args, new ScspHeaders());
            System.out.println(rcResponse.toString());
            System.out.println("Body Received");
            System.out.println(outputStream.toString());

            System.out.println(">>>>>> Copy mutable <<<<<<<<");
            ScspHeaders copyHeaders = new ScspHeaders();
            copyHeaders.addValue("Castor-color", "blue");
            ScspResponse ccResponse = client.copyMutable("", uuid, args, copyHeaders);
            System.out.println(ccResponse.toString());

            System.out.println(">>>>>> Append mutable <<<<<<<<");
            ScspResponse acResponse = client.appendMutable("", uuid, appendInputStream, testAppendData.getBytes().length,
                                                           args, new ScspHeaders());
            System.out.println(acResponse.toString());

            System.out.println(">>>>>> Read mutable after Append <<<<<<<<");
            outputStream.reset();
            rcResponse = client.readMutable("", uuid, outputStream, args, new ScspHeaders());
            System.out.println(rcResponse.toString());
            System.out.println("Body Received");
            System.out.println(outputStream.toString());

            System.out.println(">>>>>> Info mutable <<<<<<<<");
            ScspResponse icResponse = client.infoMutable("", uuid, args, new ScspHeaders());
            System.out.println(icResponse.toString());

            System.out.println(">>>>>> Delete mutable <<<<<<<<");
            ScspResponse dcResponse = client.deleteMutable("", uuid, args, new ScspHeaders());
            System.out.println(dcResponse.toString());
        }
        catch (Exception ex)
        {
            ex.printStackTrace();
        }
    }
    //[[DOC SNIPPET END]]

    static void RunImmutableMethodExamples(ScspClient client)
    //[[DOC SNIPPET START | Basic Examples | SCSP operations on unnamed immutable operations (no authentication)]]
    //This example shows how to WRITE, READ, INFO, and DELETE immutable objects.
    //Immutable objects are unnamed objects that can be modified in only one way, and that is to be deleted.
    {
        System.out.println(">>>>>> Immutable Method Examples <<<<<<<<");
        String testData = "Hello world";
        String testUpdateData = "I'm Here";
        ByteArrayInputStream inputStream = new ByteArrayInputStream(testData.getBytes());
        ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
        try
        {
            System.out.println(">>>>>> Write <<<<<<<<");
            ScspQueryArgs args = new ScspQueryArgs();
            ScspResponse wcResponse = client.write("", inputStream, testData.getBytes().length, args, new ScspHeaders());
            String uuid = wcResponse.getResponseHeaders().getHeaderValues("Content-UUID").get(0);
            System.out.println(wcResponse.toString());

            System.out.println(">>>>>> Read <<<<<<<<");
            ScspResponse rcResponse = client.read("", uuid, outputStream, args, new ScspHeaders());
            System.out.println(rcResponse.toString());
            System.out.println("Body Received");
            System.out.println(outputStream.toString());

            System.out.println(">>>>>> Info <<<<<<<<");
            ScspResponse icResponse = client.info("", uuid, args, new ScspHeaders());
            System.out.println(icResponse.toString());

            System.out.println(">>>>>> Delete <<<<<<<<");
            ScspResponse dcResponse = client.deleteMutable("", uuid, args, new ScspHeaders());
            System.out.println(dcResponse.toString());
        }
        catch (Exception ex)
        {
            ex.printStackTrace();
        }
    }
    //[[DOC SNIPPET END]]

    static void RunIntegritySealExamples(ScspClient client)
    //[[DOC SNIPPET START | MD5 Integrity Seal Example | Create an object with an MD5 integrity seal and upgrade the integrity seal]]
    //This example shows how to write an object with an MD5 integrity seal and then upgrade the
    //integrity seal to SHA256. An "integrity seal" is a URL containing an object's name or UUID,
    //its hash value, and the type of hash algorithm that was used for the computations. An application
    //can request an integrity seal when it performs a WRITE by including a hashtype query string.
    //By supplying the URL returned in the Location header from the WRITE request
    //(perhaps replacing the host address if connected to a different cluster or node), the application
    //can ask CAStor to validate while reading the data.
    {
        System.out.println(">>>>>> Integrity Seal Examples <<<<<<<<");

        StringBuilder testData = new StringBuilder("Hello world");
        ByteArrayInputStream inputStream = new ByteArrayInputStream(testData.toString().getBytes());
        ByteArrayOutputStream outputStream = new ByteArrayOutputStream();

        ScspQueryArgs hashTypeArgs = new ScspQueryArgs();
        hashTypeArgs.setValue("hashtype", ScspIntegritySeal.MD5);

        try
        {
            System.out.println(">>>>>> Write <<<<<<<<");
            ScspResponse wcResponse = client.write("", inputStream,testData.toString().getBytes().length, hashTypeArgs, new ScspHeaders());
            String uuid = wcResponse.getResponseHeaders().getHeaderValues("Content-UUID").get(0);
            ScspHeader header = new ScspHeader("Location", wcResponse.getResponseHeaders().getHeaderValues("Location").get(0));
            ScspIntegritySeal integritySeal = ScspIntegritySeal.fromLocationHeader(header);
            System.out.println(wcResponse.toString());
            System.out.println(integritySeal.toQueryArgString());

            System.out.println(">>>>>> Read <<<<<<<<");
            ScspResponse rcResponse = client.read("", uuid, outputStream, integritySeal.toQueryArgs(), new ScspHeaders());
            System.out.println(rcResponse.toString());
            System.out.println("Body Received");
            System.out.println(outputStream.toString());

            System.out.println(">>>>>> Read and Upgrade <<<<<<<<");
            ScspQueryArgs rehashArgs = integritySeal.toQueryArgs();
            rehashArgs.setValue("newhashtype", ScspIntegritySeal.SHA256);
            outputStream.reset();
            ScspResponse rcResponseRehash = client.read("", uuid, outputStream, rehashArgs, new ScspHeaders());
            header = new ScspHeader("Location", rcResponseRehash.getResponseHeaders().getHeaderValues("Location").get(0));
            integritySeal = ScspIntegritySeal.fromLocationHeader(header);
            System.out.println(rcResponseRehash.toString());
            System.out.println(integritySeal.toQueryArgs().toQueryArgString());

        }
        catch (Exception ex)
        {
            ex.printStackTrace();
        }
    }
    //[[DOC SNIPPET END]]

	static void RunLifepoint(ScspClient client, ScspLifepoint lifepoint, String headerMessage)
    //[[DOC SNIPPET START | Lifepoint Example | Creates objects and adds 'lifepoint' headers to constrain deletion]]
    //Creates objects with a 'lifepoint' header, gets the value of the header, and deletes the object.
	{

        System.out.println(">>>>> LP Example <<<<<< ");
        System.out.println(" "  + headerMessage);
		System.out.println("   LP: " + lifepoint.toString());

		String testData = "Hello world";
		ByteArrayInputStream inputStream = new ByteArrayInputStream(testData.getBytes());
		ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
		try
		{
			System.out.println(">>>>>> Write <<<<<<<<");
			ScspHeaders headers = new ScspHeaders();
            headers.addLifepoint(lifepoint.getEndPolicyDate(),lifepoint.getDeleteConstraint(), lifepoint.getMinReps());
			ScspResponse wcResponse = client.write("", inputStream, testData.toString().getBytes().length, new ScspQueryArgs(), headers);
            String uuid = wcResponse.getResponseHeaders().getHeaderValues("Content-UUID").get(0);
			System.out.println(wcResponse.toString());

			System.out.println(">>>>>> Read <<<<<<<<");
            ScspResponse rcResponse = client.read("", uuid, outputStream, new ScspQueryArgs(), new ScspHeaders());
			System.out.println(rcResponse.toString());
			System.out.println("Body Received");
			System.out.println(outputStream.toString());

			System.out.println(">>>>>> Info <<<<<<<<");
			ScspResponse icResponse = client.info("", uuid, new ScspQueryArgs(), new ScspHeaders());
			System.out.println(icResponse.toString());
            // retrieve and print the lifepoint
			if (icResponse.getResponseHeaders().getHeaderValues("Lifepoint") != null) {
				ScspLifepoint readLp = ScspLifepoint.valueOf(icResponse.getResponseHeaders().getHeaderValues("Lifepoint").get(0));
				System.out.println(readLp.toString());
			}
			else {
				System.out.println("Empty lifepoint"); // perfectly reasonable result if the constraint is delete
			}

			System.out.println(">>>>>> Delete <<<<<<<<");
			ScspResponse dcResponse = client.delete("", uuid, new ScspQueryArgs(), new ScspHeaders());
			System.out.println(dcResponse.toString());
		}
		catch (Exception ex)
		{
			ex.printStackTrace();
		}
	}
	static void RunLifepointExamples(ScspClient client)
	{
		System.out.println(">>>>>> Lifepoint Examples <<<<<<<<");
	    //public Lifepoint(Integer minreps)
		ScspLifepoint lpMinReps = new ScspLifepoint(2);
		RunLifepoint(client, lpMinReps, "Set the stream replicas to 2.");
	    //public Lifepoint(DeleteConstraint  delete)
		ScspLifepoint lpDeleteConstraint = new ScspLifepoint(ScspDeleteConstraint.MUST_DELETE);
		RunLifepoint(client, lpDeleteConstraint, "Set a terminal lifepoint to delete. Note that the stream will be deleted as soon as it's written.");
	    //public Lifepoint(Integer minreps, DeleteConstraint delete)
		ScspLifepoint lpMRD = new ScspLifepoint(3,ScspDeleteConstraint.DELETABLE);
		RunLifepoint(client, lpMRD, "Set the stream replicas to 2 and allow the stream to be deleted,");
	    //public Lifepoint(PolicyDate endDate, DeleteConstraint delete)
		ScspLifepoint lpDateDC = new ScspLifepoint(new ScspDate(24*60*60), ScspDeleteConstraint.NOT_DELETABLE);
		RunLifepoint(client, lpDateDC, "Don't allow the stream to be deleted until 1 day from now.");
	}
    //[[DOC SNIPPET END]]
    static void RunMutableExamples(ScspClient client) {
        RunMutableMethodExamples(client);
    }
    static void RunImmutableExamples(ScspClient client) {
        RunImmutableMethodExamples(client);
    }

    static void RunReadExamples(ScspClient client)
    {
    //[[DOC SNIPPET START | Other Read Examples | Reads objects using a 'Range' header, validated read, and with an 'Expect: Content-MD5' header]]
        System.out.println(">>>>>> Read Examples <<<<<<<<");

        StringBuilder testData = new StringBuilder("Hello world");
        ByteArrayInputStream inputStream = new ByteArrayInputStream(testData.toString().getBytes());
        ByteArrayOutputStream outputStream = new ByteArrayOutputStream();

        try
        {
            System.out.println(">>>>>> Write <<<<<<<<");
            ScspResponse wcResponse = client.write("", inputStream,testData.toString().getBytes().length, new ScspQueryArgs(), new ScspHeaders());
            String uuid = wcResponse.getResponseHeaders().getHeaderValues("Content-UUID").get(0);
            ScspHeader header = new ScspHeader("Location", wcResponse.getResponseHeaders().getHeaderValues("Location").get(0));
            System.out.println(wcResponse.toString());

            //Performs a validated read of the object. The content digest is computed and compared at the end.
            //A hash mismatch causes the socket connection to be dropped before the final bytes are sent.
			System.out.println(">>>>>> Read Validated <<<<<<<<");
            ScspQueryArgs validateArgs = new ScspQueryArgs();
            validateArgs.setValue("validate", "yes");
            ScspResponse rcResponse = client.read("", uuid, outputStream, validateArgs, new ScspHeaders());
            System.out.println(rcResponse.toString());
            System.out.println("Body Received");
            System.out.println(outputStream.toString());

            //Performs a range read of the object. When a READ request includes one or more Range headers,
            //a successful response includes only the requested bytes range(s) (in the example, bytes 1 through 5).
            //A range read, instead of returning a 200 OK response, returns a 206 Partial Content response,
            //indicating that only part of the stored stream is being returned.
            System.out.println(">>>>>> Read Range <<<<<<<<");
            outputStream.reset();
            ScspHeaders headers = new ScspHeaders();
            headers.addRange(1,5);
            rcResponse = client.read("", uuid, outputStream, new ScspQueryArgs(), headers);
            System.out.println(rcResponse.toString());
            System.out.println("Body Received");
            System.out.println(outputStream.toString());

            //Writes an object with an 'Expect: Content-MD5' header, which provides an end-to-end message
            //integrity check of the content (excluding metadata) as it is sent to and returned from CAStor.
            //CAStor computes the Content-MD5 as it receives the content. CAStor then writes the data and
            //returns the computed hash with the 201 (OK) response.
            System.out.println(">>>>>> Expect: Content-MD5 Read <<<<<<<<");
            outputStream.reset();
            ScspHeaders ecHeaders = new ScspHeaders();
            ecHeaders.addValue("Expect", "Content-MD5");
            outputStream.reset();
            ScspResponse rcResponseEC = client.read(uuid, "", outputStream, new ScspQueryArgs(), ecHeaders);
            System.out.println(rcResponseEC.toString());
            System.out.println("Body Received");
            System.out.println(outputStream.toString());
		}
		catch (Exception ex)
		{
			ex.printStackTrace();
        }
    }
    //[[DOC SNIPPET END]]

    static void RunInfoExamples(ScspClient client)
    {
        System.out.println(">>>>>> Info Examples <<<<<<<<");

        StringBuilder testData = new StringBuilder("Hello world");
        ByteArrayInputStream inputStream = new ByteArrayInputStream(testData.toString().getBytes());
        ByteArrayOutputStream outputStream = new ByteArrayOutputStream();

        try
        {
            System.out.println(">>>>>> Write <<<<<<<<");
            ScspResponse wcResponse = client.write("", inputStream, testData.toString().getBytes().length, new ScspQueryArgs(), new ScspHeaders());
            String uuid = wcResponse.getResponseHeaders().getHeaderValues("Content-UUID").get(0);
            ScspHeader header = new ScspHeader("Location", wcResponse.getResponseHeaders().getHeaderValues("Location").get(0));
            System.out.println(wcResponse.toString());

			System.out.println(">>>>>> Info Countreps <<<<<<<<");
            ScspQueryArgs countRepsArgs = new ScspQueryArgs();
            countRepsArgs.setValue("countreps", "yes");
            ScspResponse rcResponse = client.info("", uuid, countRepsArgs, new ScspHeaders());
            System.out.println(rcResponse.toString());
            System.out.println("Body Received");
            System.out.println(outputStream.toString());

		}
		catch (Exception ex)
		{
			ex.printStackTrace();
        }
    }

    static void RunStatusExamples(ScspClient client)
    {
        System.out.println(">>>>>> Node and Cluster Status Examples <<<<<<<<");
        try
        {
			System.out.println(">>>>>> Node Status <<<<<<<<");
            ScspResponse ncResponse = client.nodeStatus(new ScspQueryArgs(), new ScspHeaders());
            System.out.println(ncResponse.toString());

			System.out.println(">>>>>> Cluster Status <<<<<<<<");
            ncResponse = client.info("", "", new ScspQueryArgs(), new ScspHeaders());
            System.out.println(ncResponse.toString());
		}
		catch (Exception ex)
		{
			ex.printStackTrace();
        }
    }

    static void RunMutableEtagExample(ScspClient client, String headerName) {
        String testData = "Hello world";
        String testUpdateData = "I'm Here";
        ByteArrayInputStream inputStream = new ByteArrayInputStream(testData.getBytes());
        ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
        ByteArrayInputStream updateInputStream = new ByteArrayInputStream(testUpdateData.getBytes());
        System.out.println("    >>>>>> Write ETag: <<<<<<<<");
        System.out.println("    " + headerName);

        try {
            ScspQueryArgs args = new ScspQueryArgs();
            ScspResponse wcResponse = client.writeMutable("", inputStream, testData.getBytes().length, args, null);
            String uuid = wcResponse.getResponseHeaders().getHeaderValues("Content-UUID").get(0);
            ScspHeaders responseHeaders = wcResponse.getResponseHeaders();
            System.out.println(wcResponse.toString());

            if (responseHeaders.containsName("Etag")) {
                String etag = wcResponse.getResponseHeaders().getHeaderValues("Etag").get(0);
                ScspHeaders etagHeaders = new ScspHeaders();
                etagHeaders.addValue(headerName, etag);

                System.out.println(">>>>>> Update Etag <<<<<<<<");
                ScspResponse ucResponse = client.updateMutable("", uuid, updateInputStream, testUpdateData.getBytes().length, args, etagHeaders);
                System.out.println(ucResponse.toString());

                System.out.println(">>>>>> Read Etag <<<<<<<<");
                ScspResponse rcResponse = client.readMutable("", uuid, outputStream, args, etagHeaders);
                System.out.println(rcResponse.toString());
                System.out.println("Body Received");
                System.out.println(outputStream.toString());

                System.out.println(">>>>>> Info Etag <<<<<<<<");
                ScspResponse icResponse = client.infoMutable("", uuid, args, etagHeaders);
                System.out.println(icResponse.toString());

                System.out.println(">>>>>> Delete Etag <<<<<<<<");
                ScspResponse dcResponse = client.deleteMutable("", uuid, args, etagHeaders);
                System.out.println(dcResponse.toString());
            }
        }
        catch (Exception ex) {
            ex.printStackTrace();
        }
    }

    static void RunCacheCoherencyExamples(ScspClient client) {
        System.out.println(">>>>>> Cache Coherency Examples <<<<<<<<");

        try
        {
            RunMutableEtagExample(client, "If-Match");
            RunMutableEtagExample(client, "If-None-Match");
            RunMutableModifiedExample(client, "If-Modified-Since");
            RunMutableModifiedExample(client, "If-Unmodified-Since");
        }
        catch (Exception ex)
        {
            ex.printStackTrace();
        }

    }
    static void RunMutableModifiedExampleOnDate(ScspClient client, String uuid,  ScspQueryArgs args, String headerName, String date) {
        System.out.println("    >>>>>> Date = " + date);
        ByteArrayOutputStream outputStream = new ByteArrayOutputStream();

        try {
        	ScspHeaders modifiedHeaders = new ScspHeaders();
        	modifiedHeaders.addValue(headerName, date);

	        System.out.println(">>>>>> Read Date <<<<<<<<");
	        ScspResponse rcResponse = client.readMutable("", uuid, outputStream, args, modifiedHeaders);
	        System.out.println(rcResponse.toString());
	        System.out.println("Body Received");
	        System.out.println(outputStream.toString());

	        System.out.println(">>>>>> Info Date <<<<<<<<");
	        ScspResponse icResponse = client.infoMutable("", uuid, args, modifiedHeaders);
	        System.out.println(icResponse.toString());
    	}
    	catch (Exception ex) {
    		ex.printStackTrace();
    	}
    }

    static void RunMutableModifiedExample(ScspClient client, String headerName) {
        String testData = "Hello world";
        ByteArrayInputStream inputStream = new ByteArrayInputStream(testData.getBytes());
        System.out.println("    >>>>>> Write Modified: <<<<<<<<");
        System.out.println("    " + headerName);

        try {
            ScspWrite wc = client.createWriteCommand(inputStream, testData.getBytes().length);
            ScspQueryArgs args = new ScspQueryArgs();
            ScspResponse wcResponse = client.writeMutable("", inputStream, testData.getBytes().length, args, null);
            String uuid = wcResponse.getResponseHeaders().getHeaderValues("Content-UUID").get(0);
            ScspHeaders responseHeaders = wcResponse.getResponseHeaders();
            System.out.println(wcResponse.toString());

            if (responseHeaders.containsName("Last-Modified")) {
            	// Let's give ourselves a little bit of time to make sure the header dates are valid
            	Thread.sleep(20000);
            	ScspDate scspDate = new ScspDate(responseHeaders.getHeaderValues("Last-Modified").get(0));
            	Date date = scspDate.toDate();
            	Date backDate = new Date(date.getTime() - 10000);
            	Date foreDate = new Date(date.getTime() + 10000);
            	RunMutableModifiedExampleOnDate(client, uuid, args, headerName, scspDate.toString());
            	RunMutableModifiedExampleOnDate(client, uuid, args, headerName, new ScspDate(backDate).toString());
            	RunMutableModifiedExampleOnDate(client, uuid, args, headerName, new ScspDate(foreDate).toString());

            	System.out.println(">>>>>> Delete Date <<<<<<<<");
    	        ScspResponse dcResponse = client.deleteMutable("", uuid, args, null);
    	        System.out.println(dcResponse.toString());
            }
        }
        catch (Exception ex) {
            ex.printStackTrace();
        }
    }

    static String getHtdigest(String theDomain, String theUser, String thePassword) throws NoSuchAlgorithmException {
        //
        // return the same contents that htdigest would put into a file
        //
        // taken from AuthUtil
        String pwdString = theUser +":"+ theDomain +":"+ thePassword;
        MessageDigest mm = MessageDigest.getInstance("MD5");
        mm.update(pwdString.getBytes());
        BigInteger lHashInt = new BigInteger(1, mm.digest());
        if (theDomain != "") {
            theDomain = theDomain +":";
            }
        String fileContent = theUser+":"+ theDomain + String.format("%1$032X", lHashInt).toLowerCase();
        return fileContent;
     }
    static void RunAllUsersTenantExampleNoAuth(ScspClient client, String bucketName) {
        String testData = "Hello world";
        String testUpdateData = "I'm Here.";
        String testAppendData = "\nI'm Gone";
        ByteArrayInputStream inputStream = new ByteArrayInputStream(testData.getBytes());
        ByteArrayInputStream updateInputStream = new ByteArrayInputStream(testUpdateData.getBytes());
        ByteArrayInputStream appendInputStream = new ByteArrayInputStream(testAppendData.getBytes());
        ByteArrayOutputStream outputStream = new ByteArrayOutputStream();

        System.out.println("\r\n>>>>> All Users Tenant Examples (without authentication) <<<<<<<\r\n\r\n");

        try {
            //[[DOC SNIPPET START | Named Object Examples | Create buckets and named objects without authentication]]
            //This example shows how to create a bucket in a tenant (domain) that does not use authentication.
            //The domain, which the cluster administrator creates using the Admin Console, must be named
            //'allusers.realm' and its Castor-Authorization header is: 'Castor-Authorization: allusers.realm/_administrators, post='
            //Every domain's Castor-Authorization header is created by the Admin Console.
            String domainName = "allusers.realm";
            client.setHostHeaderValue(domainName);

            System.out.println(">>>>>> Add bucket '" + bucketName + "'<<<<<<");

            ScspResponse bwResponse = client.write(bucketName, inputStream, 0, new ScspQueryArgs(), new ScspHeaders());
            System.out.println(bwResponse);

            String namedObjectPath = bucketName + "/named/object";

            System.out.println(">>>>>> Add named object 'named/object' to 'allusers.realm/" + bucketName + "' <<<<<<");
            bwResponse = client.write(namedObjectPath, inputStream, testData.getBytes().length, new ScspQueryArgs(), new ScspHeaders());
            System.out.println(bwResponse);

            System.out.println(">>>>>> Info named object 'named/object' <<<<<<");
            ScspResponse response = client.info("", namedObjectPath, new ScspQueryArgs(), new ScspHeaders());
            System.out.println(response);

            System.out.println(">>>>>> Read named object 'named/object' <<<<<<");
            response = client.read("", namedObjectPath, outputStream, new ScspQueryArgs(), new ScspHeaders());
            System.out.println(response);
            System.out.println("Body Received");
            System.out.println(outputStream.toString());

            System.out.println(">>>>>> Update named object 'named/object' <<<<<<");
            response = client.updateMutable("", namedObjectPath, updateInputStream, testUpdateData.getBytes().length, new ScspQueryArgs(), new ScspHeaders());
            System.out.println(response);

            System.out.println(">>>>>> Append named object 'named/object' <<<<<<");
            response = client.appendMutable("", namedObjectPath, appendInputStream, testAppendData.getBytes().length, new ScspQueryArgs(), new ScspHeaders());
            System.out.println(response);

            System.out.println(">>>>>> Copy named object 'named/object' <<<<<<");
            //copy ignores the client.hostHeaderValue (Bug 8366), so use an explicit Host header
            ScspHeaders hostHeader = new ScspHeaders();
            hostHeader.addValue("Host", domainName);
            response = client.copyMutable("", namedObjectPath, new ScspQueryArgs(), hostHeader);
            System.out.println(response);

            System.out.println(">>>>>> Info named object 'named/object' <<<<<<");
            response = client.info("", namedObjectPath, new ScspQueryArgs(), new ScspHeaders());
            System.out.println(response);

            //Important! We want to be able to rerun this example, so we have to delete the named streams we created.

            System.out.println(">>>>>> Delete named object 'named/object' <<<<<<");
            response = client.delete("", namedObjectPath, new ScspQueryArgs(), new ScspHeaders());
            System.out.println(response);

            System.out.println(">>>>>> Delete bucket '" + bucketName + "' <<<<<<");
            response = client.delete("", bucketName, new ScspQueryArgs(), new ScspHeaders());
            System.out.println(response);

            client.setHostHeaderValue("");
        }
        catch (Exception ex) {
            ex.printStackTrace();
            client.setHostHeaderValue("");
        }
    }
    //[[DOC SNIPPET END]]

    static void RunAllUsersTenantExampleWithAuth(ScspClient client, String bucketName) {
        String testData = "Hello world";
        String testUpdateData = "I'm Here.";
        String testAppendData = "\nI'm Gone";
        ByteArrayInputStream inputStream = new ByteArrayInputStream(testData.getBytes());
        ByteArrayInputStream updateInputStream = new ByteArrayInputStream(testUpdateData.getBytes());
        ByteArrayInputStream appendInputStream = new ByteArrayInputStream(testAppendData.getBytes());
        ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
        try {
            //[[DOC SNIPPET START | Authentication Examples | Create buckets and named objects using authentication and performs other SCSP operations on the objects]]
            //This is an example of creating a bucket in a tenant (domain) that requires authentication.
            //The domain, which the cluster administrator creates using the Admin Console, must be named
            //'allusers.realm' and its Castor-Authorization header is: 'Castor-Authorization: allusers.realm/_administrators, post='
            //Every domain's Castor-Authorization header is created by the Admin Console.
            //In this example, the 'allusers.realm/_administrators' realm is used for authentication.
            System.out.println("\r\n>>>>> All Users Tenant Examples (with an authenticated bucket) <<<<<<<\r\n\r\n");

            String domainName = "allusers.realm";
            client.setHostHeaderValue(domainName);

            //default admin credentials in CAStor. Update if you've changed your CAStor configuration.
            ScspAuthentication adminAuth = new ScspAuthentication("admin", "ourpwdofchoicehere", "CAStor Administrator", "");

            //create a password entry for user=user and password=password
            //the auth realm we'll be using will be <domainName>/<_administrators>
            String adminBucket = domainName + "/" + bucketName;
            ScspAuthentication domainAuth = new ScspAuthentication("user", "password", adminBucket, "");

            //Need to add the the password stream to the realm
            String passwordLine = getHtdigest(domainAuth.getRealm(), domainAuth.getUserName(), domainAuth.getPassword());

            ScspQueryArgs domainArgs = new ScspQueryArgs();
            domainArgs.setValue("domain", domainName);

            System.out.println(">>>>>> Add bucket '" + bucketName + "' with one user<<<<<<");

            ByteArrayInputStream passwordStream = new ByteArrayInputStream(passwordLine.getBytes());
            ScspResponse bwResponse = client.write(bucketName, passwordStream, passwordLine.getBytes().length, new ScspQueryArgs(), new ScspHeaders());
            System.out.println(bwResponse);

            String protectedObjectPath = bucketName + "/protectedobject";
            String unprotectedObjectPath = bucketName + "/unprotectedobject";

            ScspAuthorization authz = new ScspAuthorization();
            authz.addAuthorization(null, domainAuth.getRealm()); //set up domainName as the authentication realm for all HTTP methods (can use ScspAuthorization.ALL_OP instead of null)
            ScspHeaders caHeaders = new ScspHeaders();
            caHeaders.addValue("Castor-Authorization", authz.getAuthSpec()); // using ScspAuthorization ensures proper formatting of the auth specification

            System.out.println(">>>>>> Add named object 'protectedobject' to 'allusers.realm/" + bucketName + "' <<<<<<");
            bwResponse = client.write(protectedObjectPath, inputStream, testData.getBytes().length, new ScspQueryArgs(), caHeaders);
            System.out.println(bwResponse);

            System.out.println(">>>>>> Info named object 'protectedobject' without credentials<<<<<<"   );
            ScspResponse response = client.info("", protectedObjectPath, new ScspQueryArgs(), new ScspHeaders());
            System.out.println(response);

            System.out.println(">>>>>> Info named object 'protectedobject' with credentials<<<<<<");
            ScspHeaders authHeaders = new ScspHeaders();
            authHeaders.setAuthentication(domainAuth);
            response = client.info("", protectedObjectPath, new ScspQueryArgs(), authHeaders);
            System.out.println(response);

            System.out.println(">>>>>> Read named object 'protectedobject' <<<<<<");
            authHeaders = new ScspHeaders();
            authHeaders.setAuthentication(domainAuth);
            response = client.read("", protectedObjectPath, outputStream, new ScspQueryArgs(), authHeaders);
            System.out.println(response);
            System.out.println(outputStream.toString());

            System.out.println(">>>>>> Update named object 'protectedobject' <<<<<<");
            caHeaders = new ScspHeaders();
            caHeaders.addValue("Castor-Authorization", domainAuth.getRealm()); //otherwise, we'll discard the security
            caHeaders.setAuthentication(domainAuth);
            response = client.updateMutable("", protectedObjectPath, updateInputStream, testUpdateData.getBytes().length, new ScspQueryArgs(), caHeaders);
            System.out.println(response);

            System.out.println(">>>>>> Append named object 'protectedobject' <<<<<<");
            authHeaders = new ScspHeaders();
            authHeaders.setAuthentication(domainAuth);
            response = client.appendMutable("", protectedObjectPath, appendInputStream, testAppendData.getBytes().length, new ScspQueryArgs(), authHeaders);
            System.out.println(response);

            System.out.println(">>>>>> Copy named object 'protectedobject' <<<<<<");
            authz = new ScspAuthorization();
            authz.addAuthorization(null, domainAuth.getRealm()); //set up domainName as the authentication realm for all HTTP methods (can use ScspAuthorization.ALL_OP instead of null)
            caHeaders = new ScspHeaders();
            caHeaders.addValue("Castor-Authorization", authz.getAuthSpec()); //otherwise, we'll discard the security
            caHeaders.setAuthentication(domainAuth);
            response = client.copyMutable("", protectedObjectPath, new ScspQueryArgs(), caHeaders);
            System.out.println(response);

            System.out.println(">>>>>> Info named object 'protectedobject' <<<<<<");
            authHeaders = new ScspHeaders();
            authHeaders.setAuthentication(domainAuth);
            response = client.info("", protectedObjectPath, new ScspQueryArgs(), caHeaders);
            System.out.println(response);

            System.out.println(">>>>>> Add named object 'unprotectedobject' to 'allusers.realm/" + bucketName + "' <<<<<<");
            bwResponse = client.write(unprotectedObjectPath, inputStream, testData.getBytes().length, new ScspQueryArgs(), new ScspHeaders());
            System.out.println(bwResponse);

            System.out.println(">>>>>> Info named object 'unprotectedobject' without credentials<<<<<<"   );
            response = client.info("", unprotectedObjectPath, new ScspQueryArgs(), new ScspHeaders());
            System.out.println(response);

            //Important! We want to be able to rerun this example, so we have to delete the named streams we created.

            System.out.println(">>>>>> Delete named object 'protectedobject' <<<<<<");
            authHeaders = new ScspHeaders();
            authHeaders.setAuthentication(domainAuth);
            response = client.delete("", protectedObjectPath, new ScspQueryArgs(), authHeaders);
            System.out.println(response);

            System.out.println(">>>>>> Delete named object 'unprotectedobject' <<<<<<");
            response = client.delete("", unprotectedObjectPath, new ScspQueryArgs(), new ScspHeaders());
            System.out.println(response);

            System.out.println(">>>>>> Delete bucket '" + bucketName + "' <<<<<<" );
            response = client.delete("", bucketName, new ScspQueryArgs(), authHeaders);
            System.out.println(response);
        }
        catch (Exception ex) {
            ex.printStackTrace();
            client.setHostHeaderValue("");
        }
    }
    //[[DOC SNIPPET END]]

    static boolean CheckAllUsersTenantExamples(ScspClient client, String bucketName) {

        try {
            //Assumes that there's a tenant called 'allusers.realm' that's been set up through the console.
            String domainName = "allusers.realm";
            client.setHostHeaderValue(domainName);

            //default admin credentials in CAStor. Update if you've changed your CAStor configuration.
            ScspAuthentication adminAuth = new ScspAuthentication("admin", "ourpwdofchoicehere", "CAStor administrator", "");

            //create a password entry for user=user and password=password
            //the auth realm we'll be using will be <domainName>/<_administrators>
            String adminBucket = domainName + "/_administrators";
            ScspAuthentication domainAuth = new ScspAuthentication("user", "password", adminBucket, "");

            //Let's check that allusers.realm is there and is set up correctly. A lot of this relies on the current CAStor scheme for
            //creating an 'All Users' tenant in the console.
            ScspQueryArgs realmCheckArgs = new ScspQueryArgs();
            realmCheckArgs.setValue("domain", domainName); // You have to use the domain= query arg to access the realm stream itself.
            realmCheckArgs.setValue("admin","yes"); // If set up correctly, this is only accessible using Castor Administrator credentials
            ScspHeaders realmCheckHeaders = new ScspHeaders();
            realmCheckHeaders.setAuthentication(adminAuth);
            ScspResponse rcResponse = client.info("", "", realmCheckArgs, realmCheckHeaders);

            System.out.println(">>>>>> allusers.realm <<<<<<\r\n");
            System.out.println(rcResponse);

            if (rcResponse.getHttpStatusCode() != 200) {
                System.out.println("Couldn't find realm 'allusers.realm'. Returning.");
                return false;
            }
            if (! rcResponse.getResponseHeaders().containsName("Castor-Authorization")) {
                System.out.println("Realm 'allusers.realm' doesn't appear to have been set up correctly: missing Castor-Authorization header. Returning.");
                return false;
            }
            else {
                ScspAuthorization authz = ScspAuthorization.fromHeaders(rcResponse.getResponseHeaders());
                if ((0 < authz.getErrors().size()) || (!authz.validate())) {  // validate() is stricter about errors than the headers parser
                    System.out.println("Realm 'allusers.realm' doesn't appear to have been set up correctly: Castor-Authorization header is improperly formatted (see below errors). Returning.");

                    for (String error : authz.getErrors()) {
                        System.out.println(error);
                    }

                    return false;
                }

                ArrayList<String> expectedAll = new ArrayList<String>();
                ArrayList<String> expectedPost = new ArrayList<String>();

                expectedAll.add("allusers.realm/_administrators");
                expectedPost.add(ScspAuthorization.ANY_REALM);

                if ((!expectedAll.equals(authz.getAuthorization(ScspAuthorization.ALL_OP))) ||
                    (!expectedPost.equals(authz.getAuthorization(ScspAuthorization.POST_OP)))) {
                    System.out.println("Realm 'allusers.realm' doesn't appear to have been set up correctly: Castor-Authorization header doesn't look right. Returning.");
                    return false;
                }
            }

            rcResponse = client.info("", "_administrators", realmCheckArgs, realmCheckHeaders); //check that the _administrators bucket was created

            System.out.println(">>>>>> allusers.realm/_administrators <<<<<<\r\n");
            System.out.println(rcResponse);

            if (rcResponse.getHttpStatusCode() != 200) {
                System.out.println("Couldn't find _administrators bucket in 'allusers.realm'. Returning");
                return false;
            }

            if (! rcResponse.getResponseHeaders().containsName("Castor-Authorization")) {
                System.out.println("Realm 'allusers.realm' doesn't appear to have been set up correctly: missing Castor-Authorization header in _administrators bucket. Returning.");
                return false;
            }

            //one last check. Let's make sure that our test bucket isn't already there.
            rcResponse = client.info("", bucketName, new ScspQueryArgs(), new ScspHeaders());
            if (rcResponse.getHttpStatusCode() == 200) {
                System.out.println("Test bucket 'allusers.realm/" + bucketName + "' already in CAStor. Here's the response from an info query to CAStor:\r\n");
                System.out.println(rcResponse);
                return false;
            }

            return true;
        }
        catch (Exception ex) {
            ex.printStackTrace();
            client.setHostHeaderValue("");
            return false;
        }
    }

    static boolean CheckOnlyTenantUsersTenantExamples(ScspClient client) {

        try {
            //This example shows how to verify that you have a domain (tenant) named 'localusers.realm'
            //that requires authentication by a user in the domain realm to POST. The Castor-Authorization
            //header on this domain is: 'Castor-Authorization: localusers.realm/_administrators, post=localusers.realm'
            //The domain must exist with this header to pass verification.
            String domainName = "localusers.realm";
            client.setHostHeaderValue(domainName);

            //default admin credentials in CAStor. Update if you've changed your CAStor configuration.
            ScspAuthentication adminAuth = new ScspAuthentication("admin", "ourpwdofchoicehere", "CAStor administrator", "");

            //create a password entry for user=user and password=password
            //the auth realm we'll be using will be <domainName>/<_administrators>
            String adminBucket = domainName + "/_administrators";
            ScspAuthentication domainAuth = new ScspAuthentication("user", "password", adminBucket, "");

            //Let's check that allusers.realm is there and is set up correctly. A lot of this relies on the current CAStor scheme for
            //creating an 'Local Users' tenant in the console.
            ScspQueryArgs realmCheckArgs = new ScspQueryArgs();
            realmCheckArgs.setValue("domain", domainName); // You have to use the domain= query arg to access the realm stream itself.
            realmCheckArgs.setValue("admin","yes"); // If set up correctly, this is only accessible using Castor Administrator credentials
            ScspHeaders realmCheckHeaders = new ScspHeaders();
            realmCheckHeaders.setAuthentication(adminAuth);
            ScspResponse rcResponse = client.info("", "", realmCheckArgs, realmCheckHeaders);

            System.out.println(">>>>>> localusers.realm <<<<<<\r\n");
            System.out.println(rcResponse);

            if (rcResponse.getHttpStatusCode() != 200) {
                System.out.println("Couldn't find realm 'localusers.realm'. Returning.");
                return false;
            }
            if (! rcResponse.getResponseHeaders().containsName("Castor-Authorization")) {
                System.out.println("Realm 'localusers.realm' doesn't appear to have been set up correctly: missing Castor-Authorization header. Returning.");
                return false;
            }
            else {
                ScspAuthorization authz = ScspAuthorization.fromHeaders(rcResponse.getResponseHeaders());
                if ((0 < authz.getErrors().size()) || (!authz.validate())) {  // validate() is stricter about errors than the headers parser
                    System.out.println("Realm 'localusers.realm' doesn't appear to have been set up correctly: Castor-Authorization header is improperly formatted (see below errors). Returning.");

                    for (String error : authz.getErrors()) {
                        System.out.println(error);
                    }

                    return false;
                }

                ArrayList<String> expectedAll = new ArrayList<String>();
                ArrayList<String> expectedPost = new ArrayList<String>();

                expectedAll.add("localusers.realm/_administrators");
                expectedPost.add("localusers.realm");

                if ((!expectedAll.equals(authz.getAuthorization(ScspAuthorization.ALL_OP))) ||
                    (!expectedPost.equals(authz.getAuthorization(ScspAuthorization.POST_OP)))) {
                    System.out.println("Realm 'localusers.realm' doesn't appear to have been set up correctly: Castor-Authorization header doesn't look right. Returning.");
                    return false;
                }
            }

            rcResponse = client.info("", "_administrators", realmCheckArgs, realmCheckHeaders); //check that the _administrators bucket was created

            System.out.println(">>>>>> localusers.realm/_administrators <<<<<<\r\n");
            System.out.println(rcResponse);

            if (rcResponse.getHttpStatusCode() != 200) {
                System.out.println("Couldn't find _administrators bucket in 'localusers.realm'. Returning");
                return false;
            }
            if (! rcResponse.getResponseHeaders().containsName("Castor-Authorization")) {
                System.out.println("Realm 'localusers.realm' doesn't appear to have been set up correctly: missing Castor-Authorization header in _administrators bucket. Returning.");
                return false;
            }

            //one last check. Let's make sure that our test buckets aren't already there.
            rcResponse = client.info("", "protectedbucket", new ScspQueryArgs(), new ScspHeaders());
            if (rcResponse.getHttpStatusCode() == 200) {
                System.out.println("Test bucket 'localusers.realm/protectedbucket' already in CAStor. Here's the response from an info query to CAStor:\r\n");
                System.out.println(rcResponse);
                return false;
            }
            rcResponse = client.info("", "unprotectedbucket", new ScspQueryArgs(), new ScspHeaders());
            if (rcResponse.getHttpStatusCode() == 200) {
                System.out.println("Test bucket 'localusers.realm/unprotectedbucket' already in CAStor. Here's the response from an info query to CAStor:\r\n");
                System.out.println(rcResponse);
                return false;
            }

            return true;
        }
        catch (Exception ex) {
            ex.printStackTrace();
            client.setHostHeaderValue("");
            return false;
        }
    }


    static void RunOnlyTenantUsersTenantExample(ScspClient client) {
        String emptyData = "";
        String testData = "Hello world";
        String testUpdateData = "I'm Here.";
        String testAppendData = "\nI'm Gone";
        ByteArrayInputStream emptyStream = new ByteArrayInputStream(emptyData.getBytes());
        ByteArrayInputStream inputStream = new ByteArrayInputStream(testData.getBytes());
        ByteArrayInputStream updateInputStream = new ByteArrayInputStream(testUpdateData.getBytes());
        ByteArrayInputStream appendInputStream = new ByteArrayInputStream(testAppendData.getBytes());
        ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
        //[[DOC SNIPPET START | Authentication Examples | Create a bucket without authentication and one with authentication, creates an object in each, and performs other SCSP operations on the objects]]
        //This example verifies you have a tenant (domain) named 'localusers.realm'
        //that requires authentication by a user in the domain realm to POST. It then creates two buckets:
        //one that does not require authentication and one that does.
        //Finally, it adds a named object to each bucket, performs SCSP operations on the object (such as
        //COPY and APPEND, then deletes the objects and buckets.
        //Note that the example uses administrative override for authentication because we can't guess
        //the domain manager credentials created for the domain by your cluster administrator.
        //"Administrative override" means using the '?admin' query argument and authenticating with credentials
        //from a user in the 'CAStor administrators' realm. Therefore, you must change the code where indicated
        //if you are not using the default user name of 'admin' with password 'ourpwdofchoicehere'.
        //The Castor-Authorization header for this domain is: 'Castor-Authorization: localusers.realm/_administrators, post=localusers.realm'
        //The cluster administrator creates this domain using the Admin Console.
        //Every domain's Castor-Authorization header is created by the Admin Console.
        System.out.println("\r\n>>>>> Only Tenant Users Tenant Examples <<<<<<<\r\n\r\n");
        //Assumes that there's a tenant called 'localusers.realm' that's been set up through the console.
        String domainName = "localusers.realm";
        String unprotectedBucketName = "unprotectedbucket";
        String protectedBucketName = "protectedbucket";
        client.setHostHeaderValue(domainName);

        try {

        	//default admin credentials in CAStor. Update if you've changed your CAStor configuration.
	        ScspAuthentication adminAuth = new ScspAuthentication("admin", "ourpwdofchoicehere", "CAStor administrator", "");
	        //Need to add the the password stream to the realm

	        //create a password entry for user=user and password=password
	        //the auth realm we'll be using will be <domainName>/<_administrators>
	        String adminBucket = domainName + "/_administrators";
	        ScspAuthentication domainAuth = new ScspAuthentication("user", "password", adminBucket, "");

	        ScspQueryArgs domainArgs = new ScspQueryArgs();
	        domainArgs.setValue("domain", domainName);

	        //Need to add the the password stream to the realm
	        String passwordLine = getHtdigest(domainAuth.getRealm(), domainAuth.getUserName(), domainAuth.getPassword());

	        System.out.println(">>>>>> Add a user to the realm administrators <<<<<<"   );

	        //update rather than append to make sure we have exactly what we expect
	        //since we're updating, make sure we put the right headers on the _administrators stream
            ScspAuthorization authz = new ScspAuthorization();
            authz.addAuthorization(null, domainName); //set up domainName as the authentication realm for all HTTP methods (can use ScspAuthorization.ALL_OP instead of null)
	        ScspHeaders adminHeaders = new ScspHeaders();
	        adminHeaders.addValue("Castor-Authorization", authz.getAuthSpec()); // using ScspAuthorization ensures proper formatting of the auth specification
	        adminHeaders.addValue("Castor-Stream-Type", "admin");
	        adminHeaders.addLifepoint(null, null, 2);

	        //the castor-authorization on _administrators is <domainName>, which has no users set, so we have to update using admin credentials
	        adminHeaders.setAuthentication(adminAuth);
	        ScspQueryArgs adminArgs = new ScspQueryArgs();
	        adminArgs.setValue("admin","yes");
	        ByteArrayInputStream passwordStream = new ByteArrayInputStream(passwordLine.getBytes());
	        ScspResponse bwResponse = client.updateMutable("", "_administrators", passwordStream, passwordLine.getBytes().length, adminArgs, adminHeaders);
	        System.out.println(bwResponse);

	        System.out.println(">>>>>> Info Domain without Credentials<<<<<<"   );
	        ScspResponse response = client.info("", "", domainArgs, new ScspHeaders());
	        System.out.println(response); //should come back with a 401 on localusers.realm/_administrators

	        System.out.println(">>>>>> Info Domain with Admin Credentials<<<<<<");
	        adminHeaders = new ScspHeaders();
	        adminHeaders.setAuthentication(adminAuth);
	        adminArgs = new ScspQueryArgs(); //can only query the realm with domain=
	        adminArgs.setValue("admin","yes");
	        adminArgs.setValue("domain", domainName);
	        response = client.info("", "", adminArgs, adminHeaders);
	        System.out.println(response); //should come back with a 200

	        System.out.println(">>>>>> Info Domain with Local Credentials<<<<<<"   );
	        ScspHeaders domainHeaders = new ScspHeaders();
	        domainHeaders.setAuthentication(domainAuth);
	        response = client.info("", "", domainArgs, domainHeaders);
	        System.out.println(response); //should come back with a 200

	        System.out.println(">>>>>> Create an unprotected bucket '" + unprotectedBucketName + "'<<<<<<");
	        // the C-A header set on the domain by the console includes post=<domainName>, and we haven't added any users to the domain stream, so
	        // we can only create a bucket using admin credentials.
	        adminHeaders = new ScspHeaders();
	        adminHeaders.setAuthentication(adminAuth);
	        adminArgs = new ScspQueryArgs();
	        adminArgs.setValue("admin","yes");
	        bwResponse = client.write(unprotectedBucketName, emptyStream, emptyData.getBytes().length, adminArgs, adminHeaders);
	        System.out.println(bwResponse);

	        System.out.println(">>>>>> Info unprotected bucket <<<<<<"   );
	        response = client.info("", unprotectedBucketName, new ScspQueryArgs(), new ScspHeaders());
	        System.out.println(response);

	        String unprotectedObjectPath = unprotectedBucketName + "/named/object";
	        String protectedObjectPath = protectedBucketName + "/named/object";

	        System.out.println(">>>>>> Create named object 'named/object' in unprotected bucket '" + unprotectedBucketName + "'<<<<<<");
	        bwResponse = client.write(unprotectedObjectPath, inputStream, testData.getBytes().length, new ScspQueryArgs(), new ScspHeaders());
	        System.out.println(bwResponse);

	        System.out.println(">>>>>> Info named object 'named/object' in unprotected bucket <<<<<<"   );
	        response = client.info("",unprotectedObjectPath, new ScspQueryArgs(), new ScspHeaders());
	        System.out.println(response);

            authz = new ScspAuthorization();
            authz.addAuthorization(null, adminBucket); //set up adminBucket as the authentication realm for all HTTP methods (can use ScspAuthorization.ALL_OP instead of null)
	        ScspHeaders caHeaders = new ScspHeaders();
	        caHeaders.addValue("Castor-Authorization", authz.getAuthSpec()); // bucket auth realm is local.users/_administrators

	        System.out.println(">>>>>> Add a write-protected bucket '" + protectedBucketName + "'<<<<<<");
	        // the C-A header set on the realm by the console includes post=<realm>, and we don't have any users added to the realm stream, so
	        // we can only create a bucket using admin credentials.
            authz = new ScspAuthorization();
            authz.addAuthorization(ScspAuthorization.POST_OP, adminBucket); //set up adminBucket as the authentication realm for POST
	        adminHeaders = new ScspHeaders();
	        adminHeaders.addValue("Castor-Authorization", authz.getAuthSpec()); // bucket writes require localuser.realm/_administrator credentials
	        adminHeaders.setAuthentication(adminAuth);
	        adminArgs = new ScspQueryArgs();
	        adminArgs.setValue("admin","yes");
	        bwResponse = client.write(protectedBucketName, emptyStream, emptyData.getBytes().length, adminArgs, adminHeaders);
	        System.out.println(bwResponse);

	        System.out.println(">>>>>> Add named object 'named/object' to '" + protectedBucketName + "' without credentials <<<<<<");
	        bwResponse = client.write(protectedObjectPath, emptyStream, emptyData.getBytes().length, new ScspQueryArgs(), new ScspHeaders());
	        System.out.println(bwResponse); //should fail with a 401

	        System.out.println(">>>>>> Add named object 'named/object' to '" + protectedBucketName + "' with credentials <<<<<<");
	        domainHeaders = new ScspHeaders();
	        domainHeaders.setAuthentication(domainAuth);
	        bwResponse = client.write(protectedObjectPath, emptyStream, emptyData.getBytes().length, new ScspQueryArgs(), domainHeaders);
	        System.out.println(bwResponse); //should succeed with a 201

	        System.out.println(">>>>>> Info named object 'named/object' without credentials<<<<<<"   );
	        response = client.info("", protectedObjectPath, new ScspQueryArgs(), new ScspHeaders());
	        System.out.println(response);

	        System.out.println(">>>>>> Read named object 'named/object' <<<<<<");
	        response = client.read("", protectedObjectPath, outputStream, new ScspQueryArgs(), new ScspHeaders());
	        System.out.println(response);
	        System.out.println(outputStream.toString());

	        System.out.println(">>>>>> Update named object 'named/object' <<<<<<");
	        response = client.updateMutable("", protectedObjectPath, updateInputStream, testUpdateData.getBytes().length, new ScspQueryArgs(), new ScspHeaders());
	        System.out.println(response);

	        System.out.println(">>>>>> Append named object 'named/object' <<<<<<");
	        response = client.appendMutable("", protectedObjectPath, appendInputStream, testAppendData.getBytes().length, new ScspQueryArgs(), new ScspHeaders());
	        System.out.println(response);

	        System.out.println(">>>>>> Copy named object 'named/object' <<<<<<");
	        response = client.copyMutable("", protectedObjectPath, new ScspQueryArgs(), new ScspHeaders());
	        System.out.println(response);

	        System.out.println(">>>>>> Info named object 'named/object' <<<<<<");
	        response = client.info("", protectedObjectPath, new ScspQueryArgs(), new ScspHeaders());
	        System.out.println(response);

	        //Important! We want to be able to rerun this example, so we have to delete the named streams we created.

	        System.out.println(">>>>>> Delete named object 'named/object' <<<<<<");
	        response = client.delete("", protectedObjectPath, new ScspQueryArgs(), new ScspHeaders());
	        System.out.println(response);

	        System.out.println(">>>>>> Delete named object 'named/object' <<<<<<");
	        response = client.delete("", unprotectedObjectPath, new ScspQueryArgs(), new ScspHeaders());
	        System.out.println(response);

	        System.out.println(">>>>>> Delete bucket '" + unprotectedBucketName + "' <<<<<<" );
	        response = client.delete("", unprotectedBucketName, new ScspQueryArgs(), new ScspHeaders()); //note that no auth is required
	        System.out.println(response);

	        System.out.println(">>>>>> Delete bucket '" + protectedBucketName + "' <<<<<<");
	        response = client.delete("", protectedBucketName, new ScspQueryArgs(), domainHeaders); //note only domain auth is required
	        System.out.println(response);
        }
        catch (Exception ex) {
            ex.printStackTrace();
            client.setHostHeaderValue("");
        }
    }
    //[[DOC SNIPPET END]]

	static void RunAllUsersTenantExamples(ScspClient client) {
	    System.out.println("\r\n>>>>> All Users Tenant Examples <<<<<<<\r\n\r\n");
	    if (CheckAllUsersTenantExamples(client, "testbucket") && CheckAllUsersTenantExamples(client, "authbucket")) {
	        RunAllUsersTenantExampleNoAuth(client, "testbucket");
	        RunAllUsersTenantExampleWithAuth(client, "authbucket");
	    }
	}

	static void RunOnlyTenantUsersTenantExamples(ScspClient client) {
	    System.out.println("\r\n>>>>> Only Tenant Domain Users Tenant Examples <<<<<<<\r\n\r\n");
	    if (CheckOnlyTenantUsersTenantExamples(client)) {
	        RunOnlyTenantUsersTenantExample(client);
	    }
	}

    static void RunExamples(ScspClient client) {
		RunMutableExamples(client);
		RunImmutableExamples(client);
    	RunLifepointExamples(client);
		RunCacheCoherencyExamples(client);
        RunIntegritySealExamples(client);
        RunReadExamples(client);
        RunInfoExamples(client);
        RunStatusExamples(client);
        RunAllUsersTenantExamples(client);
        RunOnlyTenantUsersTenantExamples(client);
    }
}
