/*
 * Copyright (c) 2007 by Caringo, Inc. -- All rights reserved
 * Caringo Confidential and Proprietary
 */
package com.caringo.client.locate;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.Properties;

/**
 * Class: Locator <br>
 * Package: com.caringo.client.locate<br>
 * Description: LocatorFactory is a factory for Locator instances.  It can initialize Locators from Properties objects as well as Files.<br>
 * <br>
 * Copyright (c) 2008 by Caringo, Inc. -- All rights reserved<br>
 * This is free software, distributed under the terms of the New BSD license.<br>
 * See the LICENSE.txt file included in this archive.<br>
 * <br>
 * 
 * @author jmike
 * @created Jan 27, 2008
 * @id $Id: $
 */
public abstract class LocatorFactory {

    public static final String PROP_TYPE = "type";
    public static final String PROP_HOST = "host";
    public static final String PROP_PORT = "port";

    public static final String TYPE_ROUNDROBIN = "roundrobin";
    public static final String TYPE_STATIC = "static";
    public static final String TYPE_ZEROCONF = "zeroconf";

    public static final String DEFAULT_PORT = "80";

    /**
     * Construct a <code>Locator</code> based on settings from a properties file.
     * 
     * @param propsFile
     *            - The properties file.
     * @return - A <code>Locator</code> object.
     * @throws IOException If an input or output exception occurs.
     */
    public static Locator constructFromProperties(File propsFile) throws IOException {
        return constructFromProperties(propsFile, "");
    }

    /**
     * Construct a Locator based on setting from a properties file.
     * 
     * @param propsFile
     *            - The properties file.
     * @param keyPrefix
     *            - The prefix with which to prepend the config keys. This is useful
     *            when a config file is shared by multiple things (i.e. rather
     *            than "type", if the prefix is "locate.", the key used would be
     *            "locate.type".
     * @return - A <code>Locator</code> object.
     * @throws IOException If an input or output exception occurs.
     */
    public static Locator constructFromProperties(File propsFile, String keyPrefix) throws IOException {
        Properties props = new Properties();
        props.load(new FileInputStream(propsFile));
        return constructFromProperties(props, keyPrefix);
    }

    /**
     * Construct a Locator based on setting from a properties file.
     * 
     * @param props
     *            -The <code>Properties</code> instance.
     * @return A <code>Locator</code> object.
     */
    public static Locator constructFromProperties(Properties props) throws IOException {
        return constructFromProperties(props, "");
    }

    /**
     * Construct a Locator based on setting from a properties file.
     * 
     * @param props
     *            - The <code>Properties</code> instance.
     * @param keyPrefix
     *            - The prefix with which to prepend the config keys. This is useful
     *            when a config file is shared by multiple things (i.e. rather
     *            than "type", if the prefix is "locate.", the key used would be
     *            "locate.type".
     * @return - A <code>Locator</code> object.
     * @throws IOException If an input or output exception occurs.
     */
    public static Locator constructFromProperties(Properties props, String keyPrefix) throws IOException {
        Locator locator = null;

        String type = props.getProperty(keyPrefix + PROP_TYPE);

        if (TYPE_ZEROCONF.equalsIgnoreCase(type)) {
            String clusterName = props.getProperty(keyPrefix + "cluster.name");
            if (clusterName == null || clusterName.trim().length() == 0) {
                throw new IllegalArgumentException("\"cluster.name\" required when using zeroconf location");
            }
            locator = new ZeroconfLocator(clusterName.trim());

        } else if (TYPE_STATIC.equalsIgnoreCase(type)) {

            int port = Integer.parseInt(props.getProperty(keyPrefix + PROP_PORT, DEFAULT_PORT));
            locator = new StaticLocator();

            // iterate over "host.n"
            int i = 1;
            String host = null;
            do {
                host = props.getProperty(keyPrefix + "host." + i);
                if (host != null) {
                    ((StaticLocator) locator).addAddress(host, port);
                }
                i++;
            } while (host != null);

            // iterate over "range.n"
            i = 1;
            String range = null;
            do {
                range = props.getProperty(keyPrefix + "range." + i);
                if (range != null) {
                    String[] rangeStrs = range.split("-");
                    if (rangeStrs.length != 2) {
                        throw new IllegalArgumentException(range + " is not a valid range");
                    }
                    ((StaticLocator) locator).addIpRange(rangeStrs[0], rangeStrs[1], port);
                }
                i++;
            } while (range != null);

        } else if (TYPE_ROUNDROBIN.equalsIgnoreCase(type)) {

            String host = props.getProperty(keyPrefix + PROP_HOST);
            if (host == null) {
                throw new IllegalArgumentException("Host not specified in roundrobin locator configuration.");
            }
            int port = Integer.parseInt(props.getProperty(keyPrefix + PROP_PORT, DEFAULT_PORT));
            locator = new RoundRobinDnsLocator(host, port);

        } else {
            throw new IllegalArgumentException(type + " is not a valid locator type (only zeroconf, static, roundrobin supported)");
        }

        return locator;

    }

}
