/*
 * Copyright (c) 2007 by Caringo, Inc. -- All rights reserved<br>
 * This is free software, distributed under the terms of the New BSD license.<br>
 * See the LICENSE.txt file included in this archive.<br>
 * <br>
 */
package com.caringo.client.locate;

import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.UnknownHostException;

import org.apache.log4j.Logger;

/**
 * <code>RoundRobinResolver</code> extends the <code>StaticLocator</code> for use with round-robin DNS,
 * acquiring IP addresses from the DNS server rather than from a config file.
 * The DNS name is resolved at start time, and then never again, unless the
 * locator's pool is exhausted. While this means that DNS changes are not
 * reflected in a running <code>Locator</code> unless the pool is exhausted, it is expected
 * that any new nodes in a cluster will be discovered via SCSP
 * redirection/feedback, so this is not a shortcoming in expected use. <br>
 * <br>
 * Copyright (c) 2007 by Caringo, Inc. -- All rights reserved<br>
 * This is free software, distributed under the terms of the New BSD license.<br>
 * See the LICENSE.txt file included in this archive.<br>
 * <br>
 * 
 * @author jmike
 * @created Apr 11, 2007
 * @id $Id: RoundRobinDnsLocator.java 963 2008-02-29 19:57:21Z jmike $
 */
public class RoundRobinDnsLocator extends StaticLocator {

    private static final Logger logger = Logger.getLogger(RoundRobinDnsLocator.class);

    private InetSocketAddress addr;
    private String dnsName;
    private int port;

    /** 
     * Object constructor.
     * @param dnsName - Set the DNS name.
     * @param port - set the port number.
     */
    public RoundRobinDnsLocator(String dnsName, int port) {
        this.dnsName = dnsName;
        this.port = port;
    }

    /**
     * This method needs a description.
     */
    public void start() {
        super.start();
        requery();
    }

    /**
     * This method needs a description.
     * 
     * @param addr - An <code>InetSocketAddress</code> object.
     */
    @Override
    public synchronized void foundDead(InetSocketAddress addr) {
        // If we're down to one, then the pool's probably invalid, so requery
        if (poolSize() == 1) {
            requery();
        }
        super.foundDead(addr);
    }

    /**
     * This method needs a description.
     */
    private void requery() {
        try {
            InetAddress[] nodeAddrs = InetAddress.getAllByName(dnsName);
            for (InetAddress addr : nodeAddrs) {
                foundAlive(new InetSocketAddress(addr.getHostAddress(), port));
            }
        } catch (UnknownHostException ex) {
            logger.warn("Error resolving " + dnsName + ": " + ex.getMessage());
        }
    }

    /**
     * This method needs a description.
     * 
     * @return A <code>String</code> object.
     */
    @Override
    public String toString() {
        return "RoundRobinDnsLocator (addr=" + addr + ")";
    }

}
