/*
 * 
 */
package com.caringo.client.locate;

import java.net.InetSocketAddress;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.Date;

import org.apache.log4j.Logger;

/**
 * <code>StaticLocator</code> implements <code>Locator</code> and keeps a static pool of nodes to return
 * to clients. It can be created with a list of hosts in the constructor and/or
 * use the <code>addAddress()</code> and <code>addIpRange()</code> methods to add to the list. In addition,
 * any nodes that are passed to <code>foundAlive()</code> are added to the pool. Nodes passed
 * to <code>foundDead()</code> are removed from the pool for retryTime seconds and then returned to
 * the pool.<br>
 * <br>
 * Copyright (c) 2008 by Caringo, Inc. -- All rights reserved<br>
 * This is free software, distributed under the terms of the New BSD license.<br>
 * See the LICENSE.txt file included in this archive.<br>
 * <br>
 * 
 * @author jmike
 * @created Apr 11, 2007
 * @id $Id: $
 */
public class StaticLocator implements Locator {

    private class DeadpoolEntry {
        public long restoreTime;
        public InetSocketAddress address;
        public DeadpoolEntry(long restoreTime, InetSocketAddress address) {
            this.restoreTime = restoreTime;
            this.address = address;
        }
    }

    private Set<InetSocketAddress> poolSet = new HashSet<InetSocketAddress>();
    private List<InetSocketAddress> poolList = new ArrayList<InetSocketAddress>();
    private List<DeadpoolEntry> deadPool = new ArrayList<DeadpoolEntry>();
    private int retryTime;
    /**
     * Constructor.
     * @param hosts -
     *            An array of IP addresses or host names.
     * @param port -
     *            The SCSP port number.
     * @param retryTime -
     *            The numer of seconds that an address will live in the dead pool.
     */
    public StaticLocator(String[] hosts, int port, int retryTime) {
        this.retryTime = retryTime;
        for (String host : hosts) {
            addAddress(host, port);
        }
    }

    /**
     * The default constructor.
     */
    public StaticLocator() {
    }

    /**
     * @see com.caringo.client.locate.Locator#start
     */
    public void start() {
        // nothing to do here.
    }

    /**
     * @see com.caringo.client.locate.Locator#stop()
     */
    public void stop() {
        // nothing to do here.
    }

    /**
     * Add an address
     * 
     * @param host - A host name or IP address.
     * @param port - A port number.
     */
    public void addAddress(String host, int port) {
        foundAlive(new InetSocketAddress(host, port));
    }

    /**
     * Add a range of IP addresses.
     * 
     * @param startIp - An IP address.
     * @param endIp - An IP address.
     * @param port - A port number.
     */
    public void addIpRange(String startIp, String endIp, int port) {
        int[] startBytes = toAddrInts(startIp);
        int[] endBytes = toAddrInts(endIp);
        if (Arrays.equals(startBytes, endBytes)) {
            foundAlive(new InetSocketAddress(startIp, port));
        }
        for (int i = startBytes[0]; i <= endBytes[0]; i++) {
            for (int j = startBytes[1]; j <= endBytes[1]; j++) {
                for (int k = startBytes[2]; k <= endBytes[2]; k++) {
                    for (int l = startBytes[3]; l <= endBytes[3]; l++) {
                        foundAlive(new InetSocketAddress(i + "." + j + "." + k + "." + l, port));
                    }
                }
            }
        }
    }

    /**
     * @see com.caringo.client.locate.Locator#locate()
     */
    public synchronized InetSocketAddress locate() {
        InetSocketAddress result = null;

        int memberCount = poolList.size();
        if (memberCount > 0) {
            result = poolList.get(0);
            if (memberCount > 1) {
                // rotate to spread the load
                poolList.remove(0);
                poolList.add(result);
            }
        }
        invariant();
        return result;
    }

    /**
     * @see com.caringo.client.locate.Locator#locateAll()
     */
    public InetSocketAddress[] locateAll() {
        return poolList.toArray(new InetSocketAddress[poolList.size()]);
    }

    /**
     * @see com.caringo.client.locate.Locator#foundDead(java.net.InetSocketAddress)
     */
    public synchronized void foundDead(InetSocketAddress addr) {
        if (poolSet.size() > 1) {
            if (poolSet.contains(addr)) {
                poolSet.remove(addr);
                poolList.remove(addr);
                if (retryTime > 0) {
                    deadPool.add(new DeadpoolEntry(new Date().getTime() * 1000 + retryTime, addr));
                }
            }
        } else {
            // don't remove the last entry
        }
        invariant();
    }

    /**
     * @see com.caringo.client.locate.Locator#foundAlive(java.net.InetSocketAddress)
     */
    public synchronized void foundAlive(InetSocketAddress addr) {
        if (!poolSet.contains(addr)) {
            poolList.add(addr);
            poolSet.add(addr);
            List<DeadpoolEntry> removeList = new ArrayList<DeadpoolEntry>();
            for (DeadpoolEntry e : deadPool) {
                if (e.address == addr) {
                    removeList.add(e);
                }
            }
            deadPool.removeAll(removeList);
        }
        invariant();
    }

    /**
     * Return the pool size.
     * @return The pool size.
     */
    public int poolSize() {
        return poolSet.size();
    }

    // /
    // / Support
    // /

    private void checkDeadPool() {
        ArrayList<DeadpoolEntry> removeList = new ArrayList<DeadpoolEntry>();
        long now = new Date().getTime() * 1000;
        for (DeadpoolEntry e : deadPool) {
            if (e.restoreTime < now) {
                removeList.add(e);
                if (! poolList.contains(e.address)) {
                    poolList.add(e.address);
                }
            }
        }
        deadPool.removeAll(removeList);        
    }

    private void invariant() {
        if (poolSet.size() != poolList.size()) {
            poolSet.clear();
            poolSet.addAll(poolList);
        }
    }

    private int[] toAddrInts(String ipAddr) {
        String[] ipStrs = ipAddr.split("\\.");
        if (ipStrs.length != 4) {
            throw new IllegalArgumentException(ipAddr + " is not a valid IP address.");
        }
        int[] result = new int[4];
        try {
            for (int i = 0; i < result.length; i++) {
                int ipInt = Integer.parseInt(ipStrs[i]);
                if (ipInt < 0 || ipInt > 255) {
                    throw new IllegalArgumentException(ipAddr + " is not a valid IP address.");
                }
                result[i] = ipInt;
            }
        } catch (NumberFormatException ex) {
            throw new IllegalArgumentException(ipAddr + " is not a valid IP address.");
        }

        return result;
    }

    @Override
    public String toString() {
        return "StaticLocator (" + poolSet.size() + " nodes)";
    }

    /**
     * Construct the ClientFactory with a Properties file.
     * 
     * @param cfg
     * @param poolSize
     * @throws IOException
     * @deprecated Config parsing really shouldn't be being done here.
     */

}
