package com.caringo.client.request;

import java.io.InputStream;
import java.io.IOException;
import java.lang.IllegalArgumentException;

import org.apache.commons.httpclient.HttpMethod;
import org.apache.commons.httpclient.Header;
import org.apache.commons.httpclient.params.HttpClientParams;
import org.apache.commons.httpclient.methods.EntityEnclosingMethod;
import org.apache.commons.httpclient.methods.InputStreamRequestEntity;
import org.apache.commons.httpclient.params.HttpMethodParams;

import com.caringo.client.ScspHeaders;

/**
 * Copyright (c) 2010 by Caringo, Inc. -- All rights reserved<br>
 * This is free software, distributed under the terms of the New BSD license.<br>
 * See the LICENSE.txt file included in this archive.<br>
 * <br>
 * 
 * @author pray
 * @created Jan 1, 2010
 * @id $Id: EntityEnclosingMethodFactory.java 52121 2012-08-28 17:31:56Z jafshar $
 */

abstract class EntityEnclosingMethodFactory extends MethodFactory {

    protected InputStream stream;


    protected long streamLength;
    protected String contentType;

    public EntityEnclosingMethodFactory(InputStream stream, long streamLength, String contentType, ScspHeaders headers) {
        super(headers);
        this.stream = stream;
        this.streamLength = streamLength;
        this.contentType = contentType;
    }

    /**
     * Get the EntityEnclosingMethod to be used in makeMethod
     * @return my entity enclosing method extension.
     */
    @Override
    protected HttpMethod getMyMethod() {
        return getMyEntityEnclosingMethod();
    }
    
    /**
     * Get the EntityEnclosingMethod to be used in makeMethod
     * @return my entity enclosing method extension.
     */
    protected abstract EntityEnclosingMethod getMyEntityEnclosingMethod();
    
    /**
     * get instance of my method
     */
    @Override
    HttpMethod makeMethod() throws IOException, IllegalArgumentException {
        if (0 > this.streamLength) {
            throw new IllegalArgumentException("Stream length must be non-negative");
        }

        boolean useExpectContinue = this.streamLength > 32768 || this.reqHeaders.containsName("Expect");
        long requestStreamLength = this.streamLength;
        if (requestStreamLength <= 32768) {
            // CONTENT_LENGTH_AUTO causes the EntityEnclosingMethod to buffer the whole input stream in memory so that the method can be
            // re-sent. For things like redirects this is nice. For authentication, it's necessary.
            requestStreamLength = InputStreamRequestEntity.CONTENT_LENGTH_AUTO;
        }
        stream.reset();
        EntityEnclosingMethod method = getMyEntityEnclosingMethod();
        if (contentType.length() > 0) {
            method.setRequestEntity(new InputStreamRequestEntity(stream, requestStreamLength, contentType));
        }
        else {
            method.setRequestEntity(new InputStreamRequestEntity(stream, requestStreamLength));
        }
        // there's an HttpClient 3.0.1 bug setting Expect headers. See ExpectContinueMethod.addRequestHeaders. That method discards
        // all Expect headers if USE_EXPECT_CONTINUE is false on the parameters, so we'll have to send Expect:100-continue whenever we
        // send any other expect headers. Note that this has the virtuous side-effect of preserving the Expect: 100-continue header passed
        // into the ScspClient.
        HttpMethodParams params = method.getParams();
        params.setParameter(HttpClientParams.USE_EXPECT_CONTINUE, useExpectContinue);
        //params.setExpectContinueTimeout(30000);

        HttpMethod resMethod = addHeadersToMethod(method);
        if (this.streamLength > 32768) {
            // Another HttpClient 3.0.1 bug. The method writeRequest on HttpMethodBase makes the same mistake that 
            // ExpectContinueMethod.addRequestHeaders makes. It retrieves the condensed Expect header (that is, the multi-value, comma-separated
            // list of all the Expect headers) and checks the value against 100-continue. If there are multiple Expect headers, it
            // won't detect that we've requested a 100-continue handshake, and so won't do one. To fix this, we'll have to subtype the POST
            // and PUT method classes (we already have APPEND and COPY methods) and implement a getRequestHeader that returns the specific 
            // string '100-continue' for an Expect lookup if the method has an Expect: 100 header on it.
            // Anyway, the ExpectContinueMethod code doesn't set the Expect header if there already is an Expect header, so add this header back in
            // specifically so that it will be available to the subtypes if we actually do the subtyping. Yegads.
            // Oh, yeah. Case matters to HttpClient for the Expect header name.
            resMethod.addRequestHeader(new Header("Expect", "100-continue"));
        }

        return resMethod;        
    }

}
