package com.caringo.enumerator;

import java.util.ArrayList;

import com.caringo.client.ScspHeader;
import com.caringo.client.ScspHeaders;
import com.caringo.client.ScspResponse;
import com.caringo.client.ScspResponse.ScspResultCode;


/**
 * Holds a response from an Object Enumerator.
 * Has an EnumeratorType.
 * Has an EnumeratorAction.
 * 
 * The getEntries() method returns an ArrayList<EnumeratorEntry> when EnumeratorAction equals NEXT. Otherwise, it returns null.
 * The entries returned are subclassed according to the EnumeratorType of the response. 
 * 
 * Copyright (c) 2010 by Caringo, Inc. -- All rights reserved
 * Caringo Confidential and Proprietary
 * 
 * @author jcasey
 * @created January 21, 2010
 */
public class EnumeratorResponse {

	private ScspResponse scspResponse;
	private ScspHeaders responseHeaders;
	private String statusLine;
	private String responseBody;
	private String[] nextResponseBody;
	private boolean resultCode;
	private int httpStatusCode;
	private EnumeratorType enumType;
	private EnumeratorAction action;
	private ArrayList<EnumeratorEntry> entries;;
	
	private static final String CONTENT_UUID = "Content-UUID";
	
	public EnumeratorResponse(ScspResponse response, EnumeratorType enumType, EnumeratorAction action){
		
		this.scspResponse = response;
		this.resultCode = getResultCode();
		this.httpStatusCode = response.getHttpStatusCode();
		this.statusLine = response.getStatusLine();
		this.responseHeaders = response.getResponseHeaders();
		this.enumType = enumType;
		this.action = action;
		this.entries = null;
		
		// Go ahead and read the body of the ScspResponse now
		this.responseBody = this.scspResponse.getResponseBody();
		if (this.action.equals(EnumeratorAction.ENUM_ACTION_NEXT)) {
			this.nextResponseBody = this.scspResponse.getResponseBody().split("\r\n|\r|\n");
		}
		else {
			this.nextResponseBody = null;
		}	

	}
	
	/**
	 * If the EnumeratorAction is not NEXT, returns null.
	 * 
	 * If EnumeratorAction is NEXT, populates the entries field in this class if it is null.
	 * (Otherwise, it just returns it.) The subclass of EnumeratorEntry returned depends on
	 * the EnumeratorType of this response.
	 * 
	 * @return ArrayList<EnumeratorEntry>
	 */
	public ArrayList<EnumeratorEntry> getEntries() {
		
		// If the EnumeratorAction is not NEXT then return null
		if (! this.action.equals(EnumeratorAction.ENUM_ACTION_NEXT)) {
			return null;
		}
		
		// If we have already populated the ArrayList, just return it
		if (this.entries != null) {
			return this.entries;
		}
		else {
			entries = new ArrayList<EnumeratorEntry>();
		}
		
		// At present the publisher defaults EnumeratorType to METADATA
		if (this.enumType == null) {
			this.enumType = EnumeratorType.ENUM_TYPE_METADATA;
		}

		// EnumeratorType equals EVENTS
		if (this.enumType.equals(EnumeratorType.ENUM_TYPE_EVENT)) {
			
			for (int i = 0; i < nextResponseBody.length; i++) {
				
				boolean isGood = true;
				
				String line = nextResponseBody[i];
				String[] details = line.split(",");
				
				// Get and check the UUID
				String uuid = "";
				try {
					uuid = details[0];
					if (uuid == null || uuid.length() != 32) {
						isGood = false;
					}
				}
				catch (Exception ex) {
					isGood = false;
				}
				
				// Get and check the EventType number equivalent
				int eventNumber = 0;
				try {
					eventNumber = Integer.parseInt(details[1]);
				}
				catch (Exception ex) {
					isGood = false;
				}
				
				EnumeratorEventType eventType =  EnumeratorEventType.getEventTypeByNum(eventNumber);
				if (eventType == null) {
					isGood = false;
				}
				
				// If the entry is good add it to the list
				if (isGood) {
					EnumeratorEventEntry entry = new EnumeratorEventEntry(uuid, eventType);
					this.entries.add(entry);
				}
			}
			
		}
		// EnumeratorType equals UUID
		else if (this.enumType.equals(EnumeratorType.ENUM_TYPE_UUID)) {
			
			for (int i = 0; i < nextResponseBody.length; i++) {
				
				boolean isGood = true;
				
				String line = nextResponseBody[i];
				
				// Get and check the UUID
				String uuid = line;
				if (uuid == null || uuid.length() != 32) {
					
					isGood = false;
				}
				
				// If the entry is good add it to the list
				if (isGood) {
					EnumeratorUuidEntry entry = new EnumeratorUuidEntry(uuid);
					this.entries.add(entry);
				}
			}
			
		}
		// EnumeratorType equals METADATA
		else if (this.enumType.equals(EnumeratorType.ENUM_TYPE_METADATA)) {
			
			// There is only one entry returned for an Object Enumerator with type "metadata"
			
			//  Do not add entries that are incomplete or malformed 
			boolean isGood = true;
			
			// The UUID in this case is returned as a separate header
			String uuid = "";
			
			try {
				uuid = this.scspResponse.getResponseHeaders().getHeaderValues(EnumeratorResponse.CONTENT_UUID).get(0);
			}
			catch (Exception ex) {
				isGood = false;
			}
			
			ArrayList<ScspHeader> headers = new ArrayList<ScspHeader>();
			
			// No reason to try and parse headers if we did not get a UUID
			if (isGood) {
			
				// The lines of the response body contain the metadata as headers
				// Get the headers and put them into an ArrayList for the metadata entry
				for (int i = 0; i < nextResponseBody.length; i++) {
					
					
					try {
						String line = nextResponseBody[i];
						String[] details = line.split(":");
						ScspHeader header = new ScspHeader(details[0], details[1]);
						headers.add(header);
					}
					catch (Exception ex) {
						isGood = false;
					}
				}
			}
			
			if (isGood)  {
				EnumeratorMetadataEntry entry = new EnumeratorMetadataEntry(uuid, headers);
				this.entries.add(entry);
			}
			
		}
			
		return this.entries;	
		
	}
	
	/**
	 * 
	 * @return action
	 */
	public EnumeratorAction getAction() {
		
		return this.action;
	}
	
	/**
	 * 
	 * @return type
	 */
	public EnumeratorType getEnumeratorType() {
		
		return this.enumType;
	}
	
	/**
	 * 
	 * @return httpResponseCode
	 */
	public int getHttpStatusCode() {
		
		return this.httpStatusCode;
	}
	
	/**
	 * 
	 * @return scspResponse
	 */
	public ScspResponse getScspResponse() {
		
		return this.scspResponse;
	}
	
	/**
	 * 
	 * @return responseBody
	 */
	public String getResponseBody() {
		
		return this.responseBody;
	}
	
	/**
	 * 
	 * @return responseHeaders
	 */
	public ScspHeaders getResponseHeaders() {
		
		return this.responseHeaders; 
	}
	
	/**
	 * 
	 * @return statusLine
	 */
	public String getStatusLine() {
		
		return this.statusLine;
	}
	

	/**
	 * Return ScspResultCode as boolean 
	 * 
	 * @return boolean
	 */
	public boolean getResultCode() {
		
		ScspResultCode result = this.scspResponse.getResultCode();
		if (result.equals(ScspResultCode.ScspRCSuccess)) {
			return true;
		}
		else {
			return false;
		}
		
	}
	
}
