package com.caringo.enumerator;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.TimeZone;

import org.apache.commons.httpclient.NameValuePair;

import com.caringo.client.ScspClient;
import com.caringo.client.ScspExecutionException;
import com.caringo.client.ScspHeaders;
import com.caringo.client.ScspQueryArgs;
import com.caringo.client.ScspResponse;
import com.caringo.enumerator.EnumeratorAction;
import com.caringo.enumerator.EnumeratorQueryArgs;
import com.caringo.enumerator.EnumeratorResponse;
import com.caringo.enumerator.EnumeratorType;
import com.caringo.enumerator.ObjectEnumeratorException;

import java.io.ByteArrayInputStream;

/**
 * The interface for connecting to a Content Router Publisher instance, creating an Object Enumerator session or connecting to an existing session,
 * retrieving items from the session, and ending the session.
 * 
 * start -
 * 			Start an enumerator session. Returns an EnumeratorResponse containing the UUID of the Object Enumerator.
 * next - 
 * 			Retrieve items from an enumerator session.  
 * end - 
 * 			End an enumerator session.  Calling this method closes the connection to the Content Router.
 * 
 * Copyright (c) 2010 by Caringo, Inc. -- All rights reserved
 * Caringo Confidential and Proprietary
 * 
 * @author jcasey
 * @created January 20, 2010
 */
public class ObjectEnumerator {

	private String publisherHost;
	private int publisherPort;
    private String channelName;
    private EnumeratorType enumType;
    private Date startDate;
    private Date endDate;
    private Long timeout;
    private ScspQueryArgs startScspQueryArgs;
    private ScspQueryArgs nextScspQueryArgs;
    protected ScspClient scspClient;
    private String enumUuid;
    private String enumName;
    
    private static final int DEFAULT_MAX_RETRIES = 5;
    private static final int DEFAULT_MAX_POOL_SIZE = 10;
    private static final Long DEFAULT_TIMEOUT = 90000L; // 25 hours in seconds
    private static final String CONTENT_UUID = "Content-UUID";
    private static final Long MAX_ITEMS_LIMIT = 10L * 1000L;

    /**
     * Constructor
     * 
     * @param publisherHost -
     * 			  The IP address of the Content Router Publisher instance.
     * @param publisherPort -
     *            The port number of the Content Router Publisher instance. 
     *            
     * @throws IOException 
     */
    public ObjectEnumerator(String publisherHost, int publisherPort) throws IOException {
    	
        this.publisherHost = publisherHost;
        this.publisherPort = publisherPort;
    	this.channelName = null;
        this.enumType = EnumeratorType.ENUM_TYPE_METADATA;
        this.startDate = null;
        this.endDate = null;
        this.timeout = null;
        this.enumUuid = null;
        this.enumName = null;
        this.startScspQueryArgs = new ScspQueryArgs();
        this.nextScspQueryArgs = null;
        
        String[] hosts = { publisherHost };
        this.scspClient = new ScspClient(hosts, publisherPort, DEFAULT_MAX_POOL_SIZE, DEFAULT_MAX_RETRIES);
        
        // Start the ScspClient
        try {
        	
        	this.scspClient.start();
        }
        catch (IOException ioEx) {
        	
        	throw ioEx;
        }
    }
    
    /**
     * Constructor
     * 
     * @param publisherHost -
     * 				The IP address of the Content Router Publisher instance.
     * @param publisherPort -
     * 				The port number of the Content Router Publisher instance, typically 80.  
     * @param enumType -
     * 				The type of this Object Enumerator: ENUM_TYPE_EVENT, ENUM_TYPE_UUID, or ENUM_TYPE_METADATA
     * 
     * @throws IOException, ObjectEnumeratorException 
     */
    public ObjectEnumerator(String publisherHost, int publisherPort, EnumeratorType enumType) throws IOException, ObjectEnumeratorException {
    	
    	this(publisherHost, publisherPort);
    	
    	if (enumType == null) {
    		throw new ObjectEnumeratorException("EnumeratorType argument cannot be null");
    	}
    	this.enumType = enumType;
    	this.startScspQueryArgs.setValue(EnumeratorQueryArgs.TYPE.getName(), this.enumType.getName());
    }

    /**
     * Constructor
     * 
     * @param publisherHost -
     * 			  	The IP address of the Content Router Publisher instance.
     * @param publisherPort -
     *            	The port number of the Content Router Publisher instance, typically 80.  
     * @param enumType -
     * 				The type of this Object Enumerator: ENUM_TYPE_EVENT, ENUM_TYPE_UUID, or ENUM_TYPE_METADATA
     * @param enumUuid -
     * 				The UUID (as String) of an existing Object Enumerator in a Content Router Publisher instance 
     * 
     * @throws IOException, ObjectEnumeratorException  
     */
    public ObjectEnumerator(String publisherHost, int publisherPort, EnumeratorType enumType, String enumUuid) throws IOException, ObjectEnumeratorException {
    	
    	this(publisherHost, publisherPort, enumType);
    	
    	if (enumUuid == null) {
    		throw new ObjectEnumeratorException("UUID argument cannot be null");
    	}
        this.enumUuid = enumUuid;
    }
    
    /**
     * 
     * @return enumUuid
     */
    public String getUUID() {
    	
    	return this.enumUuid;
    }
    
    /**
     * 
     * @return channelName
     */
    public String getChannelName() {
    	
    	return this.channelName;
    }
    
   
    /**
     * 
     * @return enumType
     */
    public EnumeratorType getType() {
    	
    	return this.enumType;
    }
   
    /**
     * 
     * @return startDate
     */
    public Date getStartDate() {
    	
    	return this.startDate;
    }
    
    /**
     * 
     * @return endDate
     */
    public Date getEndDate() {
    	
    	return this.endDate;
    }
    
    /**
     * 
     * @return timeout
     */
    public Long getTimeout() {
    	
    	return this.timeout;
    }
    
    /**
     * 
     * @return enumName
     */
    public String getEnumerationName(){
    	
    	return this.enumName;
    }
    
    /**
     * 
     * @return startScspQueryArgs
     */
    public ScspQueryArgs getStartScspQueryArgs() {
    	
    	return this.startScspQueryArgs;
    }
    
    /**
     * 
     * @return nextScspQueryArgs
     */
    public ScspQueryArgs getNextScspQueryArgs() {
    	
    	return this.nextScspQueryArgs;
    }

    /**
     * 
     * @return publisherHost
     */
    public String getPublisherHost() {
    	
    	return this.publisherHost;
    }
    
    /**
     * 
     * @return publisherPort
     */
    public int getPublisherPort() {
    	
    	return this.publisherPort;
    }
    
    /**
     * Begins an Object Enumerator session
     * 
     * This method accepts the subscriber channel name as its only argument and retrieves the possible query args
     * from this Object Enumerator instance's properties which will be null unless they have either been set directly
     * or the overloaded start method has been called.
     * 
     * @param channelName
     * @return EnumeratorResponse
     * 
     * @throws ObjectEnumeratorException
     */
    public EnumeratorResponse start(String channelName) throws ObjectEnumeratorException, ScspExecutionException {
    	
    	// Throw exception if Object Enumerator session already started
    	if (this.enumUuid != null){
    		throw new ObjectEnumeratorException("Enumerator session already in progress");
    	}
    	
    	if (this.startScspQueryArgs == null) {
    		this.startScspQueryArgs = new ScspQueryArgs();
    	}
    	
    	if (channelName == null) {
    		if (this.channelName == null) {
    			throw new ObjectEnumeratorException("Channel name cannot be null");
    		}
    	}
    	else {
    		this.channelName = channelName;
    	}
    	
    	// Set default timeout in query args if not already there
    	if (! startScspQueryArgs.containsName(EnumeratorQueryArgs.TIMEOUT.getName())) {
    		this.timeout = ObjectEnumerator.DEFAULT_TIMEOUT;
    		startScspQueryArgs.setValue(EnumeratorQueryArgs.TIMEOUT.getName(), timeout.toString());
    	}
    	
    	// Set enumerator type to METADATA if not already set 
    	// (The Publisher will create a metadata type unless otherwise specified.)
    	if (this.enumType == null) {
    		this.enumType = EnumeratorType.ENUM_TYPE_METADATA;
    		this.startScspQueryArgs.setValue(EnumeratorQueryArgs.TYPE.getName(), EnumeratorType.ENUM_TYPE_METADATA.getName());
    	}
    	else {
    		if (! this.startScspQueryArgs.containsName(EnumeratorQueryArgs.TYPE.getName())) {
    			this.startScspQueryArgs.setValue(EnumeratorQueryArgs.TYPE.getName(), EnumeratorType.ENUM_TYPE_METADATA.getName());
    		}
    		else {
    			this.startScspQueryArgs.remove(EnumeratorQueryArgs.TYPE.getName());
    			this.startScspQueryArgs.setValue(EnumeratorQueryArgs.TYPE.getName(), this.enumType.getName());
    		}
    	}
    	
        // Empty headers for POST (write) method
        ScspHeaders headers = new ScspHeaders();
 
        // InputStream - this can be empty but the write method does not accept a null argument
        String emptyString = "";
        ByteArrayInputStream bais = new ByteArrayInputStream(emptyString.getBytes());
        
        // Do the POST
        ScspResponse response = this.scspClient.write(this.channelName, bais, 0L, startScspQueryArgs, headers);
        
        // Check the response code
        if (response.getHttpStatusCode() != 201) {
        	throw new ObjectEnumeratorException("Object Enumerator start returned unexpected response code: " + response.getHttpStatusCode() 
        			+ "\nResponse body: " + response.getResponseBody());
        }
        
        // Set the UUID of this Object Enumerator
        if (response.getResponseHeaders().getHeaderValues(ObjectEnumerator.CONTENT_UUID).get(0).toString().length() > 0) {
        	this.enumUuid = response.getResponseHeaders().getHeaderValues(ObjectEnumerator.CONTENT_UUID).get(0).toString();
        }
        else {
        	throw new ObjectEnumeratorException("Object enumerator start did not return a UUID for the enumerator." 
        			+ "\nResponse headers: " + response.getResponseHeaders().getHeaderMap().toString());
        }
        
        return new EnumeratorResponse(response, this.enumType, EnumeratorAction.ENUM_ACTION_START);
    	
    }
    
    
    /**
     * Begins an Object Enumerator session
     * 
     * This method uses the arguments passed to it to set the Object Enumerator's properties and then calls the
     * base start method.
     * 
     * @param channelName -
     * 					String. The name of the subscriber channel for this enumerator.
     * @param enumType -
     * 					EnumeratorType. The type of the enumerator: events, UUID, or metadata. 
     * @param startDate -
     * 					Date. The start date for this enumerator. The date range for the enumerator is inclusive of the start date.
     * @param endDate -
     * 					Date. The end date for this enumerator. The date range for the enumerator is exclusive of the end date. 
     * @param timeout - 
     * 					Long. The lifespan of this enumerator in milliseconds. Defaults to 24 hours.
     * @param enumName -
     * 					String. The name of this enumerator.
     * @param queryArgs -
     * 					HashMap<String, String>. Additional query arguments or this enumerator.
     * 
     * @return EnumeratorResponse
     * 
     * @throws ObjectEnumeratorException
     */
    public EnumeratorResponse start(String channelName, EnumeratorType enumType, Date startDate, Date endDate, 
    		Long timeout, String enumName, HashMap<String, String> queryArgs) throws ObjectEnumeratorException, ScspExecutionException {
    	
    	// Throw exception if Object Enumerator session already started
    	if (this.enumUuid != null){
    		throw new ObjectEnumeratorException("Enumerator session already in progress");
    	}
    	
    	if (this.startScspQueryArgs == null) {
    		this.startScspQueryArgs = new ScspQueryArgs();
    	}
    	
    	if (channelName == null) {
    		if (this.channelName == null) {
    			throw new ObjectEnumeratorException("Channel name cannot be null");
    		}
    	}
    	else {
    		this.channelName = channelName;
    	}
    	
    	// The Publisher defaults to type METADATA
    	if (enumType == null) {
    		if (! startScspQueryArgs.containsName(EnumeratorQueryArgs.TYPE.getName())){
    			this.enumType = EnumeratorType.ENUM_TYPE_METADATA;
    			startScspQueryArgs.setValue(EnumeratorQueryArgs.TYPE.getName(), this.enumType.getName());
    		}	
    	}
    	else {
    		this.enumType = enumType;
    		if (startScspQueryArgs.containsName(EnumeratorQueryArgs.TYPE.getName())){
    			startScspQueryArgs.remove(EnumeratorQueryArgs.TYPE.getName());
    		}
    		startScspQueryArgs.setValue(EnumeratorQueryArgs.TYPE.getName(), this.enumType.getName());
    	}

        if (startDate != null) {
        	if (this.enumType.equals(EnumeratorType.ENUM_TYPE_EVENT)) {
        		throw new ObjectEnumeratorException("Start and end dates cannot be used with Event type object enumerators");
        	}
        	else {
	        	this.startDate= startDate;
	        	if (startScspQueryArgs.containsName(EnumeratorQueryArgs.START_DATE.getName())){
	        		startScspQueryArgs.remove(EnumeratorQueryArgs.START_DATE.getName());
	        	}
	        	startScspQueryArgs.setValue(EnumeratorQueryArgs.START_DATE.getName(), getDateAsISO8601(this.startDate));
        	}
        }
        
        if (endDate != null) {
        	if (this.enumType.equals(EnumeratorType.ENUM_TYPE_EVENT)) {
        		throw new ObjectEnumeratorException("Start and end dates cannot be used with Event type object enumerators");
        	}
        	else {
	        	this.endDate = endDate;
	        	if (startScspQueryArgs.containsName(EnumeratorQueryArgs.END_DATE.getName())){
	        		startScspQueryArgs.remove(EnumeratorQueryArgs.END_DATE.getName());
	        	}
	        	startScspQueryArgs.setValue(EnumeratorQueryArgs.END_DATE.getName(), getDateAsISO8601(this.endDate));
        	}
        }
        
        if (timeout == null) {
        	if (! startScspQueryArgs.containsName(EnumeratorQueryArgs.TIMEOUT.getName())){
        		this.timeout = ObjectEnumerator.DEFAULT_TIMEOUT;
        		startScspQueryArgs.setValue(EnumeratorQueryArgs.TIMEOUT.getName(), this.timeout.toString());
        	}
        }
        else {
        	this.timeout = timeout;
        	if (startScspQueryArgs.containsName(EnumeratorQueryArgs.TIMEOUT.getName())){
        		startScspQueryArgs.remove(EnumeratorQueryArgs.TIMEOUT.getName());
        	}
        	startScspQueryArgs.setValue(EnumeratorQueryArgs.TIMEOUT.getName(), this.timeout.toString());
        }     	
        
        if (enumName != null) {
        	this.enumName = enumName;
        	if (startScspQueryArgs.containsName(EnumeratorQueryArgs.CONTEXT.getName())){
        		startScspQueryArgs.remove(EnumeratorQueryArgs.CONTEXT.getName());
        	}
        	startScspQueryArgs.setValue(EnumeratorQueryArgs.CONTEXT.getName(), this.enumName);
        }
        
        // If there are additional query arguments, add them at the end.
        if (queryArgs != null && queryArgs.size() > 0){
        	
        	// Validate additional query arguments
        	Set<Map.Entry<String, String>> mapSet = queryArgs.entrySet();
        	
        	for (Iterator<Map.Entry<String, String>> mapSetIter = mapSet.iterator(); mapSetIter.hasNext();){
        		Map.Entry<String, String> mapEntry = mapSetIter.next();
        		if (! EnumeratorQueryArgs.isValidQueryArgForStart(mapEntry.getKey())){
        			
        			throw new ObjectEnumeratorException("Invalid query argument passed to start: " + mapEntry.getKey());
        		}
        	}
        	
        	startScspQueryArgs.addAll(queryArgs);
        }
    	
        return this.start(this.channelName);
    	
    }
 
    /**
     * Overload of Object Enumerator start method that accepts additional query arguments as an ArrayList of NameValuePair
     * 
     * @param channelName -
     * 					String. The name of the subscriber channel for this enumerator.
     * @param enumType -
     * 					EnumeratorType. The type of the enumerator: events, UUID, or metadata. 
     * @param startDate -
     * 					Date. The start date for this enumerator. The date range for the enumerator is inclusive of the start date.
     * @param endDate -
     * 					Date. The end date for this enumerator. The date range for the enumerator is exclusive of the end date. 
     * @param timeout - 
     * 					Long. The lifespan of this enumerator in milliseconds. Defaults to 24 hours.
     * @param enumName -
     * 					String. The name of this enumerator.
     * @param queryArgs -
     * 					HashMap<String, String>. Additional query arguments or this enumerator.
     * 
     * @return EnumeratorResponse
     * 
     * @throws ObjectEnumeratorException
     * @throws ScspExecutionException
     */
    public EnumeratorResponse start(String channelName, EnumeratorType enumType, Date startDate, Date endDate, 
    		Long timeout, String enumName, ArrayList<NameValuePair> queryArgs) throws ObjectEnumeratorException, ScspExecutionException {
    	
    	HashMap<String, String> queryArgsMap = this.fromArrayList(queryArgs);
    	return start(channelName, enumType, startDate, endDate, timeout, enumName, queryArgsMap);
    }    
    
    /**
     * Retrieves items from an Object Enumerator Session
     * 
     * @param maxItems -
     * 					The number of items to return from this call (Note: If the enumerator type is metadata, then this argument 
     * 					is ignored and the call returns either 1 or 0 entries.)
     * @param queryArgs -
     * 					Additional query arguments to pass to the Object Enumerator.
     * 
     * @return EnumeratorResponse
     * 
     * @throws ObjectEnumeratorException
     * @throws ScspExecutionException
     */
    public EnumeratorResponse next(Long maxItems, HashMap<String, String> queryArgs) throws ObjectEnumeratorException, ScspExecutionException{
    	
    	// Get out of here if the enumerator session has not been started
    	if (this.enumUuid == null) {
    		throw new ObjectEnumeratorException("The Object Enumerator must be started before calling next.");
    	}
    	
    	// Reset scspQueryArgs
    	nextScspQueryArgs = new ScspQueryArgs();
    	
        // Add maxItems to query arguments
    	if (maxItems != null && !(this.enumType.equals(EnumeratorType.ENUM_TYPE_METADATA))) {
    		
    		// Throw exception if maxItem exceeds the hard limit
    		if (maxItems > MAX_ITEMS_LIMIT) {
    			throw new ObjectEnumeratorException("The maximum number of items that can be returned by the next method is 10,000.");
    		}
    		nextScspQueryArgs.setValue(EnumeratorQueryArgs.MAX_ITEMS.getName(), maxItems.toString());
    	}
    	
    	// If there are additional query arguments, add them at the end.
        if (queryArgs != null && queryArgs.size() > 0){
        	
        	// Validate additional query arguments
        	Set<Map.Entry<String, String>> mapSet = queryArgs.entrySet();
        	
        	for (Iterator<Map.Entry<String, String>> mapSetIter = mapSet.iterator(); mapSetIter.hasNext();){
        		Map.Entry<String, String> mapEntry = mapSetIter.next();
        		if (! EnumeratorQueryArgs.isValidQueryArgForNext(mapEntry.getKey())){
        			
        			throw new ObjectEnumeratorException("Invalid query argument passed to next: " + mapEntry.getKey());
        		}
        	}
        	        	
        	nextScspQueryArgs.addAll(queryArgs);
        }
    	
    	// Empty ScspHeaders object for GET (read) method
    	ScspHeaders headers = new ScspHeaders();
    	
    	// Do the GET 
    	// According to Pat the null arguments should work here.
    	ScspResponse response = this.scspClient.read(this.enumUuid.toString(), null, null, nextScspQueryArgs, headers);
    	
    	if (response.getHttpStatusCode() != 200){    		
    		if (response.getHttpStatusCode() == 404) {   			
    			throw new ObjectEnumeratorException("Object Enumerator next did not find enumerator with this UUID: " + this.enumUuid.toString());
    		}
    		else {    			
    			throw new ObjectEnumeratorException("Object Enumerator next returned an unexpected response code: " + response.getHttpStatusCode() 
    					+ "\nResponse body: " + response.getResponseBody());
    		}
    	}
    	
    	return new EnumeratorResponse(response, this.getType(), EnumeratorAction.ENUM_ACTION_NEXT);
  	
    }
    
    /**
     * Overload of Object Enumerator next that accepts additional query arguments as an ArrayList of NameValuePair
     * 
     * @param maxItems
     * @param queryArgs
     * 
     * @return EnumeratorResponse
     * 
     * @throws ObjectEnumeratorException
     * @throws ScspExecutionException
     */
    public EnumeratorResponse next(Long maxItems, ArrayList<NameValuePair> queryArgs) throws ObjectEnumeratorException, ScspExecutionException{
    	
    	HashMap<String, String> queryArgsMap = this.fromArrayList(queryArgs);
    	
    	return next(maxItems, queryArgsMap);
    	
    }
    
    /**
     * End an Object Enumerator session.
     * 
     * @return EnumeratorResponse
     * 
     * @throws ObjectEnumeratorException
     * @throws ScspExecutionException
     */
    public EnumeratorResponse end() throws ObjectEnumeratorException, ScspExecutionException {
    	
    	if (this.enumUuid == null) {
    		throw new ObjectEnumeratorException("Cannot end an Object Enumerator session that is not established.");
    	}
   	
    	// Empty query args object for DELETE
    	ScspQueryArgs queryArgs = new ScspQueryArgs();
    	
    	// Empty headers object for DELETE
    	ScspHeaders headers = new ScspHeaders();
    	
    	// Do the DELETE
    	ScspResponse response = this.scspClient.delete(this.enumUuid.toString(), null, queryArgs, headers);
    	
    	if (response.getHttpStatusCode() != 200) {
    		throw new ObjectEnumeratorException("Object Enumerator end returned an unexpected response code: " + response.getHttpStatusCode());
    	}
    	
    	// Reset the Object Enumerator properties to their initial state
    	this.enumUuid = null;
    	this.channelName = null;
    	this.startDate = null;
    	this.endDate = null;   	
    	this.timeout = null;
    	this.enumName = null;
    	this.startScspQueryArgs = null;
    	this.nextScspQueryArgs = null;
    	this.scspClient.stop();
    	this.scspClient = null;
    	
    	EnumeratorType oldType = this.enumType;
    	this.enumType = null;
    	
    	return new EnumeratorResponse(response, oldType, EnumeratorAction.ENUM_ACTION_END);
    	
    }
    
    /**
     * Return Date as a string in the ISO8601 format
     * 
     * @param date
     * 
     * @return ISO8601 converted date string
     */
    protected static String getDateAsISO8601(Date date) {
    	
        SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss");
        TimeZone tz = TimeZone.getTimeZone("GMT");
        sdf.setTimeZone(tz);
        return sdf.format(date);
    }
    
    /**
     * Return the HashMap type used by ScspClient from an ArrayList of NameValuePairs
     * 
     * @param list
     * 
     * @return map of header names to values
     */
    protected HashMap<String, String> fromArrayList(ArrayList<NameValuePair> list) {
    	
    	HashMap<String, String> aMap = new HashMap<String, String>();
    	for (Iterator<NameValuePair> listIterator = list.iterator(); listIterator.hasNext();){
    		NameValuePair listEntry = listIterator.next();
    		aMap.put(listEntry.getName(), listEntry.getValue());
    	}
    	
    	return aMap;
    }
    
}

