/**
 * 
 */
package com.caringo.enumerator.examples;

import com.caringo.enumerator.*;
import com.caringo.client.*;
import java.io.*;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.ArrayList;
import java.util.Map;
import java.util.Set;

/**
 * Copyright (c) 2010 by Caringo, Inc. -- All rights reserved<br>
 * This is free software, distributed under the terms of the New BSD license.<br>
 * See the LICENSE.txt file included in this archive.<br>
 * <br>
 * 
 * @author pray
 * @created August 2, 2010
 * @id $Id: $
 */
 
//To run this example, you must have two Content Router components running: the Publisher and the Replicator.
//These components can run on the same host or on different hosts; in any case, they must be running and
//able to communicate with each other.
 
//[[DOC SNIPPET START | ObjectEnumerator Examples | Enumerate UUIDs using Content Router Publisher]]
public final class ObjectEnumeratorExample {
    static String publisherHost = "192.168.1.128"; // must match the 'ipaddress' configuration parameter on Publisher
    static int publisherPort = 8080; // must match the 'publicationServerPort' configuration parameter on Publisher
    static String existingEnumUUID = "119a86158e84600256be7085017aeb08"; //previously created type=UUID enumerator
    static String channelName = "ReplicateAll";
	/**
	 * @param args
	 */
	public static void main(String[] args) {
        RunObjectEnumeratorExamples();
    }
    static void RunObjectEnumeratorExamples() {

        if (publisherHost == "") {
            System.out.println("Please set publisherHost in the example file");
            return;
        }

        try {
            /*
            * Constructors
            */
            ObjectEnumerator objEnum1;
            ObjectEnumerator objEnum2;
            ObjectEnumerator objEnum3;

            // Base 
            // Client's EnumeratorType will be Metadata by default.
            // Use try..catch because the constructor can throw an IOException if a connection to the Publisher cannot be established

            try {
                objEnum1 = new ObjectEnumerator(publisherHost, publisherPort);
            }
            catch (IOException ex) {
                throw ex;
            }

            // With EnumeratorType 
            EnumeratorType enumType = EnumeratorType.ENUM_TYPE_UUID;
            try {
                objEnum2 = new ObjectEnumerator(publisherHost, publisherPort, enumType);
            }
            catch (IOException ex) {
                throw ex;
            }

            // With EnumeratorType and UUID (The type and UUID of an enumerator that already exists on the Content Router Publisher.)
            EnumeratorType enumTypeEvents = EnumeratorType.ENUM_TYPE_EVENT;
            
            //With EnumeratorType and UUID (UUID of an enumerator that already exists on the Content Router Publisher).
            // This is used to resume an older enumeration session which had not been completed.
            try {
                objEnum3 = new ObjectEnumerator(publisherHost, publisherPort, enumTypeEvents, existingEnumUUID);
            }
            catch (IOException ex) {
                throw ex;
            }

            /*
            * Start method
            */
            EnumeratorResponse response1;
            EnumeratorResponse response2;
            EnumeratorResponse response3;

            //Only Mandatory argument is the name of a channel that 
            //exists on a Content Router Publisher instance.
            //An Enumerator response object (outlined below) is returned. The Enumerator 
            //object's enumUUID attribute is populated with UUID of the Enumeration Session 
            //created on Content Router Publisher Server.
            try {
                response1 = objEnum1.start(channelName);
            }
            catch (Exception ex) {
                throw ex;
            }

            // Because the EnumeratorType was not specified explicitly, the enumerator started by objEnum1
            // will be the Metadata type.

            // After the start method is called, the ObjectEnumerator instance will contain the UUID of the enumerator that was started.
            System.out.println("Enumerator #1 UUID: " + objEnum1.getUUID());

            // Named arguments that can be passed to the start method
            EnumeratorType enum2Type = EnumeratorType.ENUM_TYPE_UUID;
            Date startDate = new Date(); // Streams created after this datetime, inclusive (NOTE: The start query argument cannot be used with an EVENTS type enumerator.)
            Date endDate = new Date(); // Streams created before this datetime, exclusive (NOTE: This end query argument cannot be used with an EVENTS type enumerator.)
            String contextName = "enum1"; // This name will be displayed in the Publisher console.
            Long timeout = 90000L; // 25 hours in seconds

            // Other valid arguments can be passed to the start method in a HashMap<String, String> or ArrayList<NameValuePair>
            // The enum class EnumeratorQueryArgs contains all of the valid query arguments that can be used an ObjectEnumerator.
            String enumVersion = "Version 1.1";
            HashMap<String, String> queryArgs = new HashMap<String, String>();
            queryArgs.put(EnumeratorQueryArgs.VERSION.getName(), enumVersion);

            // Start enumerator #2 with the named query arguments, and the additional query argument contained in the HashMap
            try {
                response2 = objEnum2.start(channelName, enum2Type, startDate, endDate, timeout, contextName, queryArgs);
            }
            catch (Exception ex) {
                throw ex;
            }
            // The UUID of the second enumerator will be available after calling the start method
            System.out.println("Enumerator #2 UUID: " + objEnum2.getUUID());
            // The named arguments are available from the client after the start method is called
            System.out.println("Enumerator #2 start date: " + objEnum2.getStartDate());
            // All of the query arguments that were sent to the Publisher with the start method are available in an ScspQueryArgs object
            ScspQueryArgs scspQueryArgs = objEnum2.getStartScspQueryArgs();

            // Calling the start method on an already started enumerator will result in the client throwing an ObjectEnumeratorException
            // Expect this to throw:
            try {
                response3 = objEnum3.start(channelName);
            }
            catch (Exception ex){
                ex.printStackTrace();
            }

            /*
            * Next method
            */ 
            // Note that METADATA enumerators only return one item at a time regardless of the maxItems argument.
            // Set the maximum number of items to be returned from an enumerator.
            // (The upper limit on the numbers of items returned in one call to next is 10,000.)
  
            Long maxItems = 100L;
            HashMap<String, String> mapQueryArgs = new HashMap<String, String>();
            EnumeratorResponse nextResponse;

            // Call next on the second enumerator.  Its type is UUID.
            try {
                nextResponse = objEnum2.next(maxItems, mapQueryArgs);
            }
            catch (Exception ex){
                throw ex;
            }

            // Iterate through the UUID list returned
            for (Iterator<EnumeratorEntry> responseIterator = nextResponse.getEntries().iterator(); responseIterator.hasNext();) {
                EnumeratorUuidEntry uuidEntry = (EnumeratorUuidEntry) responseIterator.next();
                System.out.println("UUID in list: " + uuidEntry.getUuid());
            }

            // Call next on the third enumerator. This is the enumerator that was already started on the Publisher in the example above. Its type is EVENTS.
            try {
                nextResponse = objEnum3.next(maxItems, queryArgs);
            }
            catch (Exception ex) {
                throw ex;
            }

            // Iterate through the events list returned
            for (Iterator<EnumeratorEntry> responseIterator = nextResponse.getEntries().iterator(); responseIterator.hasNext();) {
                EnumeratorEventEntry eventEntry = (EnumeratorEventEntry) responseIterator.next();
                System.out.println("UUID  [" + eventEntry.getUuid() + "], " + eventEntry.getEventType().getNumber());
            }

            // Call next on the first enumerator. Its type is METADATA.
            try {
                nextResponse = objEnum1.next(maxItems, queryArgs);
            }
            catch (Exception ex) {
                throw ex;
            }

            //Iterate with a Client enumerator

            //Note that METADATA enumerators only return one item at a time regardless of the maxItems argument.
            //The upper limit on the numbers of items returned in one call to next is 10,000.

            //Manadatory arguments only (default number of entries returned for UUID and Event enumeration is 10).
            for (Iterator<EnumeratorEntry> responseIterator = nextResponse.getEntries().iterator(); responseIterator.hasNext();) {
                EnumeratorMetadataEntry metadataEntry = (EnumeratorMetadataEntry) responseIterator.next();
                System.out.println("UUID  for metadata enumerator: " + metadataEntry.getUuid());
                // A metadata entry contains a list of all the headers for the stream identified by its UUID
                for (Iterator<ScspHeader> headerIterator = metadataEntry.getScspHeaders().iterator(); headerIterator.hasNext();) {
                    ScspHeader header = headerIterator.next();
                    System.out.println(header.getName() + ": " + header.getValue());
                }
            }

            /*
            * End method
            */
            // Calling end on an enumerator removes the enumerator from the Publisher and resets the client's internal state.
            EnumeratorResponse endResponse;
            try {
                endResponse = objEnum1.end();
            }
            catch (Exception ex) {
                throw ex;
            }

        }
        catch (Exception ex) {
            ex.printStackTrace(); // might as well continue, we're going down regardless
        }
	}
}
//[[DOC SNIPPET END]]
